Package: webfakes
Title: Fake Web Apps for HTTP Testing
Version: 1.5.0
Authors@R: c(
    person("Gábor", "Csárdi", , "csardi.gabor@gmail.com", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49")),
    person(, "Civetweb contributors", role = "ctb",
           comment = "see inst/credits/ciwetweb.md"),
    person(, "Redoc contributors", role = "ctb",
           comment = "see inst/credits/redoc.md"),
    person("L. Peter", "Deutsch", role = "ctb",
           comment = "src/md5.h"),
    person("Martin", "Purschke", role = "ctb",
           comment = "src/md5.h"),
    person(, "Aladdin Enterprises", role = "cph",
           comment = "src/md5.h"),
    person("Maëlle", "Salmon", , "maelle.salmon@yahoo.se", role = "ctb",
           comment = c(ORCID = "0000-0002-2815-0399"))
  )
Description: Create a web app that makes it easier to test web clients
    without using the internet. It includes a web app framework with path
    matching, parameters and templates. Can parse various 'HTTP' request
    bodies. Can send 'JSON' data or files from the disk. Includes a web
    app that implements the 'httpbin.org' web service.
License: MIT + file LICENSE
URL: https://webfakes.r-lib.org/, https://github.com/r-lib/webfakes
BugReports: https://github.com/r-lib/webfakes/issues
Depends: R (>= 3.6)
Imports: stats, tools, utils
Suggests: brotli, callr, covr, curl, digest, glue, httpuv, httr,
        jsonlite, processx, testthat (>= 3.0.0), withr, xml2, zip (>=
        2.3.0)
Biarch: true
Config/Needs/website: tidyverse/tidytemplate
Config/testthat/edition: 3
Config/usethis/last-upkeep: 2025-04-28
Encoding: UTF-8
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-05-16 16:57:00 UTC; gaborcsardi
Author: Gábor Csárdi [aut, cre],
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>),
  Civetweb contributors [ctb] (see inst/credits/ciwetweb.md),
  Redoc contributors [ctb] (see inst/credits/redoc.md),
  L. Peter Deutsch [ctb] (src/md5.h),
  Martin Purschke [ctb] (src/md5.h),
  Aladdin Enterprises [cph] (src/md5.h),
  Maëlle Salmon [ctb] (ORCID: <https://orcid.org/0000-0002-2815-0399>)
Maintainer: Gábor Csárdi <csardi.gabor@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-16 21:50:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-17 02:15:11 UTC; unix
Archs: webfakes.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(format,webfakes_app)
S3method(format,webfakes_app_process)
S3method(format,webfakes_regexp)
S3method(format,webfakes_request)
S3method(format,webfakes_response)
S3method(print,webfakes_app)
S3method(print,webfakes_app_process)
S3method(print,webfakes_regexp)
S3method(print,webfakes_request)
S3method(print,webfakes_response)
export(git_app)
export(http_time_stamp)
export(httpbin_app)
export(local_app_process)
export(mw_cgi)
export(mw_cookie_parser)
export(mw_etag)
export(mw_json)
export(mw_log)
export(mw_multipart)
export(mw_range_parser)
export(mw_raw)
export(mw_static)
export(mw_text)
export(mw_urlencoded)
export(new_app)
export(new_app_process)
export(new_regexp)
export(oauth2_httr_login)
export(oauth2_login)
export(oauth2_resource_app)
export(oauth2_third_party_app)
export(server_opts)
export(tmpl_glue)
useDynLib(webfakes, .registration = TRUE, .fixes = "c_")

<!-- README.md is generated from README.Rmd. Please edit that file -->

# webfakes

> Your own web server for happy HTTP testing

<!-- badges: start -->

[![R build
status](https://github.com/r-lib/webfakes/workflows/R-CMD-check/badge.svg)](https://github.com/r-lib/webfakes/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/webfakes)](https://CRAN.R-project.org/package=webfakes)
[![R-CMD-check](https://github.com/r-lib/webfakes/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/r-lib/webfakes/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/r-lib/webfakes/graph/badge.svg)](https://app.codecov.io/gh/r-lib/webfakes)
<!-- badges: end -->

Lightweight fake web apps for testing. Built using the
[civetweb](https://github.com/civetweb/civetweb) embedded web server.

## Features

-   Complete web app framework, define handlers for HTTP requests in R.
-   Write your own app for your custom test cases; our use app similar
    to the `https://httpbin.org` API, so often you don’t need to write
    your own web app (e.g. if you are writing an HTTP client (httr,
    curl, crul).
-   Run one web app per test suite, per test file or per test case.
-   Flexible path matching, with parameters and regular expressions.
-   Built in templating system using glue or bring your own template
    engine.
-   Middleware to parse JSON, multipart and URL encoded request bodies.
-   A web app is just an R object. It can be saved to disk, copied to
    another R process, etc.
-   A web app is extensible, by adding new routes and middleware to it.
-   Helper functions for sending JSON, files from disk, etc.
-   App-specific environment to store any data including data from
    requests to the fake app.
-   After a web app is launched from R, you can interact with it from R
    but also from the command line, your browser, etc. Nice for
    debugging.
-   The web server runs in the R process, so it has no problems with
    local firewalls.
-   Multi-threaded web server supports concurrent HTTP requests.
-   Limit download speed to simulate low bandwidth.

## Optional dependencies

-   The jsonlite package is needed for the `mw_json()` middleware, the
    `response$send_json()` method and the `httpbin_app()` app.
-   The glue package is needed for the `tmpl_glue()` template engine.
-   The callr package is needed for `new_app_process()` and
    `local_app_process` to work.
-   The `/brotli` endpoint of `httpbin_app()` needs the brotli package.
-   The `/deflate` endpoint of `httpbin_app()` needs the zip package.
-   The `/digest-auth` endpoint of `httpbin_app()` needs the digest
    package.
-   `git_app()` requires the processx package.

## Installation

Install the release version from CRAN:

``` r
install.packages("webfakes")
```

If you need the development version of the package, install it from
GitHub:

``` r
pak::pak("r-lib/webfakes")
```

## Usage

Start a web app at the beginning of your tests or test file, and stop it
after. Here is an example with the testthat package. Suppose you want to
test that your `get_hello()` function can query an API:

`local_app_process()` helps you clean up the web server process after
the test block, or test file. It is similar to the `withr::local_*`
functions.

``` r
app <- webfakes::new_app()
app$get("/hello/:user", function(req, res) {
  res$send(paste0("Hello ", req$params$user, "!"))
})
web <- webfakes::local_app_process(app)

test_that("can use hello API", {
  url <- web$url("/hello/Gabor")
  expect_equal(get_hello(url), "Hello Gabor!")
})
```

When testing HTTP clients you can often use the built in
`httpbin_app()`:

``` r
httpbin <- webfakes::local_app_process(webfakes::httpbin_app())
```

``` r
test_that("HTTP errors are caught", {
  url <- httpbin$url("/status/404")
  resp <- httr::GET(url)
  expect_error(httr::stop_for_status(resp), class = "http_404")
})
```

    #> Test passed 🌈

## Documentation

See <https://webfakes.r-lib.org>

## Links

### Other solutions for HTTP testing in R:

-   [vcr](https://github.com/ropensci/vcr)
-   [httptest](https://github.com/nealrichardson/httptest)

### R web application frameworks

webfakes focuses on testing, these packages are for writing real web
apps:

-   [shiny](https://github.com/rstudio/shiny)
-   [opencpu](https://www.opencpu.org/)
-   [plumber](https://github.com/rstudio/plumber)
-   [fiery](https://github.com/thomasp85/fiery)
-   [RestRserve](https://github.com/rexyai/RestRserve)

## Code of Conduct

Please note that the webfakes project is released with a [Contributor
Code of Conduct](https://webfakes.r-lib.org/dev/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

## License

MIT © RStudio

#' glue based template engine
#'
#' Use this template engine to create pages with glue templates.
#' See [glue::glue()] for the syntax.
#'
#' @param sep Separator used to separate elements.
#' @param open The opening delimiter. Doubling the full delimiter escapes
#'    it.
#' @param close The closing delimiter. Doubling the full delimiter escapes
#'    it.
#' @param na Value to replace NA values with. If `NULL` missing values are
#'    propagated, that is an `NA` result will cause `NA` output.
#'    Otherwise the value is replaced by the value of `na`.
#' @param transformer A function taking three parameters `code`, `envir`
#'    and `data` used to transform the output of each block before during or
#'    after evaluation.
#' @param trim Whether to trim the input template with [glue::trim()] or not.
#' @return Template function.
#'
#' @export
#' @examples
#' # See th 'hello' app at
#' hello_root <- system.file(package = "webfakes", "examples", "hello")
#' hello_root
#'
#' app <- new_app()
#' app$engine("txt", tmpl_glue())
#' app$use(mw_log())
#'
#'
#' app$get("/view", function(req, res) {
#'   txt <- res$render("test")
#'   res$
#'     set_type("text/plain")$
#'     send(txt)
#' })
#'
#' # Switch to the app's root: setwd(hello_root)
#' # Now start the app with: app$listen(3000L)
#' # Or start it in another process: new_process(app)

tmpl_glue <- function(sep = "", open = "{", close = "}", na = "NA",
                      transformer = NULL, trim = TRUE) {
  sep; open; close; na; transformer; trim
  function(path, locals) {
    txt <- readChar(path, nchars = file.size(path))
    glue::glue_data(
      locals, txt, .sep = sep, .open = open, .close = close, .na = na,
      .transformer = transformer %||% glue::identity_transformer,
      .trim = trim
    )
  }
}

library(webfakes)

app <- new_app()
app$engine("txt", tmpl_glue())
app$use(mw_log())

app$get("/hello", function(req, res) {
  res$send("Hello there!")
})

app$get(new_regexp("^/hi(/.*)?$"), function(req, res) {
  res$send("Hi indeed!")
})

app$post("/hello", function(req, res) {
  res$send("Got it, thanks!")
})

app$get("/view", function(req, res) {
  txt <- res$render("test")
  res$set_type("text/plain")$send(txt)
})

app$listen(3000L)

`%||%` <- function(l, r) if (is.null(l)) r else l

`%|NA|%` <- function(l, r) if (is_na_scalar(l)) r else l

new_object <- function(class_name, ...) {
  structure(as.environment(list(...)), class = class_name)
}

is_string <- function(x) {
  is.character(x) && length(x) == 1 && !is.na(x)
}

is_integerish <- function(x) {
  is.integer(x) || (is.numeric(x) && all(round(x) == x))
}

is_count <- function(x) {
  is_integerish(x) && length(x) == 1 && !is.na(x)
}

is_port <- function(x) {
  is_count(x)
}

is_na_scalar <- function(x) {
  is.atomic(x) && length(x) == 1 && is.na(x)
}

cat0 <- function(..., sep = "") {
  cat(..., sep = sep, append = TRUE)
}

str_trim <- function(x) {
  sub("\\s+$", "", sub("^\\s+", "", x))
}

unquote <- function(str) {
  len <- nchar(str)
  if (substr(str, 1, 1) == '"' && substr(str, len, len) == '"') {
    substr(str, 2, len - 1)

  } else {
    str
  }
}

isFALSE <- function(x) {
  identical(x, FALSE)
}

str_is_suffix <- function(x, sfx) {
  lx <- nchar(x)
  lsfx <- nchar(sfx)
  substring(x, lx - lsfx + 1, lx) == sfx
}

read_char <- function(path, encoding = "UTF-8") {
  txt <- rawToChar(readBin(path, "raw", file.info(path)$size))
  Encoding(txt) <- encoding
  txt
}

read_bin <- function(path) {
  readBin(path, "raw", file.info(path)$size)
}

try_silently <- function(expr) {
  try(expr, silent = TRUE)
}

sseq <- function(from, to) {
  if (from > to) integer() else seq(from, to)
}

is.named <- function(x) {
  length(names(x)) == length(x) && all(names(x) != "")
}

set_envvar <- function(envs) {
  if (length(envs) == 0) return()

  stopifnot(is.named(envs))

  old <- Sys.getenv(names(envs), names = TRUE, unset = NA)
  set <- !is.na(envs)

  both_set <- set & !is.na(old)

  if (any(set))  do.call("Sys.setenv", as.list(envs[set]))
  if (any(!set)) Sys.unsetenv(names(envs)[!set])

  invisible(old)
}

mkdirp <- function(path, ...) {
  dir.create(path, showWarnings = FALSE, recursive = TRUE, ...)
}

strrep <- function(x, no) {
  paste(rep(x, no), collapse = "")
}

set_name <- function(x, nm) {
  names(x) <- nm
  x
}

random_id <- generate_token <- function(n = 30) {
  paste0(sample(c(0:9, letters[1:6]), n, replace = TRUE), collapse = "")
}

parse_url <- function(url) {
  re_url <- paste0(
    "^(?<protocol>[a-zA-Z0-9]+)://",
    "(?:(?<username>[^@/:]+)(?::(?<password>[^@/]+))?@)?",
    "(?<host>[^/]+)",
    "(?<path>.*)$"            # don't worry about query params here...
  )

  re_match(url, re_url)$groups
}

modify_path <- function(url, path) {
  purl <- parse_url(url)
  has_usr <- nzchar(purl$username)
  has_pwd <- nzchar(purl$password)
  paste0(
    purl$protocol,
    "://",
    purl$username,
    if (has_pwd) ":",
    purl$password,
    if (has_usr) "@",
    purl$host,
    if (substr(path, 1, 1) != "/") "/",
    path
  )
}

has_seed <- function() {
  exists(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
}

get_seed <- function() {
  if (has_seed()) {
    get(".Random.seed", globalenv(), mode = "integer", inherits = FALSE)
  }
}

set_seed <- function(seed) {
  if (is.null(seed)) {
    rm(".Random.seed", envir = globalenv())
  } else {
    assign(".Random.seed", seed, globalenv())
  }
}

local_options <- function(.new = list(), ...,
                          .local_envir = parent.frame()) {
  .new <- utils::modifyList(as.list(.new), list(...))
  old <- do.call(options, .new)
  defer(do.call(options, old), .local_envir)
}

map_chr <- function(X, FUN, ...) {
  vapply(X, FUN, FUN.VALUE = character(1), ...)
}

#' Format a time stamp for HTTP
#'
#' @param t Date-time value to format, defaults to the current date and
#'   time. It must be a POSIXct object.
#' @return Character vector, formatted date-time.
#' @export

http_time_stamp <- function(t = Sys.time()) {
  t <- as.POSIXlt(t, tz = "UTC")
  strftime(t, "%a, %d %b %Y %H:%M:%S GMT")
}

capitalize <- function(x) {
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

web <- local_app_process(test_response_app())

test_that("response locals", {
  url <- web$url("/local")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(rawToChar(resp$content), "foo bar")
})

# on_response is tested via mw_log()

test_that("render", {
  # if the template or engine does not exist
  url <- web$url("/badengine")
  resp <- curl::curl_fetch_memory(url)
  expect_match(
    rawToChar(resp$content),
    "Cannot find template engine for view"
  )
})

test_that("send_json", {
  url <- web$url("/badjson")
  resp <- curl::curl_fetch_memory(url)
  expect_match(
    rawToChar(resp$content),
    "Specify only one of"
  )
})

test_that("send_file", {
  url <- web$url("/file")
  resp <- curl::curl_fetch_memory(url)
  path <- system.file(
    package = "webfakes",
    "examples", "static", "public", "foo", "bar.json"
  )
  expect_equal(resp$content, read_bin(path))
})

test_that("set_type", {
  url <- web$url("/type")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`content-type`, "application/json")
})

test_that("write", {
  url <- web$url("/write")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200)
  expect_equal(rawToChar(resp$content), "hello world!")

  # header can be set
  url <- web$url("/write-header")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200)
  expect_equal(rawToChar(resp$content), "hello world!")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`foo`, "bar")
})

test_that("write-wait", {
  url <- web$url("/write-wait")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200)
  expect_equal(rawToChar(resp$content), "hello world!")
})

test_that("send_chunk", {
  url <- web$url("/send-chunk")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200)
  expect_equal(
    rawToChar(resp$content),
    "first chunk\nsecond chunk\nthird and final chunk\n"
  )
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers[["content-type"]], "text/plain")
  expect_equal(headers[["transfer-encoding"]], "chunked")
})

test_that("add_header", {
  url <- web$url("/add-header")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(
    headers[names(headers) == "foo"],
    list(foo = "bar", foo = "bar2")
  )
  expect_equal(headers[["foobar"]], "baz")
})

test_that("http_time_stamp ignores local locale", {
  skip_on_cran()
  envir <- environment()
  tryCatch(
    withr::local_locale(LC_TIME = "es_ES.UTF-8", .local_envir = envir),
    warning = function(w) skip("es_ES.UTF-8 locale not available"),
    error = function(e) skip("es_ES.UTF-8 locale not available")
  )

  expect_equal(Sys.getlocale("LC_TIME"), "es_ES.UTF-8")

  # If httr::parse_http_date works it means time is not in local locale
  expect_false(is.na(httr::parse_http_date(http_time_stamp())))
})

# HTTP methods =========================================================

test_that("/get", {
  url <- httpbin$url("/get", query = c(q1 = "one", q2 = "two"))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "foo" = "bar")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/json")
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(data$path, "/get")
  expect_equal(data$headers$foo, "bar")
  expect_equal(data$args, list(q1 = "one", q2 = "two"))
})

test_that("/post", {
  url <- httpbin$url("/post", query = c(q1 = "one", q2 = "two"))
  data <- charToRaw(jsonlite::toJSON(list(foo = "bar", foobar = 1:3)))
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "content-type" = "application/json",
    foo = "bar"
  )
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/json")
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(data$path, "/post")
  expect_equal(data$headers$foo, "bar")
  expect_equal(data$args, list(q1 = "one", q2 = "two"))
  expect_equal(
    data$json,
    list(foo = "bar", foobar = 1:3)
  )
  expect_equal(
    jsonlite::fromJSON(data$data, simplifyVector = TRUE),
    data$json
  )
})

test_that("/post and multipart data", {
  on.exit(rm(tmp), add = TRUE)
  tmp <- tempfile()
  writeBin(charToRaw("foobar\n"), con = tmp)
  url <- httpbin$url("/post")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, customrequest = "POST")
  curl::handle_setform(
    handle, a = "1", b = "2",
    c = curl::form_file(tmp, type = "application/octet-stream")
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(resp$status_code, 200L)
  expect_equal(echo$form, list(a = "1", b = "2"))
  expect_equal(
    echo$files$c,
    list(
      filename = basename(tmp),
      value = paste0(
        "data:application/octet-stream;base64,",
        base64_encode("foobar\n")
      )
    )
  )
})

# Auth =================================================================

test_that("/basic-auth", {
  # no auth supplied
  url <- httpbin$url("/basic-auth/Aladdin/OpenSesame")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 401L)
  expect_equal(resp$type, "text/plain")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, "Basic realm=\"Fake Realm\"")

  # correct auth
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "Authorization"= "Basic QWxhZGRpbjpPcGVuU2VzYW1l"
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/json")
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp$content)),
    list(authenticated = TRUE, user = "Aladdin")
  )

  # wrong auth
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "Authorization"= "Basic NOLUCK"
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 401L)
  expect_equal(resp$type, "text/plain")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, "Basic realm=\"Fake Realm\"")
})

test_that("/hidden-basic-auth", {
  # no auth supplied
  url <- httpbin$url("/hidden-basic-auth/Aladdin/OpenSesame")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)
  expect_equal(resp$type, "text/plain")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, NULL)

  # correct auth
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "Authorization"= "Basic QWxhZGRpbjpPcGVuU2VzYW1l"
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/json")
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp$content)),
    list(authenticated = TRUE, user = "Aladdin")
  )

  # wrong auth
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "Authorization"= "Basic NOLUCK"
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 404L)
  expect_equal(resp$type, "text/plain")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, NULL)
})

test_that("/digest-auth", {
  # url <- "http://localhost:3000/digest-auth/auth/user/secret"
  url <- httpbin$url("/digest-auth/auth/user/secret")
  handle <- curl::new_handle()
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_true("www-authenticate" %in% names(headers))
  creds <- parse_authorization_header(headers$`www-authenticate`)
  expect_equal(creds$scheme, "digest")
  expect_equal(creds$realm, "webfakes.r-lib.org")
  expect_equal(creds$qop, "auth")
  expect_equal(creds$algorithm, "MD5")
  expect_equal(creds$stale, "FALSE")

  auth <- list(
    username = "user",
    realm = "webfakes.r-lib.org",
    nonce = creds$nonce,
    uri = "/digest-auth/auth/user/secret",
    qop = "auth",
    nc = "00000001",
    cnonce = "0a4f113b",
    opaque = creds$opaque
  )

  hash <- function(x) digest::digest(x, algo = "md5", serialize = FALSE)

  HA1 <- hash(paste(c(auth$username, auth$realm, "secret"), collapse = ":"))
  HA2 <- hash("GET:/digest-auth/auth/user/secret")
  auth$response <- hash(paste0(
    collapse = ":",
    c(HA1, auth$nonce, auth$nc, auth$cnonce, auth$qop, HA2)
  ))

  authorize <- paste0(
    "Digest ",
    paste0(names(auth), "=", auth, collapse = ", ")
  )

  curl::handle_setheaders(handle, authorization = authorize)
  resp2 <- curl::curl_fetch_memory(handle = handle, url)
  expect_equal(resp2$status_code, 200L)
  cnt <- jsonlite::fromJSON(rawToChar(resp2$content), simplifyVector = TRUE)
  expect_equal(cnt, list(authentication = TRUE, user = "user"))
})

test_that("/bearer", {
  # no auth
  url <- httpbin$url("/bearer")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 401L)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, "bearer")

  # bad auth format
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, authorization = "foobar")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 401L)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$`www-authenticate`, "bearer")

  # correct format
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, authorization = "Bearer secret")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo, list(authenticated = TRUE, token = "secret"))
})

# Status codes =========================================================

test_that("/status", {
  codes <- c(200, 301, 401, 404)

  # get
  for (code in codes) {
    url <- httpbin$url(paste0("/status/", code))
    resp <- curl::curl_fetch_memory(url)
    expect_equal(resp$status_code, code)
  }

  # post
  data <- charToRaw(jsonlite::toJSON(list(foo = "bar", foobar = 1:3)))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "content-type" = "application/json")
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )
  for (code in codes) {
    url <- httpbin$url(paste0("/status/", code))
    resp <- curl::curl_fetch_memory(url, handle = handle)
    expect_equal(resp$status_code, code)
  }
})

# Request inspection ===================================================

test_that("/headers", {
  url <- httpbin$url("/headers")
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "header1" = "this", "header2" = "that")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo$headers$header1, "this")
  expect_equal(echo$headers$header2, "that")
})

test_that("/ip", {
  url <- httpbin$url("/ip")
  resp <- curl::curl_fetch_memory(url)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo, list(origin = "127.0.0.1"))
})

test_that("/user-agent", {
  url <- httpbin$url("/user-agent")
  ua <- "i am libcurl, that is my name"
  withr::local_options(list(HTTPUserAgent = ua))
  resp <- curl::curl_fetch_memory(url)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo, list("user-agent" = ua))
})

# Response inspection ==================================================

test_that("/etag", {
  url <- httpbin$url("/etag/foobar")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200)
  expect_equal(headers$etag, "foobar")

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "If-None-Match" = "\"foobar\"")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 304)
  expect_true(length(resp$content) == 0)

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "If-None-Match" = "\"not-foobar\"")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200)
  expect_true(length(resp$content) > 0)

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "If-Match" = "\"foobar\"")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200)
  expect_true(length(resp$content) > 0)

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "If-Match" = "\"not-foobar\"")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 412)
})

test_that("/response-headers", {
  url <- httpbin$url("/response-headers")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(rawToChar(resp$content), "{}")

  url2 <- httpbin$url("/response-headers", c(foo = "bar"))
  resp2 <- curl::curl_fetch_memory(url2)
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp2$content)),
    list(foo = "bar")
  )
  headers <- curl::parse_headers_list(resp2$headers)
  expect_equal(headers[["foo"]], "bar")

  url3 <- httpbin$url("/response-headers", c(foo = "bar", foobar = "baz"))
  resp3 <- curl::curl_fetch_memory(url3)
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp3$content)),
    list(foo = "bar", foobar = "baz")
  )
  headers <- curl::parse_headers_list(resp3$headers)
  expect_equal(headers[["foo"]], "bar")
  expect_equal(headers[["foobar"]], "baz")

  handle <- curl::new_handle()
  data <- charToRaw("{}")
  curl::handle_setheaders(
    handle,
    "content-type" = "application/json"
  )
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )
  resp31 <- curl::curl_fetch_memory(url3, handle = handle)
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp31$content)),
    list(foo = "bar", foobar = "baz")
  )
  headers <- curl::parse_headers_list(resp3$headers)
  expect_equal(headers[["foo"]], "bar")
  expect_equal(headers[["foobar"]], "baz")

  url4 <- httpbin$url("/response-headers", c(foo = "bar", foo = "bar2"))
  resp4 <- curl::curl_fetch_memory(url4)
  expect_equal(
    jsonlite::fromJSON(rawToChar(resp4$content)),
    list(foo = c("bar", "bar2"))
  )
  headers <- curl::parse_headers_list(resp4$headers)
  expect_equal(
    headers[names(headers) == "foo"],
    list(foo = "bar", foo = "bar2")
  )
})

test_that("/cache", {
  url <- httpbin$url("/cache")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_true("last-modified" %in% tolower(names(headers)))

  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "If-Modified-Since" =
      http_time_stamp(Sys.time() - as.difftime(5, units = "mins"))
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 304L)


  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "If-None-Match" = "some-etag"
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 304L)
})

test_that("/cache/:value", {
  url <- httpbin$url("/cache/10")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(
    headers[["cache-control"]],
    "public, max-age=10"
  )
})

# Response formats =====================================================

test_that("/deny", {
  url <- httpbin$url("/deny")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  path <- system.file(
    package = "webfakes",
    "examples", "httpbin", "data", "deny.txt"
  )
  expect_equal(resp$content, read_bin(path))
})

test_that("/gzip", {
  # curl seems to ungzip automatically, so we use url()
  url <- httpbin$url("/gzip")
  con <- url(url, open = "rb")
  on.exit(close(con), add = TRUE)
  echo <- readBin(con, "raw", 10000)
  expect_equal(echo[1:2], charToRaw("\x1f\x8b"))
  json <- readChar(gzcon(rawConnection(echo)), 10000)
  obj <- jsonlite::fromJSON(json, simplifyVector = FALSE)
  expect_equal(obj$path, "/gzip")
})

test_that("/deflate", {
  url <- httpbin$url("/deflate")
  con <- url(url, open = "rb")
  on.exit(close(con), add = TRUE)
  echo <- readBin(con, "raw", 10000)
  data <- jsonlite::fromJSON(rawToChar(zip::inflate(echo)$output))
  expect_true(data$deflated)
})

test_that("/brotli", {
  url <- httpbin$url("/brotli")
  con <- url(url, open = "rb")
  on.exit(close(con), add = TRUE)
  echo <- readBin(con, "raw", 10000)
  data <- jsonlite::fromJSON(rawToChar(brotli::brotli_decompress(echo)))
  expect_true(data$brotli)
})

test_that("/encoding/utf8", {
  url <- httpbin$url("/encoding/utf8")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "text/html; charset=utf-8")
  ptrn <- as.raw(c(
    0xe1, 0x8c, 0x8c, 0xe1, 0x8c, 0xa5, 0x20, 0xe1, 0x8b,
    0xab, 0xe1, 0x88, 0x88, 0xe1, 0x89, 0xa4, 0xe1, 0x89, 0xb1
  ))
  # On windows end of line is converted to \r\n
  expect_true(grepRaw(ptrn, resp$content, fixed = TRUE) %in% c(9085, 9233))
})

test_that("/html", {
  url <- httpbin$url("/html")
  resp <- curl::curl_fetch_memory(url)
  expect_match(resp$type, "^text/html")
  expect_match(rawToChar(resp$content), "<html>", fixed = TRUE)
})

test_that("/json", {
  url <- httpbin$url("/json")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "application/json")
  path <- system.file(
    package = "webfakes",
    "examples", "httpbin", "data", "example.json"
  )
  expect_equal(resp$content, read_bin(path))
})

test_that("/robots.txt", {
  url <- httpbin$url("/robots.txt")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "text/plain")
  path <- system.file(
    package = "webfakes",
    "examples", "httpbin", "data", "robots.txt"
  )
  expect_equal(resp$content, read_bin(path))
})

test_that("/xml", {
  url <- httpbin$url("/xml")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "application/xml")
  path <- system.file(
    package = "webfakes",
    "examples", "httpbin", "data", "example.xml"
  )
  expect_equal(resp$content, read_bin(path))
})

# Dynamic data =========================================================

test_that("/base64/", {
  url <- httpbin$url("/base64")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  expect_equal(rawToChar(resp$content), "Everything is Rsome")

  value <- base64_encode("hello there!")
  url <- httpbin$url(paste0("/base64/", value))
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "application/octet-stream")
  expect_equal(rawToChar(resp$content), "hello there!")
})

test_that("/bytes", {
  url <- httpbin$url("/bytes/foo")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)

  url <- httpbin$url("/bytes/1000")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$type, "application/octet-stream")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers[["content-length"]], "1000")
  expect_equal(length(resp$content), 1000)
})

test_that("/delay", {
  url <- httpbin$url("/delay/foo")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)

  url <- httpbin$url("/delay/0.2")
  st <- system.time(resp <- curl::curl_fetch_memory(url))
  expect_true(st[["elapsed"]] >= 0.1)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/json")
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo$path, "/delay/0.2")
})

test_that("/stream-bytes", {
  url <- httpbin$url("/stream-bytes/100")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200)
  expect_equal(headers[["transfer-encoding"]], "chunked")
  expect_equal(length(resp$content), 100)

  ## seed works
  url2 <- httpbin$url("/stream-bytes/10", c(seed = 100))
  resp2 <- curl::curl_fetch_memory(url2)
  expect_false(identical(resp$content, resp2$content))

  resp3 <- curl::curl_fetch_memory(url2)
  expect_identical(resp2$content, resp3$content)

  ## chunk_size works
  url <- httpbin$url("/stream-bytes/100", c(chunk_size = 30))
  resp <- curl::curl_fetch_memory(url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200)
  expect_equal(headers[["transfer-encoding"]], "chunked")
  expect_equal(length(resp$content), 100)
})

test_that("/range", {
  url <- httpbin$url("/range/100")

  # HEAD
  handle <- curl::new_handle()
  curl::handle_setopt(handle, customrequest = "HEAD")
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200L)
  expect_equal(headers$`accept-ranges`, "bytes")
  expect_equal(headers$`etag`, "range100")

  abc <- function(n) {
    l <- paste(letters, collapse = "")
    substr(strrep(l, n / nchar(l) + 1), 1, n)
  }

  # No range header
  handle <- curl::new_handle()
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200L)
  expect_equal(headers$`accept-ranges`, "bytes")
  expect_equal(headers$`etag`, "range100")
  expect_equal(resp$content, charToRaw(abc(100)))

  # Illegal range headers
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, Range = "bytes=20-30, 25-35")
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200L)
  expect_equal(headers$`accept-ranges`, "bytes")
  expect_equal(headers$`etag`, "range100")
  expect_equal(resp$content, charToRaw(abc(100)))

  # Legal header
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, Range = "bytes=20-29")
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 206L)
  expect_equal(headers$`accept-ranges`, "bytes")
  expect_equal(headers$`etag`, "range100")
  expect_equal(resp$content, charToRaw(substr(abc(30), 21, 30)))

  # In pieces, duration is for the full response
  url <- httpbin$url("/range/100?chunk_size=1&duration=10")
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, Range = "bytes=20-29")
  resp <- curl::curl_fetch_memory(handle = handle, url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 206L)
  expect_equal(headers$`accept-ranges`, "bytes")
  expect_equal(headers$`etag`, "range100")
  expect_equal(resp$content, charToRaw(substr(abc(30), 21, 30)))
  expect_true(resp$times[["total"]] > 0.5)
})

test_that("/uuid", {
  url <- httpbin$url("/uuid")
  resp <- curl::curl_fetch_memory(url)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_match(echo$uuid, "^[-0-9a-z]+$")
  expect_equal(nchar(echo$uuid), 36)
})

test_that("/stream", {
  url <- httpbin$url("/stream/10")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl:::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 200)
  expect_equal(headers[["transfer-encoding"]], "chunked")
  lines <- strsplit(rawToChar(resp$content), "\n", fixed = TRUE)[[1]]
  json <- lapply(lines, jsonlite::fromJSON, simplifyVector = TRUE)
  expect_equal(vapply(json, "[[", integer(1), "id"), 0:9)
})

# Cookies ==============================================================

test_that("/cookies", {
  url <- httpbin$url("/cookies")
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, Cookie = "foo=bar; bar=baz")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(data, list(cookies = list(foo = "bar", bar = "baz")))
})

test_that("/cookies/set/:name/:value", {
  url <- httpbin$url("/cookies/set/foo/bar")
  handle <- curl::new_handle()
  resp <- curl::curl_fetch_memory(url, handle = handle)
  headers <- curl::parse_headers(resp$headers, multiple = TRUE)
  expect_true("HTTP/1.1 302 Found" %in% headers[[1]])
  expect_true("Set-Cookie: foo=bar; Path=/" %in% headers[[1]])
  expect_true("HTTP/1.1 200 OK" %in% headers[[2]])
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(data, list(cookies = list(foo = "bar")))
  expect_snapshot(curl::handle_cookies(handle), variant = r_variant())
})

test_that("/cookies/set", {
  url <- httpbin$url("/cookies/set?foo=bar&bar=baz")
  handle <- curl::new_handle()
  resp <- curl::curl_fetch_memory(url, handle = handle)
  headers <- curl::parse_headers(resp$headers, multiple = TRUE)
  expect_true("HTTP/1.1 302 Found" %in% headers[[1]])
  expect_true("Set-Cookie: foo=bar; Path=/" %in% headers[[1]])
  expect_true("Set-Cookie: bar=baz; Path=/" %in% headers[[1]])
  expect_true("HTTP/1.1 200 OK" %in% headers[[2]])
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(data, list(cookies = list(bar="baz", foo = "bar")))
  expect_snapshot(curl::handle_cookies(handle), variant = r_variant())
})

test_that("/cookies/delete", {
  handle <- curl::new_handle()

  url1 <- httpbin$url("/cookies/set?foo=bar&bar=baz")
  resp1 <- curl::curl_fetch_memory(url1, handle = handle)

  url <- httpbin$url("/cookies/delete?foo")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  headers <- curl::parse_headers(resp$headers, multiple = TRUE)
  expect_true("HTTP/1.1 302 Found" %in% headers[[1]])
  expect_true("Set-Cookie: foo=; Expires=Thu, 01 Jan 1970 00:00:00 GMT; Max-Age=0; Path=/" %in% headers[[1]])
  expect_true("HTTP/1.1 200 OK" %in% headers[[2]])
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(data, list(cookies = list(bar = "baz")))
  expect_snapshot(curl::handle_cookies(handle), variant = r_variant())
})

# Images ===============================================================

test_that("/image", {
  url <- httpbin$url("/image")
  types <- c("image/webp", "image/svg+xml", "image/jpeg", "image/png")
  for (type in types) {
    handle <- curl::new_handle()
    curl::handle_setheaders(handle, accept = type)
    resp <- curl::curl_fetch_memory(url, handle = handle)
    expect_equal(resp$status_code, 200L)
    expect_equal(resp$type, type)
  }

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, accept = "image/*")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "image/png")

  handle <- curl::new_handle()
  curl::handle_setheaders(handle, accept = "application/json")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 406L)

  exts <- c("jpeg", "png", "svg", "webp")
  for (ext in exts) {
    handle <- curl::new_handle()
    url <- httpbin$url(paste0("/image/", ext))
    resp <- curl::curl_fetch_memory(url, handle = handle)
    expect_equal(resp$status_code, 200L)
    expect_equal(resp$type, unname(mime_find(ext)))
  }
})

# Redirects ============================================================

test_that("absolute-redirect", {
  url <- httpbin$url("/absolute-redirect/4")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = FALSE)
  resp <- curl::curl_fetch_memory(url, handle = handle)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 302L)
  expect_equal(headers$location, httpbin$url("/absolute-redirect/3"))

  url <- httpbin$url("/absolute-redirect/2")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = TRUE)
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$url, httpbin$url("/get"))

  url <- httpbin$url("/absolute-redirect/foo")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)
})

test_that("redirect", {
  url <- httpbin$url("/redirect/4")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = FALSE)
  resp <- curl::curl_fetch_memory(url, handle = handle)
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(resp$status_code, 302L)
  expect_equal(headers$location, "/redirect/3")

  url <- httpbin$url("/redirect/2")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = TRUE)
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$url, httpbin$url("/get"))

  url <- httpbin$url("/redirect/foo")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)
})

test_that("/redirect-to", {
  # default status is 302
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = FALSE)
  url <- httpbin$url("/redirect-to", query = list(url = "/get"))
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 302L)
  expect_equal(resp$type, "text/plain")
  expect_equal(curl::parse_headers_list(resp$headers)$location, "/get")
  expect_equal(rawToChar(resp$content), "302 Found. Redirecting to /get")

  # can specify status
  handle <- curl::new_handle()
  curl::handle_setopt(handle, followlocation = FALSE)
  url <- httpbin$url(
    "/redirect-to",
    query = list(url = "/status/200", status_code = 301)
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 301L)
  expect_equal(resp$type, "text/plain")
  expect_equal(curl::parse_headers_list(resp$headers)$location, "/status/200")
  expect_equal(
    rawToChar(resp$content),
    "301 Moved Permanently. Redirecting to /status/200"
  )
})

# Anything =============================================================

http_statuses <- c(
  "100" = "Continue",
  "101" = "Switching Protocols",
  "102" = "Processing (WebDAV; RFC 2518)",
  "200" = "OK",
  "201" = "Created",
  "202" = "Accepted",
  "203" = "Non-Authoritative Information",
  "204" = "No Content",
  "205" = "Reset Content",
  "206" = "Partial Content",
  "207" = "Multi-Status (WebDAV; RFC 4918)",
  "208" = "Already Reported (WebDAV; RFC 5842)",
  "226" = "IM Used (RFC 3229)",
  "300" = "Multiple Choices",
  "301" = "Moved Permanently",
  "302" = "Found",
  "303" = "See Other",
  "304" = "Not Modified",
  "305" = "Use Proxy",
  "306" = "Switch Proxy",
  "307" = "Temporary Redirect",
  "308" = "Permanent Redirect (experimental Internet-Draft)",
  "400" = "Bad Request",
  "401" = "Unauthorized",
  "402" = "Payment Required",
  "403" = "Forbidden",
  "404" = "Not Found",
  "405" = "Method Not Allowed",
  "406" = "Not Acceptable",
  "407" = "Proxy Authentication Required",
  "408" = "Request Timeout",
  "409" = "Conflict",
  "410" = "Gone",
  "411" = "Length Required",
  "412" = "Precondition Failed",
  "413" = "Request Entity Too Large",
  "414" = "Request-URI Too Long",
  "415" = "Unsupported Media Type",
  "416" = "Requested Range Not Satisfiable",
  "417" = "Expectation Failed",
  "418" = "I'm a teapot (RFC 2324)",
  "420" = "Enhance Your Calm (Twitter)",
  "422" = "Unprocessable Entity (WebDAV; RFC 4918)",
  "423" = "Locked (WebDAV; RFC 4918)",
  "424" = "Failed Dependency (WebDAV; RFC 4918)",
  "424" = "Method Failure (WebDAV)",
  "425" = "Unordered Collection (Internet draft)",
  "426" = "Upgrade Required (RFC 2817)",
  "428" = "Precondition Required (RFC 6585)",
  "429" = "Too Many Requests (RFC 6585)",
  "431" = "Request Header Fields Too Large (RFC 6585)",
  "444" = "No Response (Nginx)",
  "449" = "Retry With (Microsoft)",
  "450" = "Blocked by Windows Parental Controls (Microsoft)",
  "451" = "Unavailable For Legal Reasons (Internet draft)",
  "499" = "Client Closed Request (Nginx)",
  "500" = "Internal Server Error",
  "501" = "Not Implemented",
  "502" = "Bad Gateway",
  "503" = "Service Unavailable",
  "504" = "Gateway Timeout",
  "505" = "HTTP Version Not Supported",
  "506" = "Variant Also Negotiates (RFC 2295)",
  "507" = "Insufficient Storage (WebDAV; RFC 4918)",
  "508" = "Loop Detected (WebDAV; RFC 5842)",
  "509" = "Bandwidth Limit Exceeded (Apache bw/limited extension)",
  "510" = "Not Extended (RFC 2774)",
  "511" = "Network Authentication Required (RFC 6585)",
  "598" = "Network read timeout error (Unknown)",
  "599" = "Network connect timeout error (Unknown)"
)

#' Log requests to the standard output or other connection
#'
#' A one line log entry for every request. The output looks like this:
#' ```
#' GET http://127.0.0.1:3000/image 200 3 ms - 4742
#' ```
#' and contains
#' * the HTTP method,
#' * the full request URL,
#' * the HTTP status code of the response,
#' * how long it took to process the response, in ms,
#' * and the size of the response body, in bytes.
#'
#' @param format Log format. Not implemented currently.
#' @param stream R connection to log to. `"stdout"` means the standard
#'   output, `"stderr"` is the standard error. You can also supply a
#'   connection object, but then you need to be sure that it will be
#'   valid when the app is actually running.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_log())
#' app

mw_log <- function(format = "dev", stream = "stdout") {

  format; stream

  function(req, res) {

    start <- Sys.time()

    fmt <- function(req, res) {
      if (identical(stream, "stdout")) stream <- stdout()
      if (identical(stream, "stderr")) stream <- stderr()
      len <- if (is.null(res$.body)) {
        0L                              # nocov
      } else if (is.raw(res$.body)) {
        length(res$.body)
      } else if (is_string(res$.body)) {
        nchar(res$.body, type = "bytes")
      } else {
        "??"                            # nocov
      }
      t <- as.integer(round((Sys.time() - start) * 1000))
      msg <- sprintf(
        "%s %s %s %s ms - %s\n",
        toupper(req$method), req$url, res$.status, t, len
      )
      cat0(msg, file = stream)
      if (inherits(stream, "connection")) flush(stream)
    }
    res$on_response(fmt)

    "next"
  }
}

parse_ports <- function(opts = server_opts()) {
  paste0(
    if (!is.null(opts$interfaces)) paste0(opts$interfaces, ":"),
    opts$port %||% "0",
    collapse = ","
  )
}

server_start <- function(opts = server_opts()) {
  ports <- parse_ports(opts)

  if (!is.na(opts$access_log_file)) {
    mkdirp(dirname(opts$access_log_file))
    file.create(opts$access_log_file)
  }
  if (!is.na(opts$error_log_file)) {
    mkdirp(dirname(opts$error_log_file))
    file.create(opts$error_log_file)
  }

  throttle <- paste0(
    "*=",
    if (opts$throttle == Inf) "0" else opts$throttle
  )

  options <- c(
    "listening_ports" = ports,
    "num_threads" = opts$num_threads,
    "enable_keep_alive" = c("no", "yes")[[opts$enable_keep_alive + 1]],
    "access_log_file" = opts$access_log_file %|NA|% "",
    "error_log_file" = opts$error_log_file %|NA|% "",
    "tcp_nodelay" = c("0", "1")[[opts$tcp_nodelay + 1]],
    "throttle" = throttle,

    # These are not configurable currently
    "request_timeout_ms" = "100000",
    "enable_auth_domain_check" = "no",
    "decode_url" = if (opts$decode_url) "yes" else "no",
    "ssl_certificate" = if (!is.null(opts$ssl_certificate)) opts$ssl_certificate
  )

  srv <- call_with_cleanup(c_server_start, options)
  attr(srv, "options") <- opts
  srv
}

#' Webfakes web server options
#'
#' @param remote Meta-option. If set to `TRUE`, webfakes uses slightly
#'   different defaults, that are more appropriate for a background
#'   server process.
#' @param port Port to start the web server on. Defaults to a randomly
#'   chosen port.
#' @param num_threads Number of request handler threads to use. Typically
#'   you don't need more than one thread, unless you run test cases in
#'   parallel or you make concurrent HTTP requests.
#' @param interfaces The network interfaces to listen on. Being a test
#'   web server, it defaults to the localhost. Only bind to a public
#'   interface if you know what you are doing. webfakes was not designed
#'   to serve public web pages.
#' @param enable_keep_alive Whether the server keeps connections alive.
#' @param access_log_file `TRUE`, `FALSE`, or a path. See 'Logging'
#'   below.
#' @param error_log_file `TRUE`, `FALSE`, or a path. See 'Logging'
#'   below.
#' @param tcp_nodelay if `TRUE` then packages will be sent as soon as
#'   possible, instead of waiting for a full buffer or timeout to occur.
#' @param throttle Limit download speed for clients. If not `Inf`,
#'   then it is the maximum number of bytes per second, that is sent to
#'   as connection.
#' @param decode_url Whether the server should automatically decode
#'   URL-encodded URLs. If `TRUE` (the default), `/foo%2fbar` will be
#'   converted to `/foo/bar` automatically. If `FALSE`, URLs as not
#'   URL-decoded.
#' @param ssl_certificate Path to the SSL certificate of the server,
#'   needed if you want to server HTTPS requests.
#' @return List of options that can be passed to `webfakes_app$listen()`
#'   (see [new_app()]), and [new_app_process()].
#'
#' @section Logging:
#'
#' * For `access_log_file`, `TRUE` means `<log-dir>/access.log`.
#' * For `error_log_file`, `TRUE` means `<log-dir>/error.log`.
#'
#' `<log-dir>` is set to the contents of the `WEBFAKES_LOG_DIR`
#' environment variable, if it is set. Otherwise it is set to
#' `<tmpdir>/webfakes` for local apps and `<tmpdir>/<pid>/webfakes` for
#' remote apps (started with `new_app_procss()`).
#'
#' `<tmpdir>` is the session temporary directory of the _main process_.
#'
#' `<pid>` is the process id of the subprocess.
#'
#' @export
#' @examples
#' # See the defaults
#' server_opts()

server_opts <- function(
  remote = FALSE,
  port = NULL,
  num_threads = 1,
  interfaces = "127.0.0.1",
  enable_keep_alive = FALSE,
  access_log_file = remote,
  error_log_file = TRUE,
  tcp_nodelay = FALSE,
  throttle = Inf,
  decode_url = TRUE,
  ssl_certificate = NULL
) {
  log_dir <- Sys.getenv("WEBFAKES_LOG_DIR", file.path(tempdir(), "webfakes"))
  if (isTRUE(access_log_file)) {
    if (remote) {
      access_log_file <- file.path(log_dir, "%p", "access.log")
    } else {
      access_log_file <- file.path(log_dir, "access.log")
    }
  } else if (isFALSE(access_log_file)) {
    access_log_file <- NA_character_
  }
  if (isTRUE(error_log_file)) {
    if (remote) {
      error_log_file <- file.path(log_dir, "%p", "error.log")
    } else {
      error_log_file <- file.path(log_dir, "error.log")
    }
  } else if (isFALSE(error_log_file)) {
    error_log_file <- NA_character_
  }
  rm(log_dir)

  ssl_certificate <- ssl_certificate %||%
    system.file("cert/localhost/server.pem", package = "webfakes")

  as.list(environment())
}

server_get_ports <- function(srv) {
  as.data.frame(call_with_cleanup(c_server_get_ports, srv))
}

server_stop <- function(srv) {
  invisible(call_with_cleanup(c_server_stop, srv))
}

server_poll <- function(srv, cleanup = TRUE) {
  while (TRUE) {
    tryCatch(
      return(call_with_cleanup(c_server_poll, srv, cleanup)),
      error = function(err) {
        cat(as.character(err), file = stderr())
        stop(new_webfakes_error())
      }
    )
  }
}

response_send <- function(req) {
  tryCatch(
    call_with_cleanup(c_response_send, req),
    error = function(err) {
      cat(as.character(err), file = stderr())
      stop(new_webfakes_error())
    }
  )
  invisible(NULL)
}

response_send_chunk <- function(req, data) {
  tryCatch(
    call_with_cleanup(c_response_send_chunk, req, data),
    error = function(err) {
      cat(as.character(err), file = stderr())
      stop(new_webfakes_error())
    }
  )
  invisible(NULL)
}

response_send_error <- function(req, msg, status) {
  tryCatch(
    call_with_cleanup(c_response_send_error, req, msg, status),
    error = function(err) {
      cat(as.character(err), file = stderr())
      stop(new_webfakes_error())
    }
  )
  invisible(NULL)
}

response_delay <- function(req, secs) {
  tryCatch(
    call_with_cleanup(c_response_delay, req, secs),
    error = function(err) {
      cat(as.character(err), file = stderr())
      stop(new_webfakes_error())
    }
  )
  invisible(NULL)
}

response_write <- function(req, data) {
  tryCatch(
    call_with_cleanup(c_response_write, req, data),
    error = function(err) {
      cat(as.character(err), file = stderr())
      stop(new_webfakes_error())
    }
  )
  invisible(NULL)
}

new_webfakes_error <- function(...) {
  structure(
    list(message = "error in webfakes connection", ...),
    class = c("webfakes_error", "error", "condition")
  )
}

#' Parse a multipart HTTP request body
#'
#' Adds the parsed form fields in the `form` element of the request and
#' the parsed files to the `files` element.
#'
#' @param type Content type to match before parsing. If it does not
#'   match, then the request object is not modified.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_multipart())
#' app

mw_multipart <- function(type = "multipart/form-data") {
  type
  function(req, res) {
    ct <- req$get_header("Content-Type") %||% ""
    if (!any(vapply(
           paste0("^", type),
           function(x) grepl(x, ct),
           logical(1)))) return("next")

    parts <- str_trim(strsplit(ct, ";", fixed = TRUE)[[1]])
    bnd <- grep("boundary=", parts, value = TRUE)[1]
    if (is.na(bnd)) return("next")
    bnd <- sub("^boundary=", "", bnd)

    tryCatch({
      mp <- parse_multipart(req$.body, bnd)
      req$form <- list()
      req$files <- list()
      for (p in mp) {
        if (is.null(p$filename)) {
          req$form[[p$name]] <- rawToChar(p$value)
        } else {
          req$files[[p$name]] <- list(filename = p$filename, value = p$value)
        }
      }

    }, error = function(err) NULL)

    "next"
  }
}

parse_multipart <- function(body, boundary) {
  boundary <- paste0("--", boundary)
  boundary_length <- nchar(boundary)

  # Find the locations of the boundary string
  indexes <- grepRaw(boundary, body, fixed = TRUE, all = TRUE)

  if (!length(indexes)) stop("Boundary was not found in the body.")

  if (length(indexes) == 1) {
    if (length(body) < (boundary_length + 5)) {
      # Empty HTML5 FormData object
      return(list())
    } else {
      # Something went wrong
      stop("The 'boundary' was only found once in the ",
           "multipart/form-data message. It should appear at ",
           "least twice. The request-body might be truncated.")
    }
  }

  parts <- list()
  for (i in seq_along(utils::head(indexes, -1))) {
    from <- indexes[i] + boundary_length
    to <- indexes[i + 1] -1
    parts[[i]] <- body[from:to]
  }

  out <- lapply(parts, multipart_sub)
  names(out) <- vapply(
    out,
    function(x) as.character(x$name),
    character(1)
  )

  out
}

multipart_sub <- function(bodydata) {
  splitchar <- grepRaw("\\r\\n\\r\\n|\\n\\n|\\r\\r", bodydata)
  if (!length(splitchar)) {
    stop("Invalid multipart subpart:\n\n", rawToChar(bodydata))
  }

  headers <- bodydata[1:(splitchar-1)]
  headers <- str_trim(rawToChar(headers))
  headers <- gsub("\r\n", "\n", headers)
  headers <- gsub("\r", "\n", headers)
  headerlist <- unlist(lapply(strsplit(headers, "\n")[[1]], str_trim))

  dispindex <- grep("^Content-Disposition:", headerlist)
  if (!length(dispindex)) {
    stop("Content-Disposition header not found:", headers)
  }
  dispheader <- headerlist[dispindex]

  #get parameter name
  m <- regexpr("; name=\\\"(.*?)\\\"", dispheader)
  if (m < 0)
    stop('failed to find the name="..." header')

  namefield <- unquote(sub(
    "; name=",
    "",
    regmatches(dispheader, m),
    fixed = TRUE
  ))

  #test for file upload
  m <- regexpr("; filename=\\\"(.*?)\\\"", dispheader)
  if (m < 0) {
    filenamefield = NULL
  } else {
    filenamefield <- unquote(sub(
      "; filename=",
      "",
      regmatches(dispheader, m),
      fixed = TRUE
    ))
  }

  #filedata
  splitval <- grepRaw("\\r\\n\\r\\n|\\n\\n|\\r\\r", bodydata, value=TRUE)
  start <- splitchar + length(splitval)
  if (identical(utils::tail(bodydata, 2), charToRaw("\r\n"))) {
    end <- length(bodydata) - 2
  } else {
    end <- length(bodydata) - 1
  }

  #the actual fields
  list (
    name = namefield,
    value = bodydata[start:end],
    filename = filenamefield
  )
}

#' Middleware to parse a plain text body
#'
#' Adds the parsed object as the `text` element of the request object.
#'
#' @param default_charset Encoding to set on the text.
#' @param type Content type to match before parsing. If it does not
#'   match, then the request object is not modified.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_text())
#' app

mw_text <- function(default_charset = "utf-8",
                     type = "text/plain") {
  default_charset; type
  function(req, res) {
    ct <- req$get_header("Content-Type") %||% ""
    if (! ct %in% tolower(type)) return("next")
    req$text <- rawToChar(req$.body %||% raw())
    Encoding(req$text) <- default_charset
    "next"
  }
}

#' Middleware that calls a CGI script
#'
#' You can use it as an unconditional middleware in `app$use()`,
#' as a handler on `app$get()`, `app$post()`, etc., or you can call it
#' from a handler. See examples below.
#'
#' @param command External command to run.
#' @param args Arguments to pass to the external command.
#' @param timeout Timeout for the external command. If the command does
#'   not terminate in time, the web server kills it and returns an 500
#'   response.
#' @return A function with signature
#' ```
#' function(req, res, env = character())
#' ```
#'
#' See [RFC 3875](https://datatracker.ietf.org/doc/html/rfc3875)
#' for details on the CGI protocol.
#'
#' The request body (if any) is passed to the external command as standard
#' intput. `mw_cgi()` sets `CONTENT_LENGTH`, `CONTENT_TYPE`,
#' `GATEWAY_INTERFACE`, `PATH_INFO`, `QUERY_STRING`, `REMOTE_ADDR`,
#' `REMOTE_HOST`, `REMOTE_USER`, `REQUEST_METHOD`, `SERVER_NAME`,
#' `SERVER_PORT`, `SERVER_PROTOCOL`, `SERVER_SOFTEWARE`.
#'
#' It does not currently set the `AUTH_TYPE`, `PATH_TRANSLATED`,
#' `REMOTE_IDENT`, `SCRIPT_NAME` environment variables.
#'
#' The standard output of the external command is used to set the
#' response status code, the response headers and the response body.
#' Example output from git's CGI:
#' ```
#' Status: 200 OK
#' Expires: Fri, 01 Jan 1980 00:00:00 GMT
#' Pragma: no-cache
#' Cache-Control: no-cache, max-age=0, must-revalidate
#' Content-Type: application/x-git-upload-pack-advertisement
#'
#' 000eversion 2
#' 0015agent=git/2.42.0
#' 0013ls-refs=unborn
#' 0020fetch=shallow wait-for-done
#' 0012server-option
#' 0017object-format=sha1
#' 0010object-info
#' 0000
#' ```
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_cgi("echo", "Status: 200\n\nHello"))
#' app
#'
#' app2 <- new_app()
#' app2$get("/greet", mw_cgi("echo", "Status: 200\n\nHello"))
#' app2
#'
#' # Using `mw_cgi()` in a handler, you can pass extra environment variables
#' app3 <- new_app()
#' cgi <- mw_cgi("echo", "Status: 200\n\nHello")
#' app2$get("/greet", function(req, res) {
#'   cgi(req, res, env = c("EXTRA_VAR" = "EXTRA_VALUE"))
#' })
#' app3

mw_cgi <- function(command, args = character(),
                   timeout = as.difftime(Inf, units = "secs")) {
  command
  args
  timeout <- if (timeout == Inf) {
    -1
  } else {
    as.double(timeout, units = "secs") * 1000
  }

  function(req, res, env = character()) {
    all_env <- c("current", cgi_env(req), env)

    inp <- tempfile()
    out <- tempfile()
    err <- tempfile()
    on.exit(unlink(c(inp, out, err)), add = TRUE)
    writeBin(req$.body %||% raw(), inp)

    px <- processx::process$new(
      command,
      args,
      env = all_env,
      stdin = inp,
      stdout = out,
      stderr = err
    )
    px$wait(timeout)
    output <- parse_cgi_output(px, out, err)

    res$set_status(output$status)
    for (idx in seq_along(output$headers)) {
      res$set_header(names(output$headers)[idx], output$headers[[idx]])
    }
    res$send(output$body)
  }
}

parse_cgi_output <- function(px, out, err) {
  if (px$is_alive() || px$get_exit_status() != 0) {
    px$kill()
    err <- tryCatch(read_char(err), error = function(e) "???")
    return(list(
      status = 500L,
      headers = c("content-type" = "text/plain"),
      body = paste0("Internal git error: ", err)
    ))
  }

  out <- read_bin(out)
  err <- read_char(err)

  cgi_res <- split_cgi_output(out)
  headers <- cgi_res$headers
  names(headers) <- tolower(names(headers))

  status <- parse_status(headers[["status"]] %||% "200")
  headers <- headers[names(headers) != "status"]

  list(status = status, headers = headers, body = cgi_res$body)
}

cgi_env <- function(req) {
  url <- parse_url(req$url)
  c(
    CONTENT_LENGTH = length(req$.body),
    CONTENT_TYPE = if (!is.null(req$get_header)) req$get_header("content-type") %||% "",
    GATEWAY_INTERFACE = "CGI/1.1",
    PATH_INFO = req$path,
    QUERY_STRING = req$query_string,
    REMOTE_ADDR = req$remote_addr,
    REMOTE_HOST = req$remote_addr,
    REMOTE_USER = "anonymous",
    REQUEST_METHOD = toupper(req$method),
    SERVER_NAME = url$host,
    SERVER_PORT = url$port,
    SERVER_PROTOCOL = paste0("http/", req$http_version),
    SERVER_SOFTWARE = "https://github.com/r-lib/webfakes"
  )
}

split_cgi_output <- function(x) {
  nlnl <- grepRaw("\r?\n\r?\n", x)[1]
  if (is.na(nlnl)) {
    stop("Invalid response from git cgi, no headers?")
  }

  headers <- parse_headers(rawToChar(x[1:(nlnl - 1L)]))

  body <- x[nlnl:length(x)]
  ndrop <- 1L
  while (body[ndrop] != 0x0a) ndrop <- ndrop + 1L
  ndrop <- ndrop + 1L
  while (body[ndrop] != 0x0a) ndrop <- ndrop + 1L
  body <- utils::tail(body, -ndrop)

  list(headers = headers, body = body)
}

parse_status <- function(x) {
  status <- as.integer(strsplit(x, " ", fixed = TRUE)[[1]][1])
  if (is.na(status)) {
    stop("Invalid status from git cgi: ", x)
  }
  status
}

parse_headers <- function (txt) {
  headers <- grep(":", parse_headers0(txt), fixed = TRUE, value = TRUE)
  out <- lapply(headers, split_header)
  names <- tolower(vapply(out, `[[`, character(1), 1))
  values <- lapply(lapply(out, `[[`, 2), trimws)
  names(values) <- names
  values
}

parse_headers0 <- function (txt, multiple = FALSE) {
  if (!length(txt))
    return(NULL)
  if (is.raw(txt)) {
    txt <- rawToChar(txt)
  }
  stopifnot(is.character(txt))
  if (length(txt) > 1) {
    txt <- paste(txt, collapse = "\n")
  }
  sets <- strsplit(txt, "\\r\\n\\r\\n|\\n\\n|\\r\\r")[[1]]
  headers <- strsplit(sets, "\\r\\n|\\n|\\r")
  if (multiple) {
    headers
  }
  else {
    headers[[length(headers)]]
  }
}

split_header <- function(x) {
  pos <- grepRaw(":", x, fixed = TRUE)[1]
  if (is.na(pos)) {
    stop("Invalid response header from git cgi: ", x)
  }
  c(substr(x, 1, pos - 1L), substr(x, pos + 1L, nchar(x)))
}

parse_authorization_header <- function(x) {
  if (length(x) == 0) return(NULL)

  scheme <- tolower(sub("[ ].*$", "", x))
  rest <- trimws(sub("^[^ ]+[ ]", "", x))

  if (scheme == "basic") {
    username <- password <- NULL
    tryCatch({
      ptxt <- strsplit(base64_decode(rest), ":", fixed = TRUE)[[1]]
      if (length(ptxt) == 2) {
        username <- ptxt[1]
        password <- ptxt[2]
      }
    }, error = function(err) NULL)

    if (!is.null(username) && !is.null(password)) {
      return(list(
        scheme = scheme,
        username = username,
        password = password
      ))
    } else {
      return(NULL)
    }
  }

  # if it has a (non-trailing) =, then it is a dictionary, otherwise token
  if (grepl("=", sub("=+$", "", rest))) {
    return(c(list(scheme = scheme), parse_dict_header(rest)))
  }

  list(scheme = scheme, token = rest)
}

parse_dict_header <- function(x) {
  result <- list()
  for (item in parse_list_header(x)) {
    if (!grepl("=", item)) {
      result[item] <- list(NULL)
      next
    }
    key <- sub("=.*$", "", item)
    value <- sub("^[^=]+=", "", item)

    # https://www.rfc-editor.org/rfc/rfc2231#section-4
    # we always assume UTF-8
    if (grepl("[*]$", key)) {
      key <- sub("[*]$", "", key)
      value <- sub("^.*'.*'", "", value)
      value <- utils::URLdecode(value)
      Encoding(value) <- "UTF-8"
    }

    if (grepl('^".*"$', value)) {
      value <- substr(value, 2, nchar(value) - 1L)
    }

    result[[key]] <- value
  }

  result
}

parse_list_header <- function(x) {
  s <- strsplit(x, "")[[1]]
  res <- character()
  part <- character()
  escape <- quote <- FALSE

  for (cur in s) {
    if (escape) {
      part[length(part) + 1L] <- cur
      escape <- FALSE
      next
    }

    if (quote) {
      if (cur == "\\") {
        escape <- TRUE
        next
      } else if (cur == '"') {
        quote <- FALSE
      }
      part[length(part) + 1L] <- cur
      next
    }

    if (cur == ",") {
      res[length(res) + 1L] <- paste(part, collapse = "")
      part <- character()
      next
    }

    if (cur == '"') {
      quote <- TRUE
    }

    part[length(part) + 1L] <- cur
  }

  if (length(part)) {
    res[length(res) + 1L] <- paste(part, collapse = "")
  }

  trimws(res)
}

#' Middleware function to serve static files
#'
#' The content type of the response is set automatically from the
#' extension of the file. Note that this is a terminal middleware
#' handler function. If a file is served, then the rest of the handler
#' functions will not be called. If a file was not found, however,
#' the rest of the handlers are still called.
#'
#' @param root Root path of the served files. Everything under this
#' directory is served automatically. Directory lists are not currently
#' supports.
#' @param set_headers Callback function to call before a file is served.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' root <- system.file(package = "webfakes", "examples", "static", "public")
#' app <- new_app()
#' app$use(mw_static(root = root))
#' app

mw_static <- function(root, set_headers = NULL) {

  root; set_headers
  function(req, res) {
    path <- file.path(root, sub("^/", "", req$path))
    if (!file.exists(path)) return("next")
    if (file.info(path)$isdir) return("next")
    ext <- tools::file_ext(basename(path))
    ct <- mime_find(ext)
    if (!is.na(ct)) {
      res$set_header("Content-Type", ct)
    }
    if (!is.null(set_headers)) set_headers(req, res)
    res$send(read_bin(path))
  }
}

mime_find <- function(ext) {
  stopifnot(is_string(ext))
  m <- mime_types[ext]
  if (is.na(m)) {
    ew <- str_is_suffix(ext, names(mime_types_sfx))
    m <- mime_types_sfx[ew]
  }
  c(m, NA_character_)[1]
}

mime_types_sfx <- c(
  `3gpp` = "audio/3gpp",
  `jpm` = "video/jpm",
  `mp3` = "audio/mp3",
  `rtf` = "text/rtf",
  `wav` = "audio/wave",
  `x3db` = "model/x3d+binary",
  `x3dv` = "model/x3d+vrml",
  `xml` = "text/xml"
)

mime_types <- c(
  `3g2` = "video/3gpp2",
  `3gp` = "video/3gpp",
  `3gpp` = "video/3gpp",
  `3mf` = "model/3mf",
  ac = "application/pkix-attr-cert",
  adp = "audio/adpcm",
  ai = "application/postscript",
  apng = "image/apng",
  appcache = "text/cache-manifest",
  asc = "application/pgp-signature",
  atom = "application/atom+xml",
  atomcat = "application/atomcat+xml",
  atomsvc = "application/atomsvc+xml",
  au = "audio/basic",
  aw = "application/applixware",
  bdoc = "application/bdoc",
  bin = "application/octet-stream",
  bmp = "image/bmp",
  bpk = "application/octet-stream",
  buffer = "application/octet-stream",
  ccxml = "application/ccxml+xml",
  cdmia = "application/cdmi-capability",
  cdmic = "application/cdmi-container",
  cdmid = "application/cdmi-domain",
  cdmio = "application/cdmi-object",
  cdmiq = "application/cdmi-queue",
  cer = "application/pkix-cert",
  cgm = "image/cgm",
  class = "application/java-vm",
  coffee = "text/coffeescript",
  conf = "text/plain",
  cpt = "application/mac-compactpro",
  crl = "application/pkix-crl",
  css = "text/css",
  csv = "text/csv",
  cu = "application/cu-seeme",
  davmount = "application/davmount+xml",
  dbk = "application/docbook+xml",
  deb = "application/octet-stream",
  def = "text/plain",
  deploy = "application/octet-stream",
  `disposition-notification` = "message/disposition-notification",
  dist = "application/octet-stream",
  distz = "application/octet-stream",
  dll = "application/octet-stream",
  dmg = "application/octet-stream",
  dms = "application/octet-stream",
  doc = "application/msword",
  dot = "application/msword",
  drle = "image/dicom-rle",
  dssc = "application/dssc+der",
  dtd = "application/xml-dtd",
  dump = "application/octet-stream",
  ear = "application/java-archive",
  ecma = "application/ecmascript",
  elc = "application/octet-stream",
  emf = "image/emf",
  eml = "message/rfc822",
  emma = "application/emma+xml",
  eps = "application/postscript",
  epub = "application/epub+zip",
  es = "application/ecmascript",
  exe = "application/octet-stream",
  exi = "application/exi",
  exr = "image/aces",
  ez = "application/andrew-inset",
  fits = "image/fits",
  g3 = "image/g3fax",
  gbr = "application/rpki-ghostbusters",
  geojson = "application/geo+json",
  gif = "image/gif",
  glb = "model/gltf-binary",
  gltf = "model/gltf+json",
  gml = "application/gml+xml",
  gpx = "application/gpx+xml",
  gram = "application/srgs",
  grxml = "application/srgs+xml",
  gxf = "application/gxf",
  gz = "application/gzip",
  h261 = "video/h261",
  h263 = "video/h263",
  h264 = "video/h264",
  heic = "image/heic",
  heics = "image/heic-sequence",
  heif = "image/heif",
  heifs = "image/heif-sequence",
  hjson = "application/hjson",
  hlp = "application/winhlp",
  hqx = "application/mac-binhex40",
  htm = "text/html",
  html = "text/html",
  ics = "text/calendar",
  ief = "image/ief",
  ifb = "text/calendar",
  iges = "model/iges",
  igs = "model/iges",
  img = "application/octet-stream",
  `in` = "text/plain",
  ini = "text/plain",
  ink = "application/inkml+xml",
  inkml = "application/inkml+xml",
  ipfix = "application/ipfix",
  iso = "application/octet-stream",
  jade = "text/jade",
  jar = "application/java-archive",
  jls = "image/jls",
  jp2 = "image/jp2",
  jpe = "image/jpeg",
  jpeg = "image/jpeg",
  jpf = "image/jpx",
  jpg = "image/jpeg",
  jpg2 = "image/jp2",
  jpgm = "video/jpm",
  jpgv = "video/jpeg",
  jpm = "image/jpm",
  jpx = "image/jpx",
  js = "application/javascript",
  json = "application/json",
  json5 = "application/json5",
  jsonld = "application/ld+json",
  jsonml = "application/jsonml+json",
  jsx = "text/jsx",
  jxr = "image/jxr",
  kar = "audio/midi",
  ktx = "image/ktx",
  less = "text/less",
  list = "text/plain",
  litcoffee = "text/coffeescript",
  log = "text/plain",
  lostxml = "application/lost+xml",
  lrf = "application/octet-stream",
  m1v = "video/mpeg",
  m21 = "application/mp21",
  m2a = "audio/mpeg",
  m2v = "video/mpeg",
  m3a = "audio/mpeg",
  m4a = "audio/mp4",
  m4p = "application/mp4",
  ma = "application/mathematica",
  mads = "application/mads+xml",
  man = "text/troff",
  manifest = "text/cache-manifest",
  map = "application/json",
  mar = "application/octet-stream",
  markdown = "text/markdown",
  mathml = "application/mathml+xml",
  mb = "application/mathematica",
  mbox = "application/mbox",
  md = "text/markdown",
  mdx = "text/mdx",
  me = "text/troff",
  mesh = "model/mesh",
  meta4 = "application/metalink4+xml",
  metalink = "application/metalink+xml",
  mets = "application/mets+xml",
  mft = "application/rpki-manifest",
  mid = "audio/midi",
  midi = "audio/midi",
  mime = "message/rfc822",
  mj2 = "video/mj2",
  mjp2 = "video/mj2",
  mjs = "application/javascript",
  mml = "text/mathml",
  mods = "application/mods+xml",
  mov = "video/quicktime",
  mp2 = "audio/mpeg",
  mp21 = "application/mp21",
  mp2a = "audio/mpeg",
  mp3 = "audio/mpeg",
  mp4 = "video/mp4",
  mp4a = "audio/mp4",
  mp4s = "application/mp4",
  mp4v = "video/mp4",
  mpd = "application/dash+xml",
  mpe = "video/mpeg",
  mpeg = "video/mpeg",
  mpg = "video/mpeg",
  mpg4 = "video/mp4",
  mpga = "audio/mpeg",
  mrc = "application/marc",
  mrcx = "application/marcxml+xml",
  ms = "text/troff",
  mscml = "application/mediaservercontrol+xml",
  msh = "model/mesh",
  msi = "application/octet-stream",
  msm = "application/octet-stream",
  msp = "application/octet-stream",
  mxf = "application/mxf",
  mxml = "application/xv+xml",
  n3 = "text/n3",
  nb = "application/mathematica",
  nq = "application/n-quads",
  nt = "application/n-triples",
  oda = "application/oda",
  oga = "audio/ogg",
  ogg = "audio/ogg",
  ogv = "video/ogg",
  ogx = "application/ogg",
  omdoc = "application/omdoc+xml",
  onepkg = "application/onenote",
  onetmp = "application/onenote",
  onetoc = "application/onenote",
  onetoc2 = "application/onenote",
  opf = "application/oebps-package+xml",
  otf = "font/otf",
  owl = "application/rdf+xml",
  oxps = "application/oxps",
  p10 = "application/pkcs10",
  p7c = "application/pkcs7-mime",
  p7m = "application/pkcs7-mime",
  p7s = "application/pkcs7-signature",
  p8 = "application/pkcs8",
  pdf = "application/pdf",
  pfr = "application/font-tdpfr",
  pgp = "application/pgp-encrypted",
  pkg = "application/octet-stream",
  pki = "application/pkixcmp",
  pkipath = "application/pkix-pkipath",
  pls = "application/pls+xml",
  png = "image/png",
  prf = "application/pics-rules",
  ps = "application/postscript",
  pskcxml = "application/pskc+xml",
  qt = "video/quicktime",
  raml = "application/raml+yaml",
  rdf = "application/rdf+xml",
  rif = "application/reginfo+xml",
  rl = "application/resource-lists+xml",
  rld = "application/resource-lists-diff+xml",
  rmi = "audio/midi",
  rnc = "application/relax-ng-compact-syntax",
  rng = "application/xml",
  roa = "application/rpki-roa",
  roff = "text/troff",
  rq = "application/sparql-query",
  rs = "application/rls-services+xml",
  rsd = "application/rsd+xml",
  rss = "application/rss+xml",
  rtf = "application/rtf",
  rtx = "text/richtext",
  s3m = "audio/s3m",
  sbml = "application/sbml+xml",
  scq = "application/scvp-cv-request",
  scs = "application/scvp-cv-response",
  sdp = "application/sdp",
  ser = "application/java-serialized-object",
  setpay = "application/set-payment-initiation",
  setreg = "application/set-registration-initiation",
  sgi = "image/sgi",
  sgm = "text/sgml",
  sgml = "text/sgml",
  shex = "text/shex",
  shf = "application/shf+xml",
  shtml = "text/html",
  sieve = "application/sieve",
  sig = "application/pgp-signature",
  sil = "audio/silk",
  silo = "model/mesh",
  siv = "application/sieve",
  slim = "text/slim",
  slm = "text/slim",
  smi = "application/smil+xml",
  smil = "application/smil+xml",
  snd = "audio/basic",
  so = "application/octet-stream",
  spp = "application/scvp-vp-response",
  spq = "application/scvp-vp-request",
  spx = "audio/ogg",
  sru = "application/sru+xml",
  srx = "application/sparql-results+xml",
  ssdl = "application/ssdl+xml",
  ssml = "application/ssml+xml",
  stk = "application/hyperstudio",
  stl = "model/stl",
  styl = "text/stylus",
  stylus = "text/stylus",
  svg = "image/svg+xml",
  svgz = "image/svg+xml",
  t = "text/troff",
  t38 = "image/t38",
  tei = "application/tei+xml",
  teicorpus = "application/tei+xml",
  text = "text/plain",
  tfi = "application/thraud+xml",
  tfx = "image/tiff-fx",
  tif = "image/tiff",
  tiff = "image/tiff",
  tr = "text/troff",
  ts = "video/mp2t",
  tsd = "application/timestamped-data",
  tsv = "text/tab-separated-values",
  ttc = "font/collection",
  ttf = "font/ttf",
  ttl = "text/turtle",
  txt = "text/plain",
  u8dsn = "message/global-delivery-status",
  u8hdr = "message/global-headers",
  u8mdn = "message/global-disposition-notification",
  u8msg = "message/global",
  uri = "text/uri-list",
  uris = "text/uri-list",
  urls = "text/uri-list",
  vcard = "text/vcard",
  vrml = "model/vrml",
  vtt = "text/vtt",
  vxml = "application/voicexml+xml",
  war = "application/java-archive",
  wasm = "application/wasm",
  wav = "audio/wav",
  weba = "audio/webm",
  webm = "video/webm",
  webmanifest = "application/manifest+json",
  webp = "image/webp",
  wgt = "application/widget",
  wmf = "image/wmf",
  woff = "font/woff",
  woff2 = "font/woff2",
  wrl = "model/vrml",
  wsdl = "application/wsdl+xml",
  wspolicy = "application/wspolicy+xml",
  x3d = "model/x3d+xml",
  x3db = "model/x3d+fastinfoset",
  x3dbz = "model/x3d+binary",
  x3dv = "model/x3d-vrml",
  x3dvz = "model/x3d+vrml",
  x3dz = "model/x3d+xml",
  xaml = "application/xaml+xml",
  xdf = "application/xcap-diff+xml",
  xdssc = "application/dssc+xml",
  xenc = "application/xenc+xml",
  xer = "application/patch-ops-error+xml",
  xht = "application/xhtml+xml",
  xhtml = "application/xhtml+xml",
  xhvml = "application/xv+xml",
  xm = "audio/xm",
  xml = "application/xml",
  xop = "application/xop+xml",
  xpl = "application/xproc+xml",
  xsd = "application/xml",
  xsl = "application/xml",
  xslt = "application/xslt+xml",
  xspf = "application/xspf+xml",
  xvm = "application/xv+xml",
  xvml = "application/xv+xml",
  yaml = "text/yaml",
  yang = "application/yang",
  yin = "application/yin+xml",
  yml = "text/yaml",
  zip = "application/zip"
)

#' Middleware to parse a Range header
#'
#' Adds the requested ranges to the `ranges` element of the request
#' object. `request$ranges` is a data frame with two columns, `from` and
#' `to`. Each row corresponds one requested interval.
#'
#' When the last `n` bytes of the file are requested, the matrix row is set
#' to `c(0, -n)`. When all bytes after a `p` position are requested, the
#' matrix row is set to `c(p, Inf)`.
#'
#' If the intervals overlap, then `ranges` is not set, i.e. the `Range`
#' header is ignored.
#'
#' If its syntax is invalid or the unit is not `bytes`, then the
#' `Range` header is ignored.
#'
#' @return Handler function.
#'
#' @family middleware
#' @export

mw_range_parser <- function() {
  function(req, res) {
    rh <- req$get_header("Range")
    if (length(rh) == 0) return("next")
    req$ranges <- parse_range(rh)
    "next"
  }
}

parse_range <- function(rh) {
  rh <- trimws(rh)
  if (length(rh) == 0 || !startsWith(rh, "bytes=")) return()
  rh <- sub("^bytes=[ ]*", "", rh)
  rngs <- trimws(strsplit(rh, ",", fixed = TRUE)[[1]])
  res <- matrix(integer(1), nrow = length(rngs), ncol = 2)

  for (i in seq_along(rngs)) {
    rng <- strsplit(rngs[i], "-")[[1]]
    if (length(rng) < 1 || length(rng) > 3) {
      return()
    } else if (length(rng) == 1) {
      res[i, 1] <- parse_int(rng[1])
      res[i, 2] <- Inf
      if (is.na(res[i, 1]) || res[i, 1] < 0) return()
    } else if (rng[1] == "") {
      res[i, 1] <- 0
      res[i, 2] <- -parse_int(rng[2])
      if (is.na(res[i, 2]) || res[i, 2] > 0) return()
    } else {
      res[i, 1] <- parse_int(rng[1])
      res[i, 2] <- parse_int(rng[2])
      if (is.na(res[i, 1]) || is.na(res[i, 2]) ||
          res[i, 1] < 0 || res[i, 2] < 0 || res[i, 1] > res[i, 2]) {
        return()
      }
    }
  }

  res <- res[order(res[,1]), , drop = FALSE]

  # check for overlapping intervals
  if (intervals_overlap(res)) return()

  res
}

parse_int <- function(x) {
  suppressWarnings(as.integer(x))
}

intervals_overlap <- function(x) {
  # assume that it is sorted on first column
  # then every interval needs to finish before the next one starts
  if (nrow(x) <= 1) return(FALSE)
  any(x[,2][-nrow(x)] >= x[,1][-1])
}

test_that("parse_range", {
  expect_snapshot({
    parse_range("foobar=1-100")
    parse_range("bytes=0-100, 50-150")
    parse_range("bytes=200-100")
    parse_range("bytes=x-100")

    parse_range("bytes=1-100")
    parse_range("bytes=1-")
    parse_range("bytes=-100")
    parse_range("bytes=0-100, 200-")
    parse_range("bytes=200-300, 0-100")
  })
})

test_that("intervals_overlap", {
  ok <- list(
    matrix(0, nrow = 0, ncol = 2),
    rbind(1:2),
    rbind(1:2, 3:4)
  )
  for (x in ok) expect_false(intervals_overlap(x), info = x)

  bad <- list(
    rbind(c(1, 10), c(5, 10)),
    rbind(c(1, 20), c(1, 20)),
    rbind(c(1, 1), c(1, 4))
  )
  for (x in bad) expect_true(intervals_overlap(x), info = x)
})

library(webfakes)

app <- new_app()
app$engine("txt", tmpl_glue())
app$use(mw_log())

app$get("/hello", function(req, res) {
  res$send("Hello there!")
})

app$get(new_regexp("^/hi(/.*)?$"), function(req, res) {
  res$send("Hi indeed!")
})

app$post("/hello", function(req, res) {
  res$send("Got it, thanks!")
})

app$get("/view", function(req, res) {
  txt <- res$render("test")
  res$
    set_type("text/plain")$
    send(txt)
})

app$listen(3000L)

re_match <- function(text, pattern, perl = TRUE, ...) {

  stopifnot(is.character(pattern), length(pattern) == 1, !is.na(pattern))
  text <- as.character(text)

  match <- regexpr(pattern, text, perl = perl, ...)

  start  <- as.vector(match)
  length <- attr(match, "match.length")
  end    <- start + length - 1L

  matchstr <- substring(text, start, end)
  matchstr[ start == -1 ] <- NA_character_

  empty <- data.frame(stringsAsFactors = FALSE, .text = text)[, numeric()]
  res <- list(match = !is.na(matchstr), groups = empty)

  if (!is.null(attr(match, "capture.start"))) {

    gstart  <- attr(match, "capture.start")
    glength <- attr(match, "capture.length")
    gend    <- gstart + glength - 1L

    groupstr <- substring(text, gstart, gend)
    groupstr[ gstart == -1 ] <- NA_character_
    dim(groupstr) <- dim(gstart)

    res$groups <- cbind(groupstr, res$groups, stringsAsFactors = FALSE)
    names(res$groups) <- attr(match, "capture.names")
  }

  res
}

#' Generic web app for testing HTTP clients
#'
#' A web app similar to `https://httpbin.org`.
#' See [its specific docs](https://webfakes.r-lib.org/httpbin.html).
#' You can also see these docs locally, by starting the app:
#' ```r
#' httpbin <- new_app_process(httpbin_app())
#' browseURL(httpbin$url())
#' ```
#'
#' @param log Whether to log requests to the standard output.
#' @return A `webfakes_app`.
#'
#' @export
#' @examples
#' app <- httpbin_app()
#' proc <- new_app_process(app)
#' url <- proc$url("/get")
#' resp <- curl::curl_fetch_memory(url)
#' curl::parse_headers_list(resp$headers)
#' cat(rawToChar(resp$content))
#' proc$stop()

httpbin_app <- function(log = interactive()) {
  encode_files <- function(files) {
    for (i in seq_along(files)) {
      files[[i]]$value <- paste0(
        "data:application/octet-stream;base64,",
        base64_encode(files[[i]]$value)
      )
    }
    files
  }

  app <- new_app()

  # Log requests by default
  if (log) {
    app$use("logger" = mw_log())
  }

  # Parse all kinds of bodies
  app$use("json body parser" = mw_json())
  app$use(
    "text body parser" = mw_text(type = c("text/plain", "application/json"))
  )
  app$use("multipart body parser" = mw_multipart())
  app$use("URL encoded body parser" = mw_urlencoded())
  app$use("cookie parser" = mw_cookie_parser())

  # Add etags by default
  app$use("add etag" = mw_etag())

  # Add date by default
  app$use("add date" = function(req, res) {
    res$set_header("Date", http_time_stamp())
    "next"
  })

  make_common_response <- function(req, res) {
    ret <- list(
      args = as.list(req$query),
      data = req$text,
      files = encode_files(req$files),
      form = req$form,
      headers = req$headers,
      json = req$json,
      method = req$method,
      path = req$path,
      origin = req$remote_addr,
      url = req$url
    )
  }

  common_response <- function(req, res) {
    ret <- make_common_response(req, res)
    res$send_json(object = ret, auto_unbox = TRUE, pretty = TRUE)
  }

  # Main page

  app$get("/", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes", "examples", "httpbin", "assets"),
      "httpbin.html"
    )
  })

  # HTTP methods =========================================================

  common_get <- function(req, res) {
    ret <- list(
      args = as.list(req$query),
      headers = req$headers,
      origin = req$remote_addr,
      path = req$path,
      url = req$url
    )
    res$send_json(object = ret, auto_unbox = TRUE, pretty = TRUE)
  }

  app$get("/get", common_get)

  app$delete("/delete", common_response)
  app$patch("/patch", common_response)
  app$post("/post", common_response)
  app$put("/put", common_response)

  app$get("/forms/post", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes", "examples", "httpbin", "assets"),
      "forms-post.html"
    )
  })

  # Auth =================================================================

  basic_auth <- function(req, res, error_status = 401L) {
    exp <- paste(
      "Basic",
      base64_encode(paste0(req$params$user, ":", req$params$passwd))
    )
    hdr <- req$get_header("Authorization") %||% ""
    if (exp == hdr) {
      res$send_json(list(
        authenticated = jsonlite::unbox(TRUE),
        user = jsonlite::unbox(req$params$user)
      ))
    } else {
      if (error_status == 401L) {
        res$set_header(
          "WWW-Authenticate",
          "Basic realm=\"Fake Realm\""
        )$send_status(error_status)
      } else {
        res$send_status(error_status)
      }
    }
  }

  app$get("/basic-auth/:user/:passwd", function(req, res) {
    basic_auth(req, res, error_status = 401L)
  })

  app$get("/hidden-basic-auth/:user/:passwd", function(req, res) {
    basic_auth(req, res, error_status = 404L)
  })

  app$get("/bearer", function(req, res) {
    auth <- req$get_header("Authorization") %||% ""
    if (!grepl("^Bearer ", auth)) {
      res$set_header("WWW-Authenticate", "bearer")$send_status(401L)
    } else {
      token <- sub("^Bearer ", "", auth)
      res$send_json(
        list(authenticated = TRUE, token = token),
        auto_unbox = TRUE,
        pretty = TRUE
      )
    }
  })

  hash <- function(str, algorithm) {
    algo <- tolower(algorithm %||% "md5")
    algo <- c("md5" = "md5", "sha-256" = "sha256", "sha-512" = "sha512")[algo]
    if (is.na(algo)) {
      stop("Unknown hash algorithm for digest auth: ", algorithm)
    }
    digest::digest(str, algo = algo, serialize = FALSE)
  }

  hash1 <- function(realm, username, password, algorithm) {
    realm <- realm %||% ""
    hash(
      paste(
        collapse = ":",
        c(
          username,
          realm,
          password
        )
      ),
      algorithm
    )
  }

  hash2 <- function(credentials, req, algorithm) {
    qop <- credentials[["qop"]] %||% "auth"
    query <- if (nchar(req$query_string %||% "")) paste0("?", req$query_string)
    req_uri <- paste0(req$path, query)
    if (qop == "auth") {
      hash(
        paste(
          collapse = ":",
          c(
            toupper(req[["method"]]),
            req_uri
          )
        ),
        algorithm
      )
    } else {
      hash(
        paste0(
          collapse = ":",
          c(
            toupper(req[["method"]]),
            req_uri,
            hash(req$.body %||% "", algorithm)
          )
        ),
        algorithm
      )
    }
  }

  digest_challenge_response <- function(req, qop, algorithm, stale = FALSE) {
    nonce <- hash(
      paste0(req$remote_addr, ":", unclass(Sys.time()), ":", random_id(10)),
      algorithm
    )

    opaque <- hash(random_id(10), algorithm)
    realm <- "webfakes.r-lib.org"
    qop <- qop %||% "auth,auth-int"

    wwwauth <- paste0(
      "Digest ",
      "realm=\"",
      realm,
      "\", ",
      "qop=\"",
      qop,
      "\", ",
      "nonce=\"",
      nonce,
      "\", ",
      "opaque=\"",
      opaque,
      "\", ",
      "algorithm=",
      algorithm,
      ", ",
      "stale=",
      stale
    )

    wwwauth
  }

  next_stale_after_value <- function(x) {
    x <- suppressWarnings(as.integer(x))
    if (is.na(x)) "never" else as.character(x - 1L)
  }

  check_digest_auth <- function(req, credentials, user, passwd) {
    if (is.null(credentials)) {
      return(FALSE)
    }
    algorithm <- credentials[["algorithm"]]
    HA1_value <- hash1(
      credentials[["realm"]],
      credentials[["username"]],
      passwd,
      algorithm
    )
    HA2_value <- hash2(credentials, req, algorithm)

    qop <- credentials[["qop"]] %||% "compat"
    if (!qop %in% c("compat", "auth", "auth-int")) {
      return(FALSE)
    }

    response_hash <- if (qop == "compat") {
      hash(
        paste0(
          c(
            HA1_value,
            credentials[["nonce"]] %||% "",
            HA2_value
          ),
          collapse = ":"
        ),
        algorithm
      )
    } else {
      if (any(!c("nonce", "nc", "cnonce", "qop") %in% names(credentials))) {
        return(FALSE)
      }
      hash(
        paste0(
          c(
            HA1_value,
            credentials[["nonce"]],
            credentials[["nc"]],
            credentials[["cnonce"]],
            credentials[["qop"]],
            HA2_value
          ),
          collapse = ":"
        ),
        algorithm
      )
    }

    (credentials[["response"]] %||% "") == response_hash
  }

  digest_auth <- function(req, res, qop, user, passwd, algorithm, stale_after) {
    require_cookie_handling <-
      tolower(req$query$`require-cookie` %||% "") %in% c("1", "t", "true")
    if (!algorithm %in% c("MD5", "SHA-256", "SHA-512")) {
      algorithm <- "MD5"
    }

    if (!qop %in% c("auth", "auth-int")) {
      qop <- NULL
    }

    authorization <- req$get_header("Authorization")
    credentials <- if (!is.null(authorization)) {
      parse_authorization_header(authorization)
    }

    if (
      is.null(authorization) ||
        is.null(credentials) ||
        tolower(credentials$scheme) != "digest" ||
        (require_cookie_handling && is.null(req$get_header("Cookie")))
    ) {
      wwwauth <- digest_challenge_response(req, qop, algorithm)
      res$set_status(401L)$add_cookie("stale_after", stale_after)$add_cookie(
        "fake",
        "fake_value"
      )$set_header("WWW-Authenticate", wwwauth)$send("")
      return()
    }

    if (
      require_cookie_handling &&
        (req$cookies[["fake"]] %||% "") != "fake_value"
    ) {
      res$add_cookie("fake", "fake_value")$add_status(403L)$send_json(
        list(errors = "missing cookie set on challenge"),
        pretty = TRUE
      )
      return()
    }

    current_nonce <- credentials$nonce
    stale_after_value <- req$cookies$stale_after %||% ""
    if (
      identical(current_nonce, req$cookies[["last_nonce"]] %||% "") ||
        stale_after_value == "0"
    ) {
      wwwauth <- digest_challenge_response(req, qop, algorithm, TRUE)
      res$set_status(401L)$add_cookie("stale_after", stale_after)$add_cookie(
        "last_nonce",
        current_nonce
      )$add_cookie("fake", "fake_value")$set_header(
        "WWW-Authenticate",
        wwwauth
      )$send("")
      return()
    }

    if (!check_digest_auth(req, credentials, user, passwd)) {
      wwwauth <- digest_challenge_response(req, qop, algorithm, FALSE)
      res$set_status(401L)$add_cookie("stale_after", stale_after)$add_cookie(
        "last_nonce",
        current_nonce
      )$add_cookie("fake", "fake_value")$set_header(
        "WWW-Authenticate",
        wwwauth
      )$send("")
      return()
    }

    res$add_cookie("fake", "fake_value")
    if (!is.null(stale_after_value)) {
      res$add_cookie(
        "stale_after",
        next_stale_after_value(stale_after_value)
      )
    }

    res$send_json(
      list(authentication = TRUE, user = user),
      pretty = TRUE,
      auto_unbox = TRUE
    )
  }

  app$get("/digest-auth/:qop/:user/:passwd", function(req, res) {
    qop <- req$params$qop
    user <- req$params$user
    passwd <- req$params$passwd
    digest_auth(req, res, qop, user, passwd, "MD5", "never")
  })

  app$get("/digest-auth/:qop/:user/:passwd/:algorithm", function(req, res) {
    qop <- req$params$qop
    user <- req$params$user
    passwd <- req$params$passwd
    algorithm <- req$params$algorithm
    digest_auth(req, res, qop, user, passwd, algorithm, "never")
  })

  app$get(
    "/digest-auth/:qop/:user/:passwd/:algorithm/:stale_after",
    function(req, res) {
      qop <- req$params$qop
      user <- req$params$user
      passwd <- req$params$passwd
      algorithm <- req$params$algorithm
      stale_after <- req$params$stale_after
      digest_auth(req, res, qop, user, passwd, algorithm, stale_after)
    }
  )

  # Status codes =========================================================

  app$all(
    new_regexp("^/status/(?<status>[0-9][0-9][0-9])$"),
    function(req, res) {
      status <- req$params$status
      res$set_status(status)
      if (status == "418") {
        res$send(paste(
          sep = "\n",
          "",
          "    -=[ teapot ]=-",
          "",
          "       _...._",
          "     .'  _ _ `.",
          "    | .\"` ^ `\". _,",
          "    \\_;`\"---\"`|//",
          "      |       ;/",
          "      \\_     _/",
          "        `\"\"\"`",
          ""
        ))
      } else {
        res$send("")
      }
    }
  )

  # Request inspection ===================================================

  app$get("/headers", function(req, res) {
    ret <- list(headers = req$headers)
    res$send_json(ret, auto_unbox = TRUE, pretty = TRUE)
  })

  app$get("/ip", function(req, res) {
    ret <- list(origin = req$remote_addr)
    res$send_json(ret, auto_unbox = TRUE, pretty = TRUE)
  })

  app$get("/user-agent", function(req, res) {
    ret <- list("user-agent" = req$get_header("User-Agent"))
    res$send_json(ret, auto_unbox = TRUE, pretty = TRUE)
  })

  # Response inspection ==================================================

  app$get("/etag/:etag", function(req, res) {
    etag <- req$params$etag

    # The mw_etag() middleware is active, so we need to do this after that
    res_etag <- NULL
    res$on_response(function(req, res) {
      if (!is.null(res_etag)) res$set_header("ETag", res_etag)
    })

    parse <- function(x) {
      x <- strsplit(x, ",", fixed = TRUE)[[1]]
      re_match(x, '\\s*(W/)?"?([^"]*)"?\\s*')$groups[, 2]
    }

    if_none_match <- parse(req$get_header("If-None-Match") %||% "")
    if_match <- parse(req$get_header("If-Match") %||% "")

    if (length(if_none_match) > 0) {
      if (etag %in% if_none_match || "*" %in% if_none_match) {
        res$send_status(304)
        res_etag <- "etag"
        return()
      }
    } else if (length(if_match) > 0) {
      if ((!etag %in% if_match) && (!"*" %in% if_match)) {
        res$send_status(412)
        return()
      }
    }

    res_etag <- etag
    common_get(req, res)
  })

  rsp_hdrs <- function(req, res) {
    obj <- structure(list(), names = character())
    for (i in seq_along(req$query)) {
      key <- names(req$query)[i]
      res$add_header(key, req$query[[i]])
      obj[[key]] <- c(obj[[key]], req$query[[i]])
    }

    res$send_json(object = obj, auto_unbox = TRUE)
  }
  app$get("/response-headers", rsp_hdrs)
  app$post("/response-headers", rsp_hdrs)

  app$get("/cache", function(req, res) {
    if (
      is.null(req$get_header("If-Modified-Since")) &&
        is.null(req$get_header("If-None-Match"))
    ) {
      res$set_header("Last-Modified", http_time_stamp())
      # etag is added by default
      common_response(req, res)
    } else {
      res$send_status(304)
    }
  })

  app$get("/cache/:value", function(req, res) {
    value <- suppressWarnings(as.integer(req$params$value))
    if (is.na(value)) {
      "next"
    } else {
      res$set_header(
        "Cache-Control",
        sprintf("public, max-age=%d", value)
      )
      common_response(req, res)
    }
  })

  # Response formats =====================================================

  app$get("/deny", function(req, res) {
    res$set_type("text/plain")$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "deny.txt")
    )
  })

  app$get("/brotli", function(req, res) {
    ret <- make_common_response(req, res)
    ret$brotli <- TRUE
    json <- jsonlite::toJSON(ret, auto_unbox = TRUE, pretty = TRUE)
    data <- charToRaw(json)
    datax <- brotli::brotli_compress(data)
    res$set_type("application/json")$set_header(
      "Content-Encoding",
      "brotli"
    )$send(datax)
  })

  app$get("/gzip", function(req, res) {
    ret <- make_common_response(req, res)
    ret$gzipped <- TRUE
    json <- jsonlite::toJSON(ret, auto_unbox = TRUE, pretty = TRUE)
    tmp <- tempfile()
    on.exit(unlink(tmp), add = TRUE)
    con <- file(tmp, open = "wb")
    con2 <- gzcon(con)
    writeBin(charToRaw(json), con2)
    flush(con2)
    close(con2)
    gzipped <- readBin(tmp, "raw", file.info(tmp)$size)
    res$set_type("application/json")$set_header(
      "Content-Encoding",
      "gzip"
    )$send(gzipped)
  })

  app$get("/deflate", function(req, res) {
    ret <- make_common_response(req, res)
    ret$deflated <- TRUE
    json <- jsonlite::toJSON(ret, auto_unbox = TRUE, pretty = TRUE)
    data <- charToRaw(json)
    datax <- zip::deflate(data)
    res$set_type("application/json")$set_header(
      "Content-Encoding",
      "deflate"
    )$send(datax$output)
  })

  app$get("/encoding/utf8", function(req, res) {
    res$set_type("text/html; charset=utf-8")$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "utf8.html")
    )
  })

  app$get("/html", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "example.html")
    )
  })

  app$get("/json", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "example.json")
    )
  })

  app$get("/robots.txt", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "robots.txt")
    )
  })

  app$get("/xml", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "httpbin", "data", "example.xml")
    )
  })

  # Dynamic data =========================================================

  app$get(
    list("/base64", new_regexp("/base64/(?<value>[\\+/=a-zA-Z0-9]*)")),
    function(req, res) {
      value <- req$params$value %||% ""
      if (value == "") {
        value <- "RXZlcnl0aGluZyBpcyBSc29tZQ=="
      }
      plain <- charToRaw(base64_decode(value))
      res$set_type("application/octet-stream")$send(plain)
    }
  )

  app$get("/bytes/:n", function(req, res) {
    n <- suppressWarnings(as.integer(req$params$n))
    if (is.na(n)) {
      return("next")
    } else {
      n <- min(n, 10000)
      bytes <- as.raw(as.integer(floor(stats::runif(n, min = 0, max = 256))))
      res$set_type("application/octet-stream")$send(bytes)
    }
  })

  app$all(new_regexp("/delay/(?<delay>[0-9\\.]+)$"), function(req, res) {
    delay <- suppressWarnings(as.numeric(req$params$delay))
    if (is.na(delay)) {
      return("next")
    } else if (is.null(res$locals$seen)) {
      res$locals$seen <- TRUE
      # we need to send the headers early, to work around
      # https://github.com/r-lib/webfakes/issues/108
      # but then if we want to send content-length and etag, we
      # need to calculate the response here and save it
      res$locals$response <- jsonlite::toJSON(
        make_common_response(req, res),
        auto_unbox = TRUE,
        pretty = TRUE
      )
      content_length <- nchar(res$locals$response, type = "bytes")
      res$set_header('Content-Length', content_length)
      etag <- paste0("\"", crc32(res$locals$response), "\"")
      res$set_header("ETag", etag)
      # this actually cannot happen, because the ETag header is also
      # part of the response, so it is included in the ETag calculation
      # So there is no ETag that could ever match here. Nevertheless we
      # check for it, for future reference.
      req_etag <- req$get_header("If-None-Match")
      if (!is.null(req_etag) && req_etag == etag) {
        res$locals$response <- NULL
        res$set_status(304)
      } else {
        res$set_status(200)
      }
      res$set_type("application/json")
      res$write(raw(0))
      delay <- min(delay, 10)
      res$delay(delay)
    } else if (req$method == "head") {
      res$send_status(200L)
    } else {
      res$send(res$locals$response)
    }
  })

  app$get("/drip", function(req, res) {
    # First time?
    if (is.null(res$locals$drip)) {
      duration <- as.double(req$query$duration %||% 2)
      numbytes <- as.integer(req$query$numbytes %||% 10)
      code <- as.integer(req$query$code %||% 200L)
      delay <- as.double(req$query$delay %||% 0)
      # how much to wait between messages, at least 10ms
      pause <- max(duration / numbytes, 0.01)
      # how many messages
      nummsg <- duration / pause + 1
      # how big is a message, at least a byte
      msgsize <- max(floor(numbytes / nummsg), 1)
      res$locals$drip <- list(
        tosend = numbytes,
        msgsize = msgsize,
        pause = pause
      )

      res$set_header("Content-Length", numbytes)$set_header(
        "Content-Type",
        "application/octet-stream"
      )$set_status(code)

      if (delay > 0) return(res$delay(delay))
    }

    len <- min(res$locals$drip$tosend, res$locals$drip$msgsize)
    res$write(strrep("*", len))
    res$locals$drip$tosend <- res$locals$drip$tosend - len
    if (res$locals$drip$tosend == 0) {
      res$send("")
    } else {
      res$delay(res$locals$drip$pause)
    }
  })

  app$get(new_regexp("^/stream/(?<n>[0-9]+)$"), function(req, res) {
    n <- suppressWarnings(as.integer(req$params$n))
    n <- min(n, 100)
    if (length(n) == 0 || is.na(n)) {
      return("next")
    }
    msg <- make_common_response(req, res)[c("url", "args", "headers", "origin")]

    res$set_type("application/json")

    for (i in seq_len(n)) {
      msg$id <- i - 1L
      txt <- paste0(jsonlite::toJSON(msg, auto_unbox = TRUE), "\n")
      res$send_chunk(charToRaw(txt))
    }
  })

  app$get(new_regexp("^/stream-bytes/(?<n>[0-9]+)$"), function(req, res) {
    n <- suppressWarnings(as.integer(req$params$n))
    n <- min(n, 100 * 1024)
    seed <- suppressWarnings(as.integer(req$query$seed %||% 42))
    chunk_size <- suppressWarnings(as.integer(req$query$chunk_size %||% 10240))
    if (
      length(n) == 0 ||
        is.na(n) ||
        length(seed) == 0 ||
        is.na(seed) ||
        length(chunk_size) == 0 ||
        is.na(chunk_size)
    ) {
      return("next")
    }
    oldseed <- .GlobalEnv$.Random.seed
    on.exit(set.seed(oldseed))
    set.seed(seed)
    bytes <- as.raw(as.integer(floor(stats::runif(n, min = 0, max = 256))))
    nc <- ceiling(n / chunk_size)
    for (i in seq_len(nc)) {
      from <- (i - 1) * chunk_size + 1
      to <- min(length(bytes), i * chunk_size)
      res$send_chunk(bytes[from:to])
    }
  })

  re_range <- new_regexp("^/range/(?<numbytes>[0-9]+)$")

  # This is not in httpbin, but it is handy to get the size of the
  # response, and to see whether the server supports ranges

  app$head(re_range, function(req, res) {
    numbytes <- suppressWarnings(as.integer(req$params$n))
    if (length(numbytes) == 0 || is.na(numbytes)) {
      return("next")
    }

    res$set_header("ETag", paste0("range", numbytes))$set_header(
      "Accept-Ranges",
      "bytes"
    )$set_header("Content-Length", numbytes)$send_status(200L)
  })

  app$get(re_range, function(req, res) {
    if (is.null(res$locals$range)) {
      numbytes <- suppressWarnings(as.integer(req$params$n))
      if (length(numbytes) == 0 || is.na(numbytes)) {
        return("next")
      }

      if (numbytes < 0 || numbytes > 100 * 1024) {
        res$set_header("ETag", paste0("range", numbytes))$set_header(
          "Accept-Ranges",
          "bytes"
        )$set_status(404L)$send(
          "number of bytes must be in the range (0, 102400]."
        )
        return()
      }

      chunk_size <- max(1, as.integer(req$query$chunk_size %||% (10 * 1024)))
      duration <- as.integer(req$query$duration %||% 0)
      pause_per_byte <- duration / numbytes
      if (duration == 0) {
        chunk_size <- numbytes
      }

      ranges <- parse_range(req$get_header("Range"))

      # just like httpbin, we do not support multiple ranges
      if (NROW(ranges) != 1) {
        ranges <- NULL
      }

      # This is not exactly the same as httpbin, but rather follows
      # https://developer.mozilla.org/en-US/docs/Web/HTTP/Range_requests
      # and also how web servers seem to behave.
      #
      # In particular, in these cases we return the full response:
      # - no Range header,
      # - invalid Range header syntax,
      # - overlapping Range header ranges
      #
      # Otherwise, if a range is outside of the size of the response, we
      # return a 416 response.

      if (!is.null(ranges)) {
        ranges[ranges == Inf] <- numbytes
        if (any(ranges[, 2] >= numbytes)) {
          res$set_header("ETag", paste0("range", numbytes))$set_header(
            "Accept-Ranges",
            "bytes"
          )$set_header(
            "Content-Range",
            paste0("bytes */", numbytes)
          )$set_header("Content-Length", 0L)$send_status(416L)
          return()
        }
      }

      # we need the response for sure
      abc <- paste(letters, collapse = "")
      bytes <- substr(strrep(abc, numbytes / nchar(abc) + 1), 1, numbytes)

      res$locals$range <- list(
        bytes = bytes,
        chunk_size = chunk_size,
        pause_per_byte = pause_per_byte
      )

      # First part, so send status and headers
      if (is.null(ranges)) {
        res$set_header("ETag", paste0("range", numbytes))$set_header(
          "Accept-Ranges",
          "bytes"
        )$set_header("Content-Length", numbytes)$set_status(200L)
      } else if (nrow(ranges) == 1) {
        # A single range
        res$set_header("ETag", paste0("range", numbytes))$set_header(
          "Accept-Ranges",
          "bytes"
        )$set_header(
          "Content-Range",
          sprintf("bytes=%d-%d/%d", ranges[1, 1], ranges[1, 2], numbytes)
        )$set_header(
          "Content-Length",
          ranges[1, 2] - ranges[1, 1] + 1L
        )$set_status(206L)

        # This is all we need to send
        res$locals$range$bytes <- substr(
          bytes,
          ranges[1, 1] + 1,
          ranges[1, 2] + 1L
        )
      } else {
        # This cannot happen now, we do not support multiple ranges
        # Maybe later
      }
    }

    # send a part
    chunk_size <- res$locals$range$chunk_size
    pause <- res$locals$range$pause_per_byte

    tosend <- substr(res$locals$range$bytes, 1, chunk_size)
    res$locals$range$bytes <- substr(
      res$locals$range$bytes,
      chunk_size + 1L,
      nchar(res$locals$range$bytes)
    )
    res$write(tosend)
    if (pause * nchar(tosend) > 0) {
      res$delay(pause * nchar(tosend))
    }
  })

  app$get("/uuid", function(req, res) {
    ret <- list(uuid = uuid_random())
    res$send_json(ret, auto_unbox = TRUE, pretty = TRUE)
  })

  app$get(
    new_regexp("^/links/(?<n>[0-9]+)(/(?<offset>[0-9]+))?$"),
    function(req, res) {
      n <- suppressWarnings(as.integer(req$params$n))
      o <- suppressWarnings(as.integer(req$params$offset))
      if (length(o) == 0 || is.na(o)) {
        o <- 1
      }
      if (length(n) == 0 || is.na(n)) {
        return("next")
      }
      n <- min(max(1, n), 200)
      o <- min(max(1, o), n)
      links <- sprintf("<a href = \"/links/%d/%d\">%d</a>", n, 1:n, 1:n)
      links[o] <- o
      html <- paste0(
        "<html><head><title>Links</title></head><body>",
        paste(links, collapse = " "),
        "</body></html>"
      )
      res$set_type("html")$send(html)
    }
  )

  # Cookies ==============================================================

  app$get("/cookies", function(req, res) {
    cks <- req$cookies
    res$send_json(
      object = list(cookies = cks),
      auto_unbox = TRUE,
      pretty = TRUE
    )
  })

  app$get("/cookies/set/:name/:value", function(req, res) {
    res$add_cookie(req$params$name, req$params$value)
    res$redirect("/cookies", 302L)
  })

  app$get("/cookies/set", function(req, res) {
    for (n in names(req$query)) {
      res$add_cookie(n, req$query[[n]])
    }
    res$redirect("/cookies", 302L)
  })

  app$get("/cookies/delete", function(req, res) {
    for (n in names(req$query)) {
      res$clear_cookie(n)
    }
    res$redirect("/cookies", 302L)
  })

  # Images ===============================================================

  app$get("/image", function(req, res) {
    act <- req$get_header("Accept")
    ok <- c(
      "image/webp",
      "image/svg+xml",
      "image/jpeg",
      "image/png",
      "image/*"
    )
    msg <- list(
      message = "Client did not request a supported media type.",
      accept = ok
    )
    if (is.null(act) || !act %in% ok) {
      res$set_status(406)$set_type("application/json")$send_json(msg)
    } else {
      fls <- c(
        "image/webp" = "Rlogo.webp",
        "image/svg+xml" = "Rlogo.svg",
        "image/jpeg" = "Rlogo.jpeg",
        "image/png" = "Rlogo.png",
        "image/*" = "Rlogo.png"
      )
      res$send_file(
        root = system.file(package = "webfakes"),
        file.path("examples", "httpbin", "images", fls[act])
      )
    }
  })

  app$get(
    new_regexp("/image/(?<format>jpeg|png|svg|webp)"),
    function(req, res) {
      filename <- paste0("Rlogo.", req$params$format)
      res$send_file(
        root = system.file(package = "webfakes"),
        file.path("examples", "httpbin", "images", filename)
      )
    }
  )

  # Redirects ============================================================

  app$get("/absolute-redirect/:n", function(req, res) {
    n <- suppressWarnings(as.integer(req$params$n))
    if (is.na(n)) {
      return("next")
    } else {
      if (n == 1) {
        url <- sub("/absolute-redirect/[0-9]+$", "/get", req$url)
      } else {
        n <- min(n, 5)
        url <- paste0(sub("/[0-9]+$", "/", req$url), n - 1)
      }
      res$redirect(url, 302L)
    }
  })

  app$get(c("/redirect/:n", "/relative-redirect/:n"), function(req, res) {
    n <- suppressWarnings(as.integer(req$params$n))
    if (is.na(n)) {
      return("next")
    } else {
      if (n == 1) {
        url <- sub("/redirect/[0-9]+$", "/get", req$path)
        url <- sub("/relative-redirect/[0-9]+$", "/get", url)
      } else {
        n <- min(n, 5)
        url <- paste0(sub("/[0-9]+$", "/", req$path), n - 1)
      }
      res$redirect(url, 302L)
    }
  })

  app$all("/redirect-to", function(req, res) {
    res$redirect(req$query$url, req$query$status_code %||% 302)
  })

  # Anything =============================================================

  app$all(new_regexp("^/anything"), common_response)

  app
}

#' App process that is cleaned up automatically
#'
#' You can start the process with an explicit `$start()` call.
#' Alternatively it starts up at the first `$url()` or `$get_port()`
#' call.
#'
#' @inheritParams new_app_process
#' @param ... Passed to [new_app_process()].
#' @param .local_envir The environment to attach the process cleanup to.
#'   Typically a frame. When this frame finishes, the process is stopped.
#'
#' @export
#' @seealso [new_app_process()] for more details.

local_app_process <- function(app, ..., .local_envir = parent.frame()) {
  proc <- new_app_process(app, ...)
  withr::defer(proc$stop(), envir = .local_envir)
  proc
}

test_that("get", {
  url <- httpbin$url("/get")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(data$path, "/get")
})

test_that("post", {
  url <- httpbin$url("/post")
  data <- charToRaw(jsonlite::toJSON(list(foo = "bar", foobar = 1:3)))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "Content-Type" = "application/json")
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  data <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
  expect_equal(
    data$json,
    list(foo = "bar", foobar = 1:3)
  )
  expect_equal(data$path, "/post")
})

test_methods <- c("connect", "delete", "head", "mkcol", "options",
                  "patch", "propfind", "put", "report")

app <- new_app()
handler <- function(req, res) res$send_json(list(method = req$method))
for (method in test_methods) app[[method]](paste0("/", method), handler)
web2 <- local_app_process(app, port = NA)

test_that("the rest", {
  for (method in test_methods) {
    url <- web2$url(paste0("/", method))
    handle <- curl::new_handle()
    curl::handle_setheaders(handle, "content-type" = "application/json")
    curl::handle_setopt(handle, customrequest = toupper(method))

    resp <- curl::curl_fetch_memory(url, handle = handle)
    expect_equal(resp$status, 200)
    if (method == "head") {
      expect_equal(length(resp$content), 0)
    } else {
      echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = TRUE)
      expect_equal(echo$method, method)
    }
  }
})

test_that("git_app", {
  skip_on_cran()
  dir.create(tmp <- tempfile())
  on.exit(unlink(tmp, recursive = TRUE), add = TRUE)
  untar(testthat::test_path("fixtures/git-repo.tar.gz"), exdir = tmp)
  app <- git_app(file.path(tmp, "repo"))
  git <- webfakes::local_app_process(app)

  expect_snapshot(
    system(
      paste("git ls-remote", git$url("/pak-test.git")),
      intern = TRUE
    )
  )
})

#' A webfakes request object
#'
#' webfakes creates a `webfakes_request` object for every incoming HTTP
#' request. This object is passed to every matched route and middleware,
#' until the response is sent. It has reference semantics, so handlers
#' can modify it.
#'
#' Fields and methods:
#'
#' * `app`: The `webfakes_app` object itself.
#' * `headers`: Named list of HTTP request headers.
#' * `hostname`: The Host header, the server hostname and maybe port.
#' * `method`: HTTP method.
#' * `path`: Server path.
#' * `protocol`: `"http"` or `"https"`.
#' * `query_string`: The raw query string, without the starting `?`.
#' * `query`: Parsed query parameters in a named list.
#' * `remote_addr`: String, the domain name or IP address of the client.
#'    webfakes runs on the localhost, so this is `127.0.0.1`.
#' * `url`: The full URL of the request.
#' * `get_header(field)`: Function to query a request header. Returns
#'   `NULL` if the header is not present.
#'
#' Body parsing middleware adds additional fields to the request object.
#' See [mw_raw()], [mw_text()], [mw_json()], [mw_multipart()] and
#' [mw_urlencoded()].
#'
#' @seealso [webfakes_response] for the webfakes response object.
#' @name webfakes_request
#' @examples
#' # This is how you can see the request and response objects:
#' app <- new_app()
#' app$get("/", function(req, res) {
#'   browser()
#'   res$send("done")
#' })
#' app
#'
#' # Now start this app on a port:
#' # app$listen(3000)
#' # and connect to it from a web browser: http://127.0.0.1:3000
#' # You can also use another R session to connect:
#' # httr::GET("http://127.0.0.1:3000")
#' # or the command line curl tool:
#' # curl -v http://127.0.0.1:3000
#' # The app will stop while processing the request.
NULL

new_request <- function(app, self) {
  if (isTRUE(self$.has_methods)) return(self)
  parsed_headers <- self$headers
  self$.has_methods <- TRUE
  self$app <- app
  self$headers <- parsed_headers
  self$hostname <- parsed_headers$host
  self$method <- tolower(self$method)
  self$protocol <- "http"
  self$query = parse_query(self$query_string)
  self$get_header <- function(field) {
    h <- self$headers
    names(h) <- tolower(names(h))
    h[[tolower(field)]]
  }
  rm(parsed_headers)

  self$res <- new_response(app, self)

  class(self) <- c("webfakes_request", class(self))
  self
}

parse_query <- function(query) {
  query <- sub("^[?]", "", query)
  query <- chartr("+", " ", query)
  argstr <- strsplit(query, "&", fixed = TRUE)[[1]]
  argstr <- strsplit(argstr, "=", fixed = TRUE)
  keys <- vapply(argstr, function(x) utils::URLdecode(x[[1]]), character(1))
  vals <- lapply(argstr, function(x) {
    if (length(x) == 2) utils::URLdecode(x[[2]]) else ""
  })
  structure(vals, names = keys)
}

#' Middleware to parse an url-encoded request body
#'
#' This is typically data from a form. The parsed data is added
#' as the `form` element of the request object.
#'
#' @param type Content type to match before parsing. If it does not
#'   match, then the request object is not modified.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_urlencoded())
#' app

mw_urlencoded <- function(type = "application/x-www-form-urlencoded") {
  function(req, res) {
    ct <- req$get_header("Content-Type") %||% ""
    if (! ct %in% tolower(type)) return("next")
    if (!is.null(req$.body)) {
      req$form <- parse_query(rawToChar(req$.body))
    }
    "next"
  }
}

#' Web app that acts as a git http server
#'
#' It is useful for tests that need an HTTP git server.
#'
#' @param git_root Path to the root of the directory tree to be served.
#' @param git_cmd Command to call, by default it is `"git"`. It may also
#'   be a full path to git.
#' @param git_timeout A `difftime` object, time limit for the git
#'   command.
#' @param filter Whether to support the `filter` capability in the server.
#' @param cleanup Whether to clean up `git_root` when the app is
#'   garbage collected.
#'
#' @export
#' @examplesIf FALSE
#' dir.create(tmp <- tempfile())
#' setwd(tmp)
#' system("git clone --bare https://github.com/cran/crayon")
#' system("git clone --bare https://github.com/cran/glue")
#' app <- git_app(tmp)
#' git <- new_app_process(app)
#' system(paste("git ls-remote", git$url("/crayon")))

git_app <- function(git_root,
                    git_cmd = "git",
                    git_timeout = as.difftime(1, units = "mins"),
                    filter = TRUE,
                    cleanup = TRUE) {

  app <- webfakes::new_app()
  app$locals$git_root <- git_root
  app$locals$git_timeout <- as.double(git_timeout, units = "secs") * 1000
  app$locals$git_config <- tempfile()

  reg.finalizer(app, function(app0) unlink(app$locals$git_config), TRUE)
  writeLines(
    c(
      "[uploadpack]",
      paste0("\tallowFilter = ", if (isTRUE(filter)) "true" else "false")
    ),
    app$locals$git_config
  )

  if (cleanup) {
    reg.finalizer(
      app,
      function(app) unlink(app$locals$git_root, recursive = TRUE),
      TRUE
    )
  }

  cgi <- mw_cgi(git_cmd, "http-backend", timeout = git_timeout)

  handler <- function(req, res) {
    env <- c(
      GIT_CONFIG_GLOBAL = req$app$locals$git_config,
      GIT_HTTP_EXPORT_ALL = "true",
      GIT_PROJECT_ROOT = req$app$locals$git_root,
      GIT_PROTOCOL = req$get_header("Git-Protocol") %||% "",
      HTTP_GIT_PROTOCOL = req$get_header("Git-Protocol") %||% ""
    )
    cgi(req, res, env = env)
  }

  re_all <- new_regexp("^(?<path>.*)$")

  app$get(re_all, handler)
  app$post(re_all, handler)

  app
}

#' Middleware to read the raw body of a request
#'
#' Adds the raw body, as a raw object to the `raw` field of the request.
#'
#' @param type Content type to match. If it does not match, then the
#' request object is not modified.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_raw())
#' app

mw_raw <- function(type = "application/octet-stream") {
  function(req, res) {
    ct <- req$get_header("Content-Type") %||% ""
    if (!ct %in% tolower(type)) {
      return("next")
    }
    req$raw <- req$.body
    "next"
  }
}

#' Middleware to parse Cookies
#'
#' Adds the cookies as the `cookies` element of the request object.
#'
#' It ignores cookies in an invalid format. It ignores duplicate cookies:
#' if two cookies have the same name, only the first one is included.
#'
#' @return Handler function.
#'
#' @family middleware
#' @export

mw_cookie_parser <- function() {
  function(req, res) {
    ch <- req$get_header("Cookie") %||% ""
    req$cookies <- parse_cookies(ch)
    "next"
  }
}

parse_cookies <- function(x) {
  parts <- strsplit(x, ";", fixed = TRUE)[[1]]
  dict <- structure(list(), names = character())
  lapply(parts, function(ck) {
    ck <- trimws(ck)
    key <- sub("^([^=]+)=.*$", "\\1", ck)
    if (key == ck) return()
    if (!is.null(dict[[key]])) return()
    value <- sub("^[^=]+=", "", ck)
    dict[[key]] <<- value
  })
  dict
}

format_named_list <- function(name, data) {
  c(paste0(name, ":"), if (length(data)) paste0("  ", names(data), ": ", data))
}

format_path <- function(patterns) {
  # Make sure patterns is a list
  if (inherits(patterns, "webfakes_regexp")) {
    patterns <- list(patterns)
  } else if (is.character(patterns)) {
    patterns <- as.list(patterns)
  }
  paste(collapse = ", ", vapply(patterns, format, character(1)))
}

#' @export

format.webfakes_app <- function(x, ...) {
  header <- "<webfakes_app>"
  data <- vapply(x$.stack, FUN.VALUE = character(1), function(x) {
    name <- if (nchar(x$name %||% "")) paste0(" # ", x$name)
    paste0("  ", x$method, " ", format_path(x$path), name)
  })
  methods <- c(
    "  all(path, ...)         # add route for *all* HTTP methods",
    "  delete(path, ...)      # add route for DELETE",
    "  engine(ext, engine)    # add template engine for file extension",
    "  head(path, ...)        # add route for HEAD",
    "  listen(port)           # start web app on port",
    "  patch(path, ...)       # add route for PATCH",
    "  post(path, ...)        # add route for POST",
    "  put(path, ...)         # add route for PUT",
    "  use(...)               # add middleware",
    "  locals                 # app-wide shared data"
  )
  help <- "# see ?webfakes_app for all methods"
  c(header, "routes:", data, "fields and methods:", methods, help)
}

#' @export

print.webfakes_app <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export

format.webfakes_request <- function(x, ...) {
  header <- "<webfakes_request>"
  data <- c(
    "method:",
    paste0("  ", x$method),
    "url:",
    paste0("  ", x$url),
    "client:",
    paste0("  ", x$remote_addr),
    format_named_list("query", x$query),
    format_named_list("headers", x$headers)
  )
  methods <- c(
    "  app                    # the webfakes_app the request belongs to",
    "  headers                # HTTP request headers",
    "  hostname               # server hostname, the Host header",
    "  method                 # HTTP method of request (lowercase)",
    "  path                   # server path",
    "  protocol               # http or https",
    "  query_string           # raw query string without '?'",
    "  query                  # named list of query parameters",
    "  remote_addr            # IP address of the client",
    "  url                    # full URL of the request",
    "  get_header(field)      # get a request header"
  )
  help <- " # see ?webfakes_request for details"
  c(header, data, "fields and methods:", methods, help)
}

#' @export

print.webfakes_request <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export

format.webfakes_response <- function(x, ...) {
  header <- "<webfakes_reeponse>"
  methods <- c(
    "  app                    # the webfakes_app the response belongs to",
    "  locals                 # response-wide shared data",
    "  get_header(field)      # query response header",
    "  on_response(fun)       # call handler function for complete response",
    "  redirect(path, status) # send redirect response",
    "  render(view, locals)   # render template",
    "  send(body)             # send text or raw data",
    "  send_file(path, root)  # send a file (automatic Content-Type)",
    "  send_json(object, text, ...)",
    "                         # send JSON data",
    "  send_status(status)    # send HTTP status and empty body",
    "  set_header(field, value)
                              # set a response header",
    "  set_status(status)     # set response status code",
    "  set_type(type)         # set Content-Type"
  )
  help <- " # see ?webfakes_response for details"
  c(header, "fields and methods:", methods, help)
}

#' @export

print.webfakes_response <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export

format.webfakes_regexp <- function(x, ...) {
  paste0("<webfakes_regexp> ", encodeString(x, quote = "\""))
}

#' @export

print.webfakes_regexp <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' @export

format.webfakes_app_process <- function(x, ...) {
  header <- "<webfakes_app_process>"
  state <- x$get_state()
  data <- c(
    "state:",
    paste0("  ", state),
    "auto_start:",
    paste0("  ", x$.auto_start),
    "process id:",
    paste0("  ", if (state == "not running") "none" else x$.process$get_pid()),
    "http url:",
    paste0("  ", if (state == "live") x$url() else "NA")
  )
  methods <- c(
    "  get_app()              # get the app object",
    "  get_port()             # query (first) port of the app",
    "  get_ports()            # query all ports of the app",
    "  get_state()            # query web server process state",
    "  local_env(envvars)     # set temporary environment variables",
    "  start()                # start the app",
    "  url(path, query)       # query url for an api path",
    "  stop()                 # stop web server process"
  )
  help <- "# see ?webfakes_app_process for details"
  c(header, data, "fields and methods:", methods, help)
}

#' @export

print.webfakes_app_process <- function(x, ...) {
  cat(format(x, ...), sep = "\n")
  invisible(x)
}

#' Middleware to parse a JSON body
#'
#' Adds the parsed object as the `json` element of the request object.
#'
#' @param type Content type to match before parsing. If it does not
#'   match, then the request object is not modified.
#' @param simplifyVector Whether to simplify lists to vectors, passed to
#'   [jsonlite::fromJSON()].
#' @param ... Arguments to pass to [jsonlite::fromJSON()], that performs
#'   the JSON parsing.
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_json())
#' app

mw_json <- function(type = "application/json", simplifyVector = FALSE, ...) {
  type
  simplifyVector
  list(...)
  function(req, res) {
    ct <- req$get_header("Content-Type") %||% ""
    if (!ct %in% tolower(type)) {
      return("next")
    }
    if (!is.null(req$.body)) {
      req$json <- jsonlite::fromJSON(
        rawToChar(req$.body),
        simplifyVector = simplifyVector,
        ...
      )
    }
    "next"
  }
}

#' A webfakes response object
#'
#' webfakes creates a `webfakes_response` object for every incoming HTTP
#' request. This object is passed to every matched route and middleware,
#' until the HTTP response is sent. It has reference semantics, so handlers
#' can modify it.
#'
#' Fields and methods:
#'
#' * `app`: The `webfakes_app` object itself.
#' * `req`: The request object.
#' * `headers_sent`: Whether the response headers were already sent out.
#' * `locals`: Local variables, the are shared between the handler
#'   functions. This is for the end user, and not for the middlewares.
#' * `delay(secs)`: delay the response for a number of seconds. If a
#'   handler calls `delay()`, the same handler will be called again,
#'   after the specified number of seconds have passed. Use the `locals`
#'   environment to distinguish between the calls. If you are using
#'   `delay()`, and want to serve requests in parallel, then you probably
#'   need a multi-threaded server, see [server_opts()].
#' * `add_header(field, value)`: Add a response header. Note that
#'   `add_header()` may create duplicate headers. You usually want
#'   `set_header()`.
#' * `get_header(field)`: Query the currently set response headers. If
#'   `field` is not present it return `NULL`.
#' * `on_response(fun)`: Run the `fun` handler function just before the
#'   response is sent out. At this point the headers and the body are
#'   already properly set.
#' * `redirect(path, status = 302)`: Send a redirect response. It sets
#'   the `Location` header, and also sends a `text/plain` body.
#' * `render(view, locals = list())`: Render a template page. Searches
#'   for the `view` template page, using all registered engine extensions,
#'   and calls the first matching template engine. Returns the filled
#'   template.
#' * `send(body)`. Send the specified body. `body` can be a raw vector,
#'   or HTML or other text. For raw vectors it sets the content type to
#'   `application/octet-stream`.
#' * `send_json(object = NULL, text = NULL, ...)`: Send a JSON response.
#'   Either `object` or `text` must be given. `object` will be converted
#'   to JSON using [jsonlite::toJSON()]. `...` are passed to
#'   [jsonlite::toJSON()]. It sets the content type appropriately.
#' * `send_file(path, root = ".")`: Send a file. Set `root = "/"` for
#'   absolute file names. It sets the content type automatically, based
#'   on the extension of the file, if it is not set already.
#' * `send_status(status)`: Send the specified HTTP status code, without
#'   a response body.
#' * `send_chunk(data)`: Send a chunk of a response in chunked encoding.
#'   The first chunk will automatically send the HTTP response headers.
#'   Webfakes will automatically send a final zero-lengh chunk, unless
#'   `$delay()` is called.
#' * `set_header(field, value)`: Set a response header. If the headers have
#'   been sent out already, then it throws a warning, and does nothing.
#' * `set_status(status)`: Set the response status code. If the headers
#'   have been sent out already, then it throws a warning, and does nothing.
#' * `set_type(type)`: Set the response content type. If it contains a `/`
#'   character then it is set as is, otherwise it is assumed to be a file
#'   extension, and the corresponding MIME type is set. If the headers have
#'   been sent out already, then it throws a warning, and does nothing.
#' * `add_cookie(name, value, options)`: Adds a cookie to the response.
#'   `options` is a named list, and may contain:
#'    * `domain`: Domain name for the cookie, not set by default.
#'    * `expires`: Expiry date in GMT. It must be a POSIXct object, and
#'       will be formatted correctly.
#'    * 'http_only': if TRUE, then it sets the 'HttpOnly' attribute, so
#'      Javasctipt cannot access the cookie.
#'    * `max_age`: Maximum age, in number of seconds.
#'    * `path`: Path for the cookie, defaults to "/".
#'    * `same_site`: The 'SameSite' cookie attribute. Possible values are
#'      "strict", "lax" and "none".
#'    * `secure`: if TRUE, then it sets the 'Secure' attribute.
#' * `clear_cookie(name, options = list())`: clears a cookie. Typically,
#'    web browsers will only clear a cookie if the options also match.
#' * `write(data)`: writes (part of) the body of the response. It also
#'   sends out the response headers, if they haven't been sent out before.
#'
#' Usually you need one of the `send()` methods, to send out the HTTP
#' response in one go, first the headers, then the body.
#'
#' Alternatively, you can use `$write()` to send the response in parts.
#'
#' @seealso [webfakes_request] for the webfakes request object.
#' @name webfakes_response
#' @examples
#' # This is how you can see the request and response objects:
#' app <- new_app()
#' app$get("/", function(req, res) {
#'   browser()
#'   res$send("done")
#' })
#' app
#'
#' # Now start this app on a port:
#' # app$listen(3000)
#' # and connect to it from a web browser: http://127.0.0.1:3000
#' # You can also use another R session to connect:
#' # httr::GET("http://127.0.0.1:3000")
#' # or the command line curl tool:
#' # curl -v http://127.0.0.1:3000
#' # The app will stop while processing the request.
NULL

new_response <- function(app, req) {
  self <- new_object(
    "webfakes_response",

    app = app,
    req = req,
    locals = as.environment(as.list(app$locals)),
    headers_sent = FALSE,

    delay = function(secs) {
      self$.stackptr <- self$.i
      self$.delay <- secs
      response_delay(self$req, secs)
      invisible(NULL)
    },

    get_header = function(field) {
      # this is case insensitive
      h <- self$.headers
      names(h) <- tolower(names(h))
      h[[tolower(field)]]
    },

    on_response = function(fun) {
      self$.on_response <- c(self$.on_response, list(fun))
      invisible(self)
    },

    redirect = function(path, status = 302) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      self$set_status(status)$set_header("Location", path)$set_type(
        "text/plain"
      )$send(paste0(
        status,
        " ",
        http_statuses[as.character(status)],
        ". Redirecting to ",
        path
      ))
      invisible(self)
    },

    render = function(view, locals = list()) {
      locals <- as.environment(as.list(locals))
      parent.env(locals) <- self$locals
      root <- self$app$get_config("views")
      for (eng in self$app$.engines) {
        f <- file.path(root, paste0(view, ".", eng$ext))
        if (file.exists(f)) return(eng$engine(f, locals))
      }
      stop("Cannot find template engine for view '", view, "'")
    },

    send = function(body) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      # We need to do these here, on_response middleware might depend on it
      self$.body <- body
      self$.set_defaults()
      for (fn in self$.on_response) {
        fn(self$req, self)
      }

      response_send(self$req)

      self$headers_sent <- TRUE
      self$.sent <- TRUE
      invisible(self)
    },

    send_chunk = function(data) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      # The first chunk sends the headers automatically, but we make
      # sure to set chunked encoding
      if (!self$headers_sent) {
        self$set_header("Transfer-Encoding", "chunked")
        if (is.null(self$get_header("Content-Type"))) {
          self$set_header("Content-Type", "application/octet-stream")
        }
        if (is.null(self$.status)) {
          self$set_status(200L)
        }
        self$.set_defaults()
      }
      enc <- self$get_header("Transfer-Encoding")
      if (enc != "chunked") {
        warning("Headers sent, cannot set chunked encoding now")
        return(invisible(self))
      }
      if (is.character(data)) {
        data <- charToRaw(paste(data, collapse = "\n"))
      }
      response_send_chunk(self$req, data)
      self$headers_sent <- TRUE
      invisible(self)
    },

    send_json = function(object = NULL, text = NULL, ...) {
      if (!is.null(object) && !is.null(text)) {
        stop("Specify only one of `object` and `text` in `send_json()`")
      }

      if (is.null(text)) {
        text <- jsonlite::toJSON(object, ...)
      }

      self$set_header("Content-Type", "application/json")$send(text)
    },

    send_file = function(path, root = ".") {
      # Set content type automatically
      if (is.null(self$get_header("Content-Type"))) {
        ext <- tools::file_ext(basename(path))
        ct <- mime_find(ext)
        if (!is.na(ct)) {
          self$set_header("Content-Type", ct)
        }
      }

      if (
        root == "/" &&
          .Platform$OS.type == "windows" &&
          grepl("^[a-zA-Z]:", path)
      ) {
        abs_path <- path
      } else {
        abs_path <- file.path(root, path)
      }

      self$send(read_bin(normalizePath(abs_path)))
    },

    send_status = function(status) {
      self$set_status(status)$send("")
    },

    set_header = function(field, value) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      self$.headers[[field]] <- as.character(value)
      invisible(self)
    },

    add_header = function(field, value) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      h <- structure(list(value), names = field)
      self$.headers <- append(self$.headers, h)
      invisible(self)
    },

    set_status = function(status) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      self$.status <- as.integer(status)
      invisible(self)
    },

    set_type = function(type) {
      if (self$.check_sent()) {
        return(invisible(self))
      }
      if (grepl("/", type)) {
        self$set_header("Content-Type", type)
      } else {
        ct <- mime_find(type)
        if (!is.na(ct)) {
          self$set_header("Content-Type", ct)
        }
      }
      invisible(self)
    },

    add_cookie = function(name, value, options = list()) {
      if (!is_string(name)) {
        stop("Cookie name must be a string.")
      }
      if (grepl("[=;]", name)) {
        stop("Cookie name cannot contain ';' and '=' characters.")
      }
      if (!is_string(value)) {
        stop("Cookie value must be a string.")
      }
      if (grepl("[=;]", value)) {
        stop("Cookie value cannot contain ';' and '=' characters.")
      }

      ck <- paste0(
        name,
        "=",
        value,
        "; ",
        format_cookie_options(options)
      )
      self$add_header("Set-Cookie", ck)
      invisible(self)
    },

    clear_cookie = function(name, options = list()) {
      if (!is_string(name)) {
        stop("Cookie name must be a string.")
      }
      if (grepl("[=;]", name)) {
        stop("Cookie name cannot contain ';' and '=' characters.")
      }

      options$expires <- .POSIXct(0)
      options$max_age <- 0L
      ck <- paste0(name, "=; ", format_cookie_options(options))
      self$add_header("Set-Cookie", ck)
      invisible(self)
    },

    write = function(data) {
      if (is.null(self$get_header("content-length"))) {
        warning("response$write() without a Content-Length header")
      }
      if (is.null(self$.status)) {
        self$set_status(200L)
      }
      if (is.character(data)) {
        data <- charToRaw(paste(data, collapse = "\n"))
      }
      response_write(self$req, data)
      self$headers_sent <- TRUE
      invisible(self)
    },

    .check_sent = function() {
      if (isTRUE(self$.sent)) {
        warning("Response is sent already")
      }
      self$.sent
    },

    .set_defaults = function() {
      if (is.null(self$.status)) {
        if (is.null(self$.body)) {
          # No status, no body, that's 404
          self$.status <- 404L
          self$.body <- "Not found"
        } else {
          # No status, but body, set status
          self$.status <- 200L
        }
      }

      # Set Content-Type if not set
      if (is.null(self$get_header("Content-Type"))) {
        if (is.raw(self$body)) {
          ct <- "application/octet-stream"
        } else {
          ct <- "text/plain"
        }
        self$set_header("Content-Type", ct)
      }

      # Set Content-Length if not set
      if (
        is.null(self$get_header("Content-Length")) &&
          (self$get_header("Transfer-Encoding") %||% "") != "chunked"
      ) {
        if (is.raw(self$.body)) {
          cl <- length(self$.body)
        } else if (is.character(self$.body)) {
          cl <- nchar(self$.body, type = "bytes")
        } else if (is.null(self$.body)) {
          cl <- 0L
        }
        self$set_header("Content-Length", cl)
      }

      # Make sure response to HEAD has empty body
      if (self$req$method == "head") {
        self$.body <- raw(0)
        self$set_header("Content-Length", "0")
      }
    },

    .body = NULL,
    .status = NULL,
    .headers = if (!app$.enable_keep_alive) {
      list("Connection" = "close")
    } else {
      list()
    },
    .on_response = NULL,
    .sent = FALSE,
    .stackptr = 1L
  )

  self
}

format_cookie_options <- function(options) {
  options$path <- options$path %||% "/"

  bad <- unique(setdiff(
    names(options),
    c(
      "domain",
      "expires",
      "http_only",
      "max_age",
      "path",
      "same_site",
      "secure"
    )
  ))
  if (length(bad)) {
    stop(
      "Unknown or unsupported cookie attribute(s): ",
      paste0("\"", bad, "\"", collapse = ", "),
      "."
    )
  }

  parts <- c(
    if (!is.null(options$domain)) {
      paste0("Domain=", options$domain)
    },

    if (!is.null(options$expires)) {
      if (!inherits(options$expires, "POSIXct")) {
        stop("The 'expires' cookie attribute must be a POSIXct object")
      }
      paste0("Expires=", http_time_stamp(options$expires))
    },

    if (isTRUE(options$http_only)) {
      "HttpOnly"
    },

    if (!is.null(options$max_age)) {
      paste0("Max-Age=", options$max_age)
    },

    paste0("Path=", options$path),

    if (!is.null(options$same_site)) {
      if (tolower(!options$same_site) %in% c("strict", "lax", "none")) {
        stop(
          "Invalid value for 'SameSite' cookie atrribute: ",
          options$same_site,
          ", must be \"strict\", \"lax\" or \"none\"."
        )
      }
      paste0("SameSite=", capitalize(options$same_site))
    },

    if (isTRUE(options$secure)) {
      "Secure"
    }
  )

  paste(parts, collapse = "; ")
}

app <- new_app()
app$get("/delay", function(req, res) {
  if (is.null(res$locals$seen)) {
    res$locals$seen <- TRUE
    res$delay(.5)
  } else {
    res$send_json(
      list(message = "Sorry, running late..."),
      auto_unbox = TRUE
    )
  }
})
app$get("/nodelay", function(req, res) {
  res$send_json(
    list(message = "I am fast, aren't I?"),
    auto_unbox = TRUE
  )
})

web <- local_app_process(app, opts = server_opts(num_threads = 2))

 test_that("delay", {
  p <- curl::new_pool(multiplex = FALSE)
  h1 <- curl::new_handle(url = web$url("/delay"))
  h2 <- curl::new_handle(url = web$url("/nodelay"))
  resp1 <- resp2 <- NULL
  curl::multi_add(h1, done = function(x) resp1 <<- x, fail = stop, pool = p)
  curl::multi_add(h2, done = function(x) resp2 <<- x, fail = stop, pool = p)
  curl::multi_run(timeout = 2, pool = p)
  curl::multi_cancel(h1)
  curl::multi_cancel(h2)

  expect_false(is.null(resp1))
  expect_false(is.null(resp2))
  expect_true(resp1$times[["total"]] > 0.5)
  expect_true(resp2$times[["total"]] < resp1$times[["total"]])
})

# nocov start --- compat-defer --- 2020-06-16

# This drop-in file implements withr::defer(). Please find the most
# recent version in withr's repository.

defer <- function(
  expr,
  envir = parent.frame(),
  priority = c("first", "last")
) {}

local({
  defer <<- defer <- function(
    expr,
    envir = parent.frame(),
    priority = c("first", "last")
  ) {
    priority <- match.arg(priority)
    if (identical(envir, .GlobalEnv) && is.null(get_handlers(envir))) {
      message(
        "Setting deferred event(s) on global environment.\n",
        "  * Execute (and clear) with `withr::deferred_run()`.\n",
        "  * Clear (without executing) with `withr::deferred_clear()`."
      )
    }
    invisible(
      add_handler(
        envir,
        handler = list(expr = substitute(expr), envir = parent.frame()),
        front = priority == "first"
      )
    )
  }

  get_handlers <- function(envir) {
    attr(envir, "handlers")
  }

  set_handlers <- function(envir, handlers) {
    has_handlers <- "handlers" %in% names(attributes(envir))
    attr(envir, "handlers") <- handlers
    if (!has_handlers) {
      call <- make_call(execute_handlers, envir)

      # We have to use do.call here instead of eval because of the way on.exit
      # determines its evaluation context
      # (https://stat.ethz.ch/pipermail/r-devel/2013-November/067867.html)
      do.call(base::on.exit, list(call, TRUE), envir = envir)
    }
  }

  execute_handlers <- function(envir) {
    handlers <- get_handlers(envir)
    errors <- list()
    for (handler in handlers) {
      tryCatch(eval(handler$expr, handler$envir), error = function(e) {
        errors[[length(errors) + 1]] <<- e
      })
    }

    for (error in errors) {
      stop(error)
    }
  }

  add_handler <- function(envir, handler, front) {
    if (front) {
      handlers <- c(list(handler), get_handlers(envir))
    } else {
      handlers <- c(get_handlers(envir), list(handler))
    }

    set_handlers(envir, handlers)
    handler
  }

  make_call <- function(...) {
    as.call(list(...))
  }
}) # defer() namespace

# nocov end

pkg_data <- new.env(parent = emptyenv())

#' Create a new web application
#'
#' @details
#' The typical workflow of creating a web application is:
#'
#' 1. Create a `webfakes_app` object with `new_app()`.
#' 1. Add middleware and/or routes to it.
#' 1. Start is with the `webfakes_app$listen()` method, or start it in
#'    another process with [new_app_process()].
#' 1. Make queries to the web app.
#' 1. Stop it via `CTRL+C` / `ESC`, or, if it is running in another
#'    process, with the `$stop()` method of [new_app_process()].
#'
#' A web application can be
#' * restarted,
#' * saved to disk,
#' * copied to another process using the callr package, or a similar way,
#' * embedded into a package,
#' * extended by simply adding new routes and/or middleware.
#'
#' The webfakes API is very much influenced by the
#' [express.js](https://expressjs.com/) project.
#'
#' ## Create web app objects
#'
#' ```r
#' new_app()
#' ```
#'
#' `new_app()` returns a `webfakes_app` object the has the methods listed
#' on this page.
#'
#' An app is an environment with S3 class `webfakes_app`.
#'
#' ## The handler stack
#'
#' An app has a stack of handlers. Each handler can be a route or
#' middleware. The differences between the two are:
#' * A route is bound to one or more paths on the web server. Middleware
#'   is not (currently) bound to paths, but run for all paths.
#' * A route is usually (but not always) the end of the handler stack for
#'   a request. I.e. a route takes care of sending out the response to
#'   the request. Middleware typically performs some action on the request
#'   or the response, and then the next handler in the stack is invoked.
#'
#' ## Routes
#'
#' The following methods define routes. Each method corresponds to the
#' HTTP verb with the same name, except for `app$all()`, which creates a
#' route for all HTTP methods.
#'
#' ```r
#' app$all(path, ...)
#' app$delete(path, ...)
#' app$get(path, ...)
#' app$head(path, ...)
#' app$patch(path, ...)
#' app$post(path, ...)
#' app$put(path, ...)
#' ... (see list below)
#' ```
#'
#' * `path` is a path specification, see 'Path specification' below.
#' * `...` is one or more handler functions. These will be placed in the
#'   handler stack, and called if they match an incoming HTTP request.
#'   See 'Handler functions' below.
#'
#' webfakes also has methods for the less frequently used HTTP verbs:
#' `CONNECT`, `MKCOL`, `OPTIONS`, `PROPFIND`, `REPORT`. (The method
#' names are always in lowercase.)
#'
#' If a request is not handled by any routes (or handler functions in
#' general), then webfakes will send a simple HTTP 404 response.
#'
#' ## Middleware
#'
#' `app$use()` adds a middleware to the handler stack. A middleware is
#' a handler function, see 'Handler functions' below. webfakes comes with
#' middleware to perform common tasks:
#'
#' * [mw_cookie_parser()] parses `Cookie` headers.
#' * [mw_etag()] adds an `ETag` header to the response.
#' * [mw_json()] parses JSON request bodies.
#' * [mw_log()] logs each requests to standard output, or another connection.
#' * [mw_multipart()] parses multipart request bodies.
#' * [mw_range_parser()] parses `Range` headers.
#' * [mw_raw()] parses raw request bodies.
#' * [mw_static()] serves static files from a directory.
#' * [mw_text()] parses plain text request bodies.
#' * [mw_urlencoded()] parses URL encoded request bodies.
#'
#' ```r
#' app$use(..., .first = FALSE)
#' ```
#'
#' * `...` is a set of (middleware) handler functions. They are added to
#' the handler stack, and called for every HTTP request. (Unless an HTTP
#' response is created before reaching this point in the handler stack.)
#' * `.first` set to `TRUE` is you want to add the handler function
#' to the bottom of the stack.
#'
#' ## Handler functions
#'
#' A handler function is a route or middleware. A handler function is
#' called by webfakes with the incoming HTTP request and the outgoing
#' HTTP response objects (being built) as arguments. The handler function
#' may query and modify the members of the request and/or the response
#' object. If it returns the string `"next"`, then it is _not_ a terminal
#' handler, and once it returns, webfakes will move on to call the next
#' handler in the stack.
#'
#' A typical route:
#'
#' ```r
#' app$get("/user/:id", function(req, res) {
#'   id <- req$params$id
#'   ...
#'   res$
#'     set_status(200L)$
#'     set_header("X-Custom-Header", "foobar")$
#'     send_json(response, auto_unbox = TRUE)
#' })
#' ```
#'
#' * The handler belongs to an API path, which is a wildcard path in
#'   this case. It matches `/user/alice`, `/user/bob`, etc. The handler
#'   will be only called for GET methods and matching API paths.
#' * The handler receives the request (`req`) and the response (`res`).
#' * It sets the HTTP status, additional headers, and sends the data.
#'   (In this case the `webfakes_response$send_json()` method automatically
#'   converts `response` to JSON and sets the `Content-Type` and
#'   `Content-Length` headers.
#' * This is a terminal handler, because it does _not_ return `"next"`.
#'   Once this handler function returns, webfakes will send out the HTTP
#'   response.
#'
#' A typical middleware:
#'
#' ```r
#' app$use(function(req, res) {
#'   ...
#'   "next"
#' })
#' ````
#'
#' * There is no HTTP method and API path here, webfakes will call the
#'   handler for each HTTP request.
#' * This is not a terminal handler, it does return `"next"`, so after it
#'   returns webfakes will look for the next handler in the stack.
#'
#' ## Errors
#'
#' If a handler function throws an error, then the web server will return
#' a HTTP 500 `text/plain` response, with the error message as the
#' response body.
#'
#' ## Request and response objects
#'
#' See [webfakes_request] and [webfakes_response] for the methods of the
#' request and response objects.
#'
#' ## Path specification
#'
#' Routes are associated with one or more API paths. A path specification
#' can be
#'
#' * A "plain" (i.e. without parameters) string. (E.g. `"/list"`.)
#' * A parameterized string. (E.g. `"/user/:id"`.)
#' * A regular expression created via [new_regexp()] function.
#' * A list or character vector of the previous ones. (Regular expressions
#'   must be in a list.)
#'
#' ## Path parameters
#'
#' Paths that are specified as parameterized strings or regular expressions
#' can have parameters.
#'
#' For parameterized strings the keys may contain letters, numbers and
#' underscores. When webfakes matches an API path to a handler with a
#' parameterized string path, the parameters will be added to the
#' request, as `params`. I.e. in the handler function (and subsequent
#' handler functions, if the current one is not terminal), they are
#' available in the `req$params` list.
#'
#' For regular expressions, capture groups are also added as parameters.
#' It is best to use named capture groups, so that the parameters are in
#' a named list.
#'
#' If the path of the handler is a list of parameterized strings or
#' regular expressions, the parameters are set according to the first
#' matching one.
#'
#' ## Templates
#'
#' webfakes supports templates, using any template engine. It comes with
#' a template engine that uses the glue package, see [tmpl_glue()].
#'
#' `app$engine()` registers a template engine, for a certain file
#' extension. The `$render()` method of [webfakes_response]
#' can be called from the handler function to evaluate a template from a
#' file.
#'
#' ```r
#' app$engine(ext, engine)
#' ```
#'
#' * `ext`: the file extension for which the template engine is added.
#'   It should not contain the dot. E.g. `"html"', `"brew"`.
#' * `engine`: the template engine, a function that takes the file path
#'   (`path`) of the template, and a list of local variables (`locals`)
#'   that can be used in the template. It should return the result.
#'
#' An example template engine that uses glue might look like this:
#'
#' ```r
#' app$engine("txt", function(path, locals) {
#'   txt <- readChar(path, nchars = file.size(path))
#'   glue::glue_data(locals, txt)
#' })
#' ```
#'
#' (The built-in [tmpl_glue()] engine has more features.)
#'
#' This template engine can be used in a handler:
#'
#' ```r
#' app$get("/view", function(req, res) {
#'  txt <- res$render("test")
#'  res$
#'    set_type("text/plain")$
#'    send(txt)
#' })
#' ```
#'
#' The location of the templates can be set using the `views` configuration
#' parameter, see the `$set_config()` method below.
#'
#' In the template, the variables passed in as `locals`, and also the
#' response local variables (see `locals` in [webfakes_response]), are
#' available.
#'
#' ## Starting and stopping
#'
#' ```r
#' app$listen(port = NULL, opts = server_opts(), cleanup = TRUE)
#' ```
#'
#' * `port`: port to listen on. When `NULL`, the operating system will
#'   automatically select a free port. Add an `"s"` suffix to the port
#'   to use HTTPS. Use `"0s"` to use an OS assigned port with HTTPS.
#'   See the [how-to] manual page if you want to start the web server
#'   on more than one ports.
#'
#' * `opts`: options to the web server. See [server_opts()] for the
#'   list of options and their default values.
#'
#' * `cleanup`: stop the server (with an error) if the standard input
#'   of the process is closed. This is handy when the app runs in a
#'   `callr::r_session` subprocess, because it stops the app (and the
#'   subprocess) if the main process has terminated.
#'
#' This method does not return, and can be interrupted with `CTRL+C` / `ESC`
#' or a SIGINT signal. See [new_app_process()] for interrupting an app that
#' is running in another process.
#'
#' When `port` is `NULL`, the operating system chooses a port where the
#' app will listen. To be able to get the port number programmatically,
#' before the listen method blocks, it advertises the selected port in a
#' `webfakes_port` condition, so one can catch it:
#'
#' webfakes by default binds only to the loopback interface at 127.0.0.1, so
#' the webfakes web app is never reachable from the network.
#'
#' ```r
#' withCallingHandlers(
#'   app$listen(),
#'   "webfakes_port" = function(msg) print(msg$port)
#' )
#' ```
#'
#' ## Logging
#'
#' webfakes can write an access log that contains an entry for all incoming
#' requests, and also an error log for the errors that happen while
#' the server is running. This is the default behavior for local app
#' (the ones started by `app$listen()` and for remote apps (the ones
#' started via `new_app_process()`:
#'
#' * Local apps do not write an access log by default.
#' * Remote apps write an access log into the
#'   `<tmpdir>/webfakes/<pid>/access.log` file, where `<tmpdir>` is the
#'   session temporary directory of the _main process_, and `<pid>` is
#'   the process id of the _subprocess_.
#' * Local apps write an error log to `<tmpdir>/webfakes/error.log`, where
#'   `<tmpdir>` is the session temporary directory of the current process.
#' * Remote app write an error log to the `<tmpdir>/webfakes/<pid>/error.log`,
#'   where `<tmpdir>` is the session temporary directory of the
#'   _main process_ and `<pid>` is the process id of the _subprocess_`.
#'
#' See [server_opts()] for changing the default logging behavior.
#'
#' ## Shared app data
#'
#' ```r
#' app$locals
#' ```
#'
#' It is often useful to share data between handlers and requests in an
#' app. `app$locals` is an environment that supports this. E.g. a
#' middleware that counts the number of requests can be implemented as:
#'
#' ```
#' app$use(function(req, res) {
#'   locals <- req$app$locals
#'   if (is.null(locals$num)) locals$num <- 0L
#'   locals$num <- locals$num + 1L
#'   "next"
#' })
#' ```
#'
#' [webfakes_response] objects also have a `locals` environment, that is
#' initially populated as a copy of `app$locals`.
#'
#' ## Configuration
#'
#' ```r
#' app$get_config(key)
#' app$set_config(key, value)
#' ```
#'
#' * `key`: configuration key.
#' * `value`: configuration value.
#'
#' Currently used configuration values:
#'
#' * `views`: path where webfakes searches for templates.
#'
#' @return A new `webfakes_app`.
#' @aliases webfakes_app
#' @seealso [webfakes_request] for request objects, [webfakes_response] for
#' response objects.
#' @export
#' @examples
#' # see example web apps in the `/examples` directory in
#' system.file(package = "webfakes", "examples")
#'
#' app <- new_app()
#' app$use(mw_log())
#'
#' app$get("/hello", function(req, res) {
#'   res$send("Hello there!")
#' })
#'
#' app$get(new_regexp("^/hi(/.*)?$"), function(req, res) {
#'   res$send("Hi indeed!")
#' })
#'
#' app$post("/hello", function(req, res) {
#'   res$send("Got it, thanks!")
#' })
#'
#' app
#'
#' # Start the app with: app$listen()
#' # Or start it in another R session: new_app_process(app)

new_app <- function() {
  self <- new_object(
    "webfakes_app",

    all = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("all", path, ...))
      invisible(self)
    },

    connect = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("connect", path, ...))
      invisible(self)
    },

    delete = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("delete", path, ...))
      invisible(self)
    },

    get = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("get", path, ...))
      invisible(self)
    },

    engine = function(ext, engine) {
      rec <- list(ext = ext, engine = engine)
      self$.engines <- c(self$.engines, list(rec))
      invisible(self)
    },

    get_config = function(key) {
      self$.config[[key]]
    },

    head = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("head", path, ...))
      invisible(self)
    },

    listen = function(port = NULL, opts = server_opts(), cleanup = TRUE) {
      if (is_na_scalar(port)) {
        port <- NULL
      }
      opts$port <- port
      self$.enable_keep_alive <- opts$enable_keep_alive
      opts$access_log_file <- sub("%p", Sys.getpid(), opts$access_log_file)
      opts$error_log_file <- sub("%p", Sys.getpid(), opts$error_log_file)
      self$.opts <- opts

      tryCatch(srv <- server_start(opts), error = function(err) {
        err$message <- paste(sep = "\n", err$message, self$.get_error_log())
        stop(err)
      })
      self$.ports <- server_get_ports(srv)
      self$.port <- self$.ports$port[1]
      port_nums <- paste0(
        self$.ports$port,
        ifelse(self$.ports$ssl, " (SSL)", "")
      )
      message(
        "Running webfakes web app on port",
        if (length(port_nums) > 1) "s " else " ",
        paste(port_nums, collapse = ", ")
      )
      if (!is.na(opts$access_log_file)) {
        message("Access log file: ", opts$access_log_file)
      }
      if (!is.na(opts$error_log_file)) {
        message("Error log file: ", opts$error_log_file)
      }
      msg <- structure(
        list(
          ports = self$.ports,
          port = self$.port,
          access_log = attr(srv, "options")$access_log_file,
          error_log = attr(srv, "options")$error_log_file
        ),
        class = c("webfakes_port", "callr_message", "condition")
      )
      message(msg)

      on.exit(server_stop(srv), add = TRUE)

      while (TRUE) {
        req <- server_poll(srv, cleanup)
        tryCatch(
          self$.process_request(req),
          error = function(err) {
            cat(as.character(err), file = stderr())
            response_send_error(req, as.character(err), 500L)
          }
        )
      }
    },

    mkcol = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("mkcol", path, ...))
      invisible(self)
    },

    options = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("options", path, ...))
      invisible(self)
    },

    patch = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("patch", path, ...))
      invisible(self)
    },

    post = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("post", path, ...))
      invisible(self)
    },

    propfind = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("propfind", path, ...))
      invisible(self)
    },

    put = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("put", path, ...))
      invisible(self)
    },

    report = function(path, ...) {
      self$.stack <- c(self$.stack, parse_handlers("report", path, ...))
      invisible(self)
    },

    set_config = function(key, value) {
      self$.config[[key]] <- value
      invisible(self)
    },

    use = function(..., .first = FALSE) {
      mw <- parse_handlers("use", "*", ...)
      if (.first) {
        self$.stack <- c(mw, self$.stack)
      } else {
        self$.stack <- c(self$.stack, mw)
      }
      invisible(self)
    },

    # Public data to be used in handlers
    locals = new.env(parent = parent.frame()),

    # Private data
    .ports = NULL,
    .port = NULL,
    .enable_keep_alive = NULL,
    .opts = NULL,

    # middleware stack
    .stack = list(),

    # view engines
    .engines = list(),

    # config
    .config = as.environment(list(
      views = file.path(getwd(), "views")
    )),

    # The request processing function
    .process_request = function(req) {
      req <- new_request(self, req)
      res <- req$res
      res$.delay <- NULL

      tryCatch(
        {
          for (i in sseq(res$.stackptr, length(self$.stack))) {
            handler <- self$.stack[[i]]
            m <- path_match(req$method, req$path, handler)
            if (!isFALSE(m)) {
              res$.i <- i
              if (is.list(m)) {
                req$params <- m$params
              }
              out <- handler$handler(req, res)
              if (!identical(out, "next")) break
            }
          }

          if (!res$.sent && is.null(res$.delay)) {
            if (!res$headers_sent) {
              res$send_status(404)
            } else if (
              (res$get_header("Transfer-Encoding") %||% "") == "chunked"
            ) {
              res$send_chunk(raw(0))
              res$headers_sent <- TRUE
              res$send("")
            } else {
              res$send("")
            }
          }
        },
        webfakes_error = function(err) {}
      )
    },

    .get_error_log = function() {
      if (!is.na(self$.opts$error_log_file)) {
        paste0("Error log:\n", read_char(self$.opts$error_log_file))
      }
    }
  )

  self
}

parse_handlers <- function(method, path, ...) {
  handlers <- list(...)
  ans <- list()
  for (h in seq_along(handlers)) {
    handler <- handlers[[h]]
    if (is.function(handler)) {
      rec <- list(
        method = method,
        path = path,
        handler = handler,
        name = names(handlers)[h]
      )
      ans <- c(ans, list(rec))
    } else {
      stop("Invalid webfakes handler")
    }
  }

  ans
}

test_that("uuid_random() format", {
  uu <- replicate(1000, uuid_random())

  # format is right
  regex <- paste0(
    "^[0-9a-f]{8}-",
    "[0-9a-f]{4}-",
    "4[0-9a-f]{3}-",
    "[89ab][0-9a-f]{3}-",
    "[0-9a-f]{12}$"
  )
  expect_true(all(grepl(regex, uu, perl = TRUE)))

  # all bits are used, except for the dashes and M and N
  pos <- setdiff(1:36, c(9, 14, 19, 24, 15, 20))
  xd <- format(as.hexmode(0:15))
  for (p in pos) expect_true(all(xd %in% substr(uu, p, p)))

  # all bits are used for N
  expect_true(all(c("8", "9", "a", "b") %in% substr(uu, 20, 20)))
})

app <- new_app()
app$use(mw_urlencoded())
app$post("/form", function(req, res) {
  ret <- list(
    form = req$form
  )
  res$send_json(ret, pretty = TRUE, auto_unbox = TRUE)
})
web <- local_app_process(app)

test_that("mw-urlencoded", {
  url <- web$url("/form")
  handle <- curl::new_handle()
  data <- charToRaw("foo=bar&foobar=100")
  curl::handle_setheaders(
    handle,
    "content-type" = "application/x-www-form-urlencoded"
  )
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )
  resp <- curl::curl_fetch_memory(url, handle = handle)
  echo <- jsonlite::fromJSON(
    rawToChar(resp$content),
    simplifyVector = FALSE
  )
  expect_equal(echo, list(form = list("foo" = "bar", "foobar" = "100")))
})

test_that("decode_url option", {
  app <- webfakes::new_app()
  app$get(
    webfakes::new_regexp("^/hello/(?<path>.*)$"),
    function(req, res) {
      res$send(paste0("Return content of ", req$params$path, "!"))
    }
  )
  web <- webfakes::local_app_process(app)

  resp <- curl::curl_fetch_memory(web$url("/hello/foo%2fbar/suffix"))
  expect_equal(resp$status_code, 200L)
  expect_equal(
    rawToChar(resp$content),
    "Return content of foo/bar/suffix!"
  )

  app <- webfakes::new_app()
  app$get(
    webfakes::new_regexp("^/hello/(?<path>.*)$"),
    function(req, res) {
      res$send(paste0("Return content of ", req$params$path, "!"))
    }
  )
  web <- webfakes::local_app_process(
    app,
    opts = server_opts(remote = TRUE, decode_url = FALSE)
  )

  resp <- curl::curl_fetch_memory(web$url("/hello/foo%2fbar/suffix"))
  expect_equal(resp$status_code, 200L)
  expect_equal(
    rawToChar(resp$content),
    "Return content of foo%2fbar/suffix!"
  )
})

app <- new_app()
app$engine("html", tmpl_glue())
app$set_config("views", test_path("fixtures", "views"))
app$get("/hello/:user", function(req, res) {
  locals <- c(as.list(req$params), greeting = "hello")
  html <- res$render("test-view", locals = locals)
  res$
    set_type("text/html")$
    send(html)
})
web <- local_app_process(app)

test_that("glue templating", {
  url <- web$url("/hello/gabor")
  resp <- curl::curl_fetch_memory(url)
  expect_match(rawToChar(resp$content), "hello gabor")
})

test_that("invalid handler", {
  app <- new_app()
  expect_snapshot(error = TRUE, {
    app$use("foobar")
    app$get("/foo", 1:100)
  })
})

test_that("handler can declare a locals argument", {
  app <- new_app()
  app$use(function(req, res, locals) {
    if (is.null(locals$num)) locals$num <- 0L
    locals$num <- locals$num + 1L
    "next"
  })
  app$get("/count", function(req, res, locals) {
    res$send(as.character(locals$num))
  })

  web <- local_app_process(app)
  expect_equal(
    rawToChar(curl::curl_fetch_memory(web$url("/count"))$content),
    "1"
  )
  expect_equal(
    rawToChar(curl::curl_fetch_memory(web$url("/count"))$content),
    "2"
  )
})

app <- new_app()
app$use(function(req, res) {
  tmp <- tempfile()
  saveRDS(list(req = req, res = res), file = tmp)
  res$
    set_status(200)$
    send(normalizePath(tmp))
})
proc <- local_app_process(app)
withr::local_options(list(HTTPUserAgent = "It is me, libcurl"))
resp <- curl::curl_fetch_memory(proc$url())
tmp <- rawToChar(resp$content)

withr::defer(unlink(tmp))

# Verify_output uses a png() graphics device, and fails if there is
# no png() support. So we skip theses tests then. capabilities()
# is very slow on macOS, because it starts up X11, so we'll just assume
# that macOS has a png device.

skip_without_png_device <- function() {
  if (.Platform$OS.type == "windows") return()
  if (! capabilities("png") || ! capabilities("X11")) {
    skip("Needs a PNG device")
  }
}

test_that("webfakes_app", {
  skip_without_png_device()
  app <- new_app()
  app$use("add etag" = mw_etag())
  app$get("/api", function(req, res) res$send("foobar"))
  verify_output(
    test_path("fixtures", "output", "webfakes_app.txt"),
    app
  )
})

test_that("webfakes_request", {
  skip_without_png_device()
  req <- readRDS(tmp)$req
  req$url <- "http://127.0.0.1:3000/"
  req$headers$Host <- "127.0.0.1:3000"
  req$headers$`Accept-Encoding` <- "deflate, gzip"
  verify_output(
    test_path("fixtures", "output", "webfakes_request.txt"),
    req
  )
})

test_that("webfakes_response", {
  skip_without_png_device()
  res <- readRDS(tmp)$res
  verify_output(
    test_path("fixtures", "output", "webfakes_response.txt"),
    res
  )
})

test_that("webfakes_regexp", {
  skip_without_png_device()
  verify_output(
    test_path("fixtures", "output", "webfakes_regexp.txt"),
    new_regexp("^(foo|bar)$")
  )
})

test_that("webfakes_app_process", {
  skip_without_png_device()
  app <- new_app()
  proc <- new_app_process(app, start = TRUE)
  proc$stop()
  # make the output deterministic
  proc$.port <- 3000
  verify_output(
    test_path("fixtures", "output", "webfakes_app_process.txt"),
    proc
  )
})

app <- new_app()
app$use(mw_etag())
app$use(mw_static(root = test_path("fixtures", "static")))
set_headers <- function(req, res) {
  res$set_header("foo", "bar")
}
app$use(mw_static(
  root = test_path("fixtures", "static2"),
  set_headers = set_headers
))
app$get("/static.html", function(req, res) {
  res$send("this is never reached")
})
app$get("/fallback", function(req, res) {
  res$send("this is the fallback")
})
web <- local_app_process(app)

test_that("static file", {
  url <- web$url("/static.html")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  # type is set from the file extension
  expect_equal(resp$type, "text/html")
  expect_equal(
    rawToChar(resp$content),
    read_char(test_path("fixtures", "static", "static.html"))
  )
})

test_that("static file in subdirectory", {
  url <- web$url("/subdir/static.json")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  # type is set from the file extension
  expect_equal(resp$type, "application/json")
  expect_equal(
    rawToChar(resp$content),
    read_char(test_path("fixtures", "static", "subdir", "static.json"))
  )
})

test_that("file not found", {
  url <- web$url("/notfound")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)
})

test_that("file not found falls back", {
  url <- web$url("/fallback")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "text/plain")
  expect_equal(
    rawToChar(resp$content),
    "this is the fallback"
  )
})

test_that("directory is 404", {
  url <- web$url("/subdir")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 404L)
})

test_that("set_headers callback", {
  url <- web$url("/static.tar.gz")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  expect_equal(resp$type, "application/gzip")
  headers <- curl::parse_headers_list(resp$headers)
  expect_equal(headers$foo, "bar")
  expect_equal(
    resp$content,
    read_bin(test_path("fixtures", "static2", "static.tar.gz"))
  )
})

test_that("if-none-match is respected", {
  url <- web$url("/static.html")
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  etag <- curl::parse_headers_list(resp$headers)$etag

  h <- curl::new_handle()
  curl::handle_setheaders(h, "If-None-Match" = etag)
  resp2 <- curl::curl_fetch_memory(url, handle = h)
  expect_equal(resp2$status_code, 304L)
  expect_equal(resp2$content, raw(0))
})

app <- new_app()$use(mw_etag())$get("/txt", function(req, res) {
  res$set_type("text/plain")$send("textual")
})$get("/txt-empty", function(req, res) {
  res$set_type("text/plain")$send("")
})$get("/raw", function(req, res) {
  res$set_type("applicartion/octet-stream")$send(charToRaw("textual"))
})$get("/raw-empty", function(req, res) {
  res$set_type("applicartion/octet-stream")$send(raw(0))
})

web <- local_app_process(app)

test_that("text/plain response", {
  url <- web$url("/txt")
  resp <- curl::curl_fetch_memory(url)
  headers <- curl::parse_headers_list(resp$headers)
  expect_true("etag" %in% names(headers))
  expect_match(headers$etag, "\"[-a-zA-Z0-9]+\"")
})

test_that("raw response", {
  txt <- curl::curl_fetch_memory(web$url("/txt"))
  raw <- curl::curl_fetch_memory(web$url("/raw"))
  htxt <- curl::parse_headers_list(txt$headers)
  hraw <- curl::parse_headers_list(raw$headers)
  expect_equal(htxt$etag, hraw$etag)
})

test_that("etag for empty response", {
  txt <- curl::curl_fetch_memory(web$url("/txt-empty"))
  raw <- curl::curl_fetch_memory(web$url("/raw-empty"))
  htxt <- curl::parse_headers_list(txt$headers)
  hraw <- curl::parse_headers_list(raw$headers)
  expect_equal(htxt$etag, "\"00000000\"")
  expect_equal(htxt$etag, hraw$etag)
})

app <- new_app()
app$use(mw_multipart())
app$put("/form", function(req, res) {
  ret = list(form = req$form, files = req$files)
  res$send_json(ret, pretty = TRUE, auto_unbox = TRUE)
})
web <- local_app_process(app)

test_that("mw_multipart", {
  on.exit(rm(tmp), add = TRUE)
  tmp <- tempfile()
  writeBin(charToRaw("foobar\n"), con = tmp)
  url <- web$url("/form")
  handle <- curl::new_handle()
  curl::handle_setopt(handle, customrequest = "PUT")
  curl::handle_setform(
    handle,
    a = "1",
    b = "2",
    c = curl::form_file(tmp, type = "text/plain")
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  echo <- jsonlite::fromJSON(rawToChar(resp$content), simplifyVector = FALSE)
  expect_equal(echo$form, list(a = "1", b = "2"))
  expect_equal(
    echo$files$c,
    list(
      filename = basename(tmp),
      value = base64_encode("foobar\n")
    )
  )
})

test_that("HTTPS", {
  skip_on_cran()
  app <- new_app()
  app$get("/hello", function(req, res) {
    res$send("Hello there!")
  })
  http <- local_app_process(app, port = "0s")
  cainfo <- system.file("cert/localhost/ca.crt", package = "webfakes")
  resp <- if (.Platform$OS.type == "windows") {
    callr_curl(http$url("/hello"), list(cainfo = cainfo))
  } else {
    curl::curl_fetch_memory(
      http$url("/hello"),
      handle = curl::new_handle(cainfo = cainfo)
    )
  }
  expect_equal(resp$status_code, 200L)
  expect_equal(tolower(resp$scheme), "https")
})

test_that("HTTP + HTTPS", {
  skip_on_cran()
  app <- new_app()
  app$get("/hello", function(req, res) {
    res$send("Hello there!")
  })
  http <- local_app_process(app, port = c("0", "0s"))
  cainfo <- system.file("cert/localhost/ca.crt", package = "webfakes")

  resp1 <- if (.Platform$OS.type == "windows") {
    callr_curl(http$url("/hello"), list(cainfo = cainfo))
  } else {
    curl::curl_fetch_memory(
      http$url("/hello"),
      handle = curl::new_handle(cainfo = cainfo)
    )
  }
  expect_equal(resp1$status_code, 200L)
  expect_equal(tolower(resp1$scheme), "http")

  resp2 <- if (.Platform$OS.type == "windows") {
    callr_curl(http$url("/hello", https = TRUE), list(cainfo = cainfo))
  } else {
    curl::curl_fetch_memory(
      http$url("/hello", https = TRUE),
      handle = curl::new_handle(cainfo = cainfo)
    )
  }
  expect_equal(resp2$status_code, 200L)
  expect_equal(tolower(resp2$scheme), "https")
})

test_that("Redirect HTTP to HTTPS", {
  skip_on_cran()
  app <- new_app()
  app$get("/hello", function(req, res) {
    res$send("Hello there!")
  })
  http <- local_app_process(app, port = c("0r", "0s"))
  cainfo <- system.file("cert/localhost/ca.crt", package = "webfakes")

  resp1 <- if (.Platform$OS.type == "windows") {
    callr_curl(http$url("/hello"), list(cainfo = cainfo))
  } else {
    curl::curl_fetch_memory(
      http$url("/hello"),
      handle = curl::new_handle(cainfo = cainfo)
    )
  }
  expect_equal(resp1$status_code, 200L)
  expect_equal(tolower(resp1$scheme), "https")

  resp2 <- if (.Platform$OS.type == "windows") {
    callr_curl(http$url("/hello", https = TRUE), list(cainfo = cainfo))
  } else {
    curl::curl_fetch_memory(
      http$url("/hello", https = TRUE),
      handle = curl::new_handle(cainfo = cainfo)
    )
  }
  expect_equal(resp2$status_code, 200L)
  expect_equal(tolower(resp2$scheme), "https")
})

app <- new_app()
app$use(mw_raw())
app$post("/raw", function(req, res) {
  res$set_type("application/octet-stream")$send(req$raw)
})
web <- local_app_process(app)

test_that("raw body parser", {
  url <- web$url("/raw")
  data <- charToRaw(jsonlite::toJSON(list(foo = "bar", foobar = 1:3)))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "content-type" = "application/octet-stream")
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
  expect_equal(data, resp$content)
})

test_that("non-matching content-type", {
  url <- web$url("/raw")
  data <- charToRaw(jsonlite::toJSON(list(foo = "bar", foobar = 1:3)))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "content-type" = "application/json")
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )

  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 404L)
})

app <- new_app()$use(mw_log())$get("/txt", function(req, res) {
  res$set_type("text/plain")$send("textual")
})$get("/html", function(req, res) {
  res$set_type("text/html")$send("<html><head></head><body>hello</body></html>")
})$get("/raw", function(req, res) {
  res$set_type("applicartion/octet-stream")$send(charToRaw("raw"))
})
web <- local_app_process(app, callr_opts = list(stdout = "|"))

test_that("text/plain response", {
  url <- web$url("/txt")
  resp <- curl::curl_fetch_memory(url)
  plr <- web$.process$poll_io(1000)
  expect_equal(plr[["output"]], "ready")
  log <- web$.process$read_output_lines()
  expect_match(log, "GET http://127\\.0\\.0\\.1:[0-9]*/txt 200 [0-9]+ ms - 7")
})

test_that("text/html response", {
  url <- web$url("/html")
  resp <- curl::curl_fetch_memory(url)
  plr <- web$.process$poll_io(1000)
  expect_equal(plr[["output"]], "ready")
  log <- web$.process$read_output_lines()
  expect_match(log, "GET http://127\\.0\\.0\\.1:[0-9]+/html 200 [0-9]+ ms - 44")
})

test_that("application/octet-stream response", {
  url <- web$url("/raw")
  resp <- curl::curl_fetch_memory(url)
  plr <- web$.process$poll_io(1000)
  expect_equal(plr[["output"]], "ready")
  log <- web$.process$read_output_lines()
  expect_match(log, "GET http://127\\.0\\.0\\.1:[0-9]+/raw 200 [0-9]+ ms - 3")
})

#' Fake OAuth 2.0 resource and authorization app
#'
#' @includeRmd man/rmd-fragments/oauth2.Rmd description
#'
#' @details
#' The app has the following endpoints:
#' * `GET /register` is the endpoint that you can use to register
#'   your third party app. It needs to receive the `name` of the
#'   third party app, and its `redirect_uri` as query parameters,
#'   otherwise returns an HTTP 400 error. On success it returns a
#'   JSON dictionary with entries `name` (the name of the third party
#'   app), `client_id`, `client_secret` and `redirect_uri`.
#' * `GET /authorize` is the endpoint where the user of the third
#'   party app is sent. You can change the URL of this endpoint with
#'   the `authorize_endpoint` argument. It needs to receive the `client_id`
#'   of the third party app, and its correct `redirect_uri` as query
#'   parameters. It may receive a `state` string as well, which can
#'   be used by a client to identify the request. Otherwise it
#'   generates a random `state` string. On error it fails with a HTTP
#'   400 error. On success it returns a simple HTML login page.
#' * `POST /authorize/decision` is the endpoint where the HTML login
#'   page generated at `/authorize` connects back to, either with a
#'   positive or negative result. The form on the login page will send
#'   the `state` string and the user's choice in the `action` variable.
#'   If the user authorized the third party app, then they are
#'   redirected to the `redirect_uri` of the app, with a temporary
#'   `code` and the `state` string supplied as query parameters.
#'   Otherwise a simple HTML page is returned.
#' * `POST /token` is the endpoint where the third party app requests
#'   a temporary access token. It is also uses for refreshing an
#'   access token with a refresh token. You can change the URL of this
#'   endpoint with the `token_endpoint` argument.
#'   To request a new token or refresh an existing one, the following
#'   data must be included in either a JSON or an URL encoded request body:
#'   - `grant_type`, this must be `authorization_code` for new tokens,
#'     and `refresh_token` for refreshing.
#'   - `code`, this must be the temporary code obtained from the
#'     `/authorize/decision` redirection, for new tokens. It is not
#'     needed when refreshing.
#'   - `client_id` must be the client id of the third party app.
#'   - `client_secret` must be the client secret of the third party
#'     app.
#'   - `redirect_uri` must be the correct redirection URI of the
#'     third party app. It is not needed when refreshing tokens.
#'   - `refresh_token` must be the refresh token obtained previously,
#'     when refreshing a token. It is not needed for new tokens.
#'   On success a JSON dictionary is returned with entries:
#'   `access_token`, `expiry` and `refresh_token`. (The latter is
#'   omitted if the `refresh` argument is `FALSE`).
#' * `GET /locals` returns a list of current apps, access tokens and
#'   refresh tokens.
#' * `GET /data` is an endpoint that returns a simple JSON response,
#'   and needs authorization.
#'
#' ## Notes
#'
#' * Using this app in your tests requires the glue package, so you
#'   need to put it in `Suggests`.
#' * You can add custom endpoints to the app, as needed.
#' * If you need authorization in your custom endpoint, call
#'   `app$is_authorized()` in your handler:
#'   ```
#'   if (!app$is_authorized(req, res)) return()
#'   ```
#'   `app$is_authorized()` returns an HTTP 401 response if the
#'   client is not authorized, so you can simply return from your
#'   handler.
#'
#' For more details see `vignette("oauth", package = "webfakes")`.
#'
#' @section `oauth2_resource_app()`:
#' App representing the API server (resource/authorization)
#' @return a `webfakes` app
#' @param access_duration After how many seconds should access tokens
#'   expire.
#' @param refresh_duration After how many seconds should refresh
#'   tokens expire (ignored if `refresh` is `FALSE`).
#' @param refresh Should a refresh token be returned (logical).
#' @param seed Random seed used when creating tokens. If `NULL`,
#'   we rely on R to provide a seed. The app uses its own RNG stream,
#'   so it does not affect reproducibility of the tests.
#' @param authorize_endpoint The authorization endpoint of the resource
#'   server. Change this from the default if the real app that you
#'   are faking does not use `/authorize`.
#' @param token_endpoint The endpoint to request tokens. Change this if the
#'   real app that you are faking does not use `/token`.
#' @return webfakes app
#'
#' @export
#' @family OAuth2.0 functions

oauth2_resource_app <- function(
  access_duration = 3600L,
  refresh_duration = 7200L,
  refresh = TRUE,
  seed = NULL,
  authorize_endpoint = "/authorize",
  token_endpoint = "/token"
) {
  access_duration
  refresh_duration
  refresh
  seed
  authorize_endpoint
  token_endpoint

  app <- new_app()
  app$locals$seed <- seed %||% get_seed()

  # Parse body for /authorize/decision, /token
  app$use(mw_urlencoded())
  app$use(mw_json())

  app$locals$tpapps <- data.frame(
    name = character(),
    client_id = character(),
    client_secret = character(),
    redirect_uri = character()
  )

  app$set_config("views", system.file("views", package = "webfakes"))
  app$engine("html", tmpl_glue())

  app$get("/register", function(req, res) {
    if (is.null(req$query$name) || is.null(req$query$redirect_uri)) {
      res$set_status(400L)$send(
        "Cannot register without 'name' and 'redirect_uri'"
      )
      return()
    }

    rec <- list(
      name = req$query$name,
      client_id = paste0("id-", generate_token()),
      client_secret = paste0("secret-", generate_token()),
      redirect_uri = req$query$redirect_uri
    )
    app$locals$tpapps <- rbind(app$locals$tpapps, rec)

    res$send_json(rec)
  })

  app$get(authorize_endpoint, function(req, res) {
    # Missing or invalid client id
    client_id <- req$query$client_id
    if (is.null(client_id)) {
      res$set_status(400L)$send("Invalid authorization request, no client id")
      return()
    } else if (!client_id %in% app$locals$tpapps$client_id) {
      res$set_status(400L)$send(
        "Invalid authorization request, unknown client id"
      )
      return()
    }

    tpapps <- app$locals$tpapps
    tprec <- tpapps[match(client_id, tpapps$client_id), ]

    # Bad redirect URL?
    if (req$query$redirect_uri %||% "" != tprec$redirect_uri) {
      res$set_status(400L)$send(paste0(
        "Invalid authorization request, redirect URL mismatch: ",
        req$query$redirect_uri %||% "",
        " vs ",
        tprec$redirect_uri
      ))
      return()
    }

    state <- req$query$state %||% generate_token()
    app$locals$states <- c(app$locals$states, set_name(client_id, state))

    html <- res$render("authorize", list(state = state, app = tprec$name))
    res$set_type("text/html")$send(html)
  })

  app$post(paste0(authorize_endpoint, "/decision"), function(req, res) {
    state <- req$form$state
    if (is.null(state) || !state %in% names(app$locals$states)) {
      res$set_status(400L)$send("Invalid decision, no state")
      return()
    }

    client_id <- app$locals$states[state]
    tpapps <- app$locals$tpapps
    tprec <- tpapps[match(client_id, tpapps$client_id), ]

    app$local$states <-
      app$local$states[setdiff(names(app$local$states), state)]

    if (req$form$action %||% "" == "yes") {
      code <- generate_token()
      # TODO: make this app specific
      app$locals$codes <- c(app$locals$codes, code)

      red_uri <- paste0(tprec$redirect_uri, "?code=", code, "&state=", state)
      res$redirect(red_uri)$send()
    } else {
      res$send("Maybe next time.")
    }
  })

  local_app_seed <- function(.local_envir = parent.frame()) {
    old_seed <- get_seed()
    set_seed(app$locals$seed)
    defer(
      {
        app$locals$seed <- get_seed()
        set_seed(old_seed)
      },
      envir = .local_envir
    )
  }

  new_access_token <- function(client_id, duration) {
    local_app_seed()
    token <- paste0("token-", generate_token())

    new <- data.frame(
      stringsAsFactors = FALSE,
      client_id = client_id,
      token = token,
      expiry = Sys.time() + duration
    )

    app$locals$tokens <- rbind(app$locals$tokens, new)
    token
  }

  new_refresh_token <- function(client_id, duration) {
    local_app_seed()
    token <- paste0("refresh-token-", generate_token())

    new <- data.frame(
      stringsAsFactors = FALSE,
      client_id = client_id,
      token = token,
      expiry = Sys.time() + duration
    )

    app$locals$refresh_tokens <- rbind(app$locals$refresh_tokens, new)
    token
  }

  expire_tokens <- function() {
    if (!is.null(app$locals$tokens)) {
      app$locals$tokens <- app$locals$tokens[
        app$locals$tokens$expiry > Sys.time(),
        ,
        drop = FALSE
      ]
    }
    if (!is.null(app$locals$refresh_tokens)) {
      app$locals$refresh_tokens <- app$locals$refresh_tokens[
        app$locals$refresh_tokens$expiry > Sys.time(),
        ,
        drop = FALSE
      ]
    }
  }

  is_valid_token <- function(token) {
    expire_tokens()
    token %in% app$locals$tokens$token
  }

  is_valid_refresh_token <- function(client_id, token) {
    expire_tokens()
    wh <- match(token, app$locals$refresh_tokens$token)
    if (is.na(wh)) {
      return(FALSE)
    }
    # client ID must match as well
    app$locals$refresh_tokens$client_id == client_id
  }

  # For refresh tokens
  app$post(token_endpoint, function(req, res) {
    if (req$form$grant_type != "refresh_token") {
      return("next")
    }
    client_id <- req$form$client_id
    tpapps <- app$locals$tpapps
    if (!client_id %in% tpapps$client_id) {
      res$set_status(400L)$send("Invalid client id")
      return()
    }
    tprec <- tpapps[match(client_id, tpapps$client_id), ]

    if (req$form$client_secret %||% "" != tprec$client_secret) {
      res$set_status(400L)$send("Invalid token request, client secret mismatch")
      return()
    }

    if (!is_valid_refresh_token(client_id, req$form$refresh_token)) {
      res$set_status(400L)$send_json(
        list(error = "invalid_request"),
        auto_unbox = TRUE
      )
      return()
    }

    res$send_json(
      list(
        access_token = new_access_token(client_id, access_duration),
        expiry = access_duration,
        refresh_token = new_refresh_token(client_id, refresh_duration)
      ),
      auto_unbox = TRUE
    )
  })

  # For regular tokens
  app$post(token_endpoint, function(req, res) {
    if (req$form$grant_type %||% "" != "authorization_code") {
      res$set_status(400L)$send(
        "Invalid grant type, must be 'authorization_code'"
      )
      return()
    }

    if (!req$form$code %in% app$locals$codes) {
      res$set_status(400L)$send("Unknown authorization code")
      return()
    }

    tpapps <- app$locals$tpapps

    client_id <- req$form$client_id
    if (!client_id %in% tpapps$client_id) {
      res$set_status(400L)$send("Invalid client id")
      return()
    }

    tprec <- tpapps[match(client_id, tpapps$client_id), ]

    if (req$form$client_secret %||% "" != tprec$client_secret) {
      res$set_status(400L)$send("Invalid token request, client secret mismatch")
      return()
    }

    if (req$form$redirect_uri %||% "" != tprec$redirect_uri) {
      res$set_status(400L)$send("Invalid token request, redirect URL mismatch")
      return()
    }

    app$locals$codes <- setdiff(app$locals$codes, req$query$code)

    res$send_json(
      list(
        access_token = new_access_token(client_id, access_duration),
        expiry = access_duration,
        refresh_token = if (refresh) {
          new_refresh_token(client_id, refresh_duration)
        }
      ),
      auto_unbox = TRUE,
      pretty = TRUE
    )
  })

  app$get("/locals", function(req, res) {
    res$set_status(200L)$send_json(
      list(
        apps = app$locals$tpapps,
        access = app$locals$tokens,
        refresh = app$locals$refresh_tokens
      ),
      auto_unbox = TRUE
    )
  })

  app$is_authorized <- function(req, res) {
    expire_tokens()
    if (!("Authorization" %in% names(req$headers))) {
      res$set_status(401L)$send("Missing bearer token")
      return(FALSE)
    }

    token <- gsub("Bearer ", "", req$headers$Authorization[[1]])
    if (!is_valid_token(token)) {
      res$set_status(401L)$send("Invalid bearer token")
      return(FALSE)
    }
    TRUE
  }

  app$get("/data", function(req, res) {
    if (!app$is_authorized(req, res)) {
      return()
    }
    res$send_json(list(data = "top secret!"))
  })

  app
}

#' App representing the third-party app
#'
#' @includeRmd man/rmd-fragments/oauth2.Rmd description
#'
#' @details
#' Endpoints:
#' * `POST /login/config` Use this endpoint to configure the client ID
#'   and the client secret of the app, received from
#'   [oauth2_resource_app()] (or another resource app). You need to
#'   send in a JSON or URL encoded body:
#'   - `auth_url`, the authorization URL of the resource app.
#'   - `token_url`, the token URL of the resource app.
#'   - `client_id`, the client ID, received from the resource app.
#'   - `client_secret` the client secret, received from the resource
#'     app.
#' * `GET /login` Use this endpoint to start the login process. It
#'   will redirect to the resource app for authorization and after the
#'   OAuth2.0 dance to `/login/redirect`.
#' * `GET /login/redirect`, `POST /login/redirect` This is the
#'   redirect URI of the third party app. (Some HTTP clients redirect
#'   a `POST` to a `GET`, others don't, so it has both.) This endpoint
#'   is used by the resource app, and it received the `code` that can
#'   be exchanged to an access token and the `state` which was
#'   generated in `/login`. It contacts the resource app to get an
#'   access token, and then stores the token in its `app$locals`
#'   local variables. It fails with HTTP code 500 if it cannot obtain
#'   an access token. On success it returns a JSON dictionary with
#'   `access_token`, `expiry` and `refresh_token` (optionally) by
#'   default. This behavior can be changed by redefining the
#'   `app$redirect_hook()` function.
#' * `GET /locals` returns the tokens that were obtained from the
#'   resource app.
#' * `GET /data` is an endpoint that uses the obtained token(s) to
#'   connect to the `/data` endpoint of the resource app. The `/data`
#'   endpoint of the resource app needs authorization. It responds
#'   with the response of the resource app. It tries to refresh the
#'   access token of the app if needed.
#'
#' For more details see `vignette("oauth", package = "webfakes")`.
#'
#' @param name Name of the third-party app
#' @return webfakes app
#' @export
#' @family OAuth2.0 functions

oauth2_third_party_app <- function(name = "Third-Party app") {
  app <- new_app()
  app$use(mw_urlencoded())
  app$use(mw_json())

  app$locals$auth_url <- NA_character_
  app$locals$token_url <- NA_character_
  app$locals$client_id <- NA_character_
  app$locals$client_secret <- NA_character_

  app$post("/login/config", function(req, res) {
    if (
      is.null(req$json$auth_url) ||
        is.null(req$json$token_url) ||
        is.null(req$json$client_id) ||
        is.null(req$json$client_secret)
    ) {
      res$set_status(400L)$send(
        "Need `client_id` and `client_secret` to config auth"
      )
      return()
    }

    if (!is.na(app$locals$auth_url)) {
      res$set_status(400L)$send("Auth already configured")
      return()
    }

    app$locals$auth_url <- req$json$auth_url
    app$locals$token_url <- req$json$token_url
    app$locals$client_id <- req$json$client_id
    app$locals$client_secret <- req$json$client_secret

    res$send_json(list(response = "Authorization configured"))
  })

  app$get("/login", function(req, res) {
    state <- generate_token()
    app$locals$state <- state

    url <- paste0(
      app$locals$auth_url,
      "?client_id=",
      app$locals$client_id,
      "&redirect_uri=",
      paste0(req$url, "/redirect"),
      "&state=",
      state
    )
    res$redirect(url)$send()
  })

  # I could not convince curl to redirect a POST to a GET, so both are good
  app$all("/login/redirect", function(req, res) {
    code <- req$query$code
    state <- req$query$state
    if (is.null(code) || is.null(state)) {
      res$set_status(400L)$send(
        "Invalid request via auth server, no 'code' or 'state'."
      )
      return()
    }
    if (state != app$locals$state) {
      res$set_status(400L)$send("Unknown state in request via auth server")
      return()
    }

    # Get a token
    handle <- curl::new_handle()
    data <- charToRaw(paste0(
      "grant_type=authorization_code&",
      "code=",
      code,
      "&",
      "client_id=",
      app$locals$client_id,
      "&",
      "client_secret=",
      app$locals$client_secret,
      "&",
      "redirect_uri=",
      req$url
    ))
    curl::handle_setheaders(
      handle,
      "content-type" = "application/x-www-form-urlencoded"
    )
    curl::handle_setopt(
      handle,
      customrequest = "POST",
      postfieldsize = length(data),
      postfields = data
    )

    resp <- curl::curl_fetch_memory(app$locals$token_url, handle = handle)

    if (resp$status_code != 200L) {
      res$set_status(500L)$send(paste0(
        "Failed to acquire authorization token. ",
        rawToChar(resp$content)
      ))
      return()
    }

    tokens <- rawToChar(resp$content)
    app$locals$tokens <- jsonlite::fromJSON(tokens)

    app$redirect_hook(res, tokens)
  })

  app$redirect_hook <- function(res, tokens) {
    res$send_json(text = tokens)
  }

  app$get("/locals", function(req, res) {
    res$set_status(200L)$send_json(app$locals$tokens, auto_unbox = TRUE)
  })

  get_data <- function() {
    auth <- paste("Bearer", app$locals$tokens$access_token)
    handle <- curl::new_handle()
    curl::handle_setheaders(handle, Authorization = auth)
    url <- modify_path(app$locals$token_url, "/data")
    curl::curl_fetch_memory(url, handle = handle)
  }

  try_refresh <- function() {
    refresh_token <- app$locals$tokens$refresh_token
    if (is.null(refresh_token)) {
      return(FALSE)
    }

    data <- charToRaw(paste0(
      "refresh_token=",
      refresh_token,
      "&",
      "grant_type=refresh_token"
    ))
    handle <- curl::new_handle()
    curl::handle_setheaders(
      handle,
      "content-type" = "application/x-www-form-urlencoded"
    )
    curl::handle_setopt(
      handle,
      customrequest = "POST",
      postfieldsize = length(data),
      postfields = data
    )

    resp <- curl::curl_fetch_memory(app$locals$token_url, handle = handle)

    if (resp$status_code != 200L) {
      return(FALSE)
    }

    tokens <- rawToChar(resp$content)
    app$locals$tokens <- jsonlite::fromJSON(tokens)

    TRUE
  }

  app$get("/data", function(req, res) {
    resp <- get_data()

    if (resp$status_code == 401) {
      if (try_refresh()) resp <- get_data()
    }

    res$set_status(resp$status_code)$set_type(resp$type)$send(resp$content)
  })
}

#' Helper function to log in to a third party OAuth2.0 app without a
#' browser
#'
#' It works with [oauth2_resource_app()], and any third party app,
#' including the fake [oauth2_third_party_app()].
#'
#' See `test-oauth.R` in webfakes for an example.
#'
#' @param login_url The login URL of the third party app.
#' @return A named list with
#' * `login_response` The curl HTTP response object for the login
#'   page.
#' * `token_response` The curl HTTP response object for submitting
#'   the login page.
#'
#' @family OAuth2.0 functions
#' @export

oauth2_login <- function(login_url) {
  login_resp <- curl::curl_fetch_memory(login_url)

  html <- rawToChar(login_resp$content)
  xml <- xml2::read_html(html)
  form <- xml2::xml_find_first(xml, "//form")
  input <- xml2::xml_find_first(form, "//input")

  actn <- xml2::xml_attr(form, "action")
  stnm <- xml2::xml_attr(input, "name")
  stvl <- xml2::xml_attr(input, "value")

  data <- charToRaw(paste0(
    stnm,
    "=",
    stvl,
    "&",
    "action=yes"
  ))

  handle2 <- curl::new_handle()
  curl::handle_setheaders(
    handle2,
    "content-type" = "application/x-www-form-urlencoded"
  )
  curl::handle_setopt(
    handle2,
    customrequest = "POST",
    postfieldsize = length(data),
    postfields = data
  )

  psurl <- parse_url(login_resp$url)
  actn_url <- paste0(psurl$protocol, "://", psurl$host, actn)
  token_resp <- curl::curl_fetch_memory(actn_url, handle = handle2)

  list(
    login_response = login_resp,
    token_response = token_resp
  )
}

#' Helper function to use httr's OAuth2.0 functions
#' non-interactively, e.g. in test cases
#'
#' To perform an automatic acknowledgement and log in for a
#' local OAuth2.0 app, run by httr, wrap the expression that
#' obtains the OAuth2.0 token in `oauth2_httr_login()`.
#'
#' In interactive sessions, `oauth2_httr_login()` overrides the
#' `browser` option, and when httr opens a browser page, it
#' calls [oauth2_login()] in a subprocess.
#'
#' In non-interactive sessions, httr does not open a browser page,
#' only messages the user to do it manually. `oauth2_httr_login()`
#' listens for these messages, and calls [oauth2_login()] in a
#' subprocess.
#'
#' @param expr Expression that calls [httr::oauth2.0_token()],
#'   either directly, or indirectly.
#' @return The return value of `expr`.
#'
#' @seealso See `?vignette("oauth", package = "webfakes")` for a case
#' study that uses this function.
#'
#' @export
#' @family OAuth2.0 functions

oauth2_httr_login <- function(expr) {
  proc <- NULL
  if (interactive()) {
    local_options(browser = function(url) {
      proc <<- callr::r_bg(
        oauth2_login,
        list(url),
        package = "webfakes"
      )
    })
    expr
  } else {
    withCallingHandlers(
      expr,
      message = function(msg) {
        if (
          grepl("^Please point your browser to the following url:", msg$message)
        ) {
          invokeRestart("muffleMessage")
        }
        if (grepl("^http", msg$message)) {
          proc <<- callr::r_bg(
            oauth2_login,
            list(trimws(msg$message)),
            package = "webfakes"
          )
          invokeRestart("muffleMessage")
        }
      }
    )
  }
}

test_that("oauth2", {
  skip_on_cran()

  # Create the resource server
  rsapp <- local_app_process(
    oauth2_resource_app(),
    opts = server_opts(num_threads = 3)
  )
  regi_url <- rsapp$url("/register")
  auth_url <- rsapp$url("/authorize")
  toke_url <- rsapp$url("/token")

  # Create the third party app server
  tpapp <- local_app_process(
    oauth2_third_party_app("3P app"),
    opts = server_opts(num_threads = 3)
  )
  redi_url <- tpapp$url("/login/redirect")
  conf_url <- tpapp$url("/login/config")

  # Register the third party app at the resource server
  # In real life this is done by the admin of the third party app
  url <- paste0(
    regi_url,
    "?name=3P%20app",
    "&redirect_uri=",
    redi_url
  )
  resp <- curl::curl_fetch_memory(url)
  expect_equal(resp$status_code, 200L)
  regdata <- jsonlite::fromJSON(rawToChar(resp$content))

  # Now set this data on the third party app
  # In real life this is included in the config of the third party app
  # by its admin
  auth_data <- jsonlite::toJSON(
    list(
      auth_url = auth_url,
      token_url = toke_url,
      client_id = regdata$client_id,
      client_secret = regdata$client_secret
    ),
    auto_unbox = TRUE
  )
  handle <- curl::new_handle()
  curl::handle_setheaders(
    handle,
    "content-type" = "application/json"
  )
  curl::handle_setopt(
    handle,
    customrequest = "POST",
    postfieldsize = nchar(auth_data),
    postfields = auth_data
  )
  resp2 <- curl::curl_fetch_memory(conf_url, handle = handle)
  expect_equal(resp2$status_code, 200L)

  # Now everything is set up, a user can go to the login page of the
  # third party app:
  # browseURL(tpapp$url("/login"))

  # Scripting it is a bit tedious, because we need to parse the HTML and
  # submit a form. The `oauth2_login()` helper function does this.
  resp3 <- oauth2_login(tpapp$url("/login"))
  token <- jsonlite::fromJSON(rawToChar(resp3$token_response$content))

  expect_equal(resp3$login_response$status_code, 200L)
  expect_equal(resp3$login_response$type, "text/html")
  expect_equal(resp3$token_response$status_code, 200L)
  expect_match(token$access_token, "^token-[0-9a-f]+$")
})

test_that("oauth + httr", {
  skip_on_cran()
  # Not great, the OS should allocate a port, really...
  withr::local_envvar(c(
    HTTP_SERVER = "127.0.0.1",
    HTTP_SERVER_PORT = httpuv::randomPort()
  ))

  # Create the resource server
  rsappex <- oauth2_resource_app(access_duration = 0.1)
  log <- tempfile("webfakes-log-", fileext = ".log")
  on.exit(unlink(log, recursive = TRUE), add = TRUE)
  rsappex$use(logger = mw_log(stream = log), .first = TRUE)
  rsapp <- local_app_process(
    rsappex,
    opts = server_opts(num_threads = 3)
  )
  regi_url <- rsapp$url("/register")
  auth_url <- rsapp$url("/authorize")
  toke_url <- rsapp$url("/token")

  # Register httr
  url <- paste0(
    regi_url,
    "?name=httr%20local%20app",
    "&redirect_uri=",
    httr::oauth_callback()
  )
  reg_resp <- httr::GET(url)
  httr::stop_for_status(reg_resp)
  regdata <- httr::content(reg_resp)

  app <- httr::oauth_app(
    regdata$name[[1]],
    key = regdata$client_id[[1]],
    secret = regdata$client_secret[[1]],
    redirect_uri = httr::oauth_callback()
  )

  endpoint <- httr::oauth_endpoint(
    authorize = auth_url,
    access = toke_url
  )

  token <- suppressMessages(oauth2_httr_login(
    httr::oauth2.0_token(endpoint, app, cache = FALSE)
  ))

  # This will refresh the token automatically
  Sys.sleep(0.1)
  expect_message(
    cnt <- httr::content(
      httr::GET(rsapp$url("/data"), config = token),
      as = "parsed",
      type = "application/json"
    ),
    "Auto-refreshing stale OAuth"
  )

  expect_equal(cnt, list(data = list("top secret!")))

  loglines <- readLines(log)

  endpoints <- parse_url(map_chr(strsplit(loglines, " "), "[[", 2))$path
  expect_equal(
    endpoints,
    c(
      "/register",
      "/authorize",
      "/authorize/decision",
      "/token",
      "/data",
      "/token",
      "/data"
    )
  )
  status <- map_chr(strsplit(loglines, " "), "[[", 3)
  expect_equal(
    as.numeric(status),
    c(200, 200, 302, 200, 401, 200, 200)
  )
})

#' Create a new regular expression to use in webfakes routes
#'
#' Note that webfakes uses PERL regular expressions.
#'
#' @details
#' As R does not have data type or class for regular expressions,
#' you can use `new_regexp()` to mark a string as a regular expression,
#' when adding routes.
#'
#' @param x String scalar containing a regular expression.
#' @return String with class `webfakes_regexp`.
#'
#' @aliases webfakes_regexp
#' @seealso The 'Path specification' and 'Path parameters' chapters
#' of the manual of [new_app()].
#' @export
#' @examples
#' new_regexp("^/api/match/(?<pattern>.*)$")

new_regexp <- function(x) structure(x, class = "webfakes_regexp")

path_match <- function(method, path, handler) {
  if (handler$method == "use") {
    return(TRUE)
  }
  if (
    (!handler$method %in% c("all", method)) &&
      !(handler$method == "get" && method == "head")
  ) {
    return(FALSE)
  }
  pattern_match(path, handler$path)
}

pattern_match <- function(path, patterns) {
  # Make sure patterns is a list
  if (inherits(patterns, "webfakes_regexp")) {
    patterns <- list(patterns)
  } else if (is.character(patterns)) {
    patterns <- as.list(patterns)
  }

  for (p in patterns) {
    if (!inherits(p, "webfakes_regexp") && grepl(":", p)) {
      p <- path_to_regexp(p)
    }
    if (inherits(p, "webfakes_regexp")) {
      m <- re_match(path, p)
      if (m$match) return(list(params = as.list(m$groups)))
    } else {
      if (path == p) return(TRUE)
    }
  }

  FALSE
}

path_to_regexp <- function(path) {
  tokens <- strsplit(path, "/")[[1]]
  keys <- grep("^:", tokens)
  reg <- tokens
  reg[keys] <- paste0("(?<", substring(tokens[keys], 2), ">[-A-Za-z0-9_]+)")
  new_regexp(paste0("^", paste(reg, collapse = "/"), "$"))
}

test_that("middleware", {
  expect_true(path_match("foobar", "foobar2", list(method = "use")))
})

test_that("string", {
  good <- list(
    list("/foo", "/foo"),
    list("/", "/"),
    list("/***", "/***")
  )
  for (x in good) {
    expect_true(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }

  bad <- list(
    list("/foo", "/bar"),
    list("/foo", "foo")
  )
  for (x in bad) {
    expect_false(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }
})

test_that("character vector or list", {
  good <- list(
    list("/foo", c("/foo2", "/foo")),
    list("/", list("notthis", "/"))
  )
  for (x in good) {
    expect_true(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }

  bad <- list(
    list("/foo", list())
  )
  for (x in bad) {
    expect_false(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }
})

named_list <- function() structure(list(), names = character())

test_that("regexp", {
  good <- list(
    list("/foo", new_regexp("^/fo+$"), list(params = named_list())),
    list(
      "/",
      list(new_regexp("/foobar"), new_regexp("^/$")),
      list(params = named_list())
    )
  )
  for (x in good) {
    expect_equal(
      path_match("get", x[[1]], list(method = "get", path = x[[2]])),
      x[[3]]
    )
  }

  bad <- list(
    list("/foo", list("/foobar"))
  )
  for (x in bad) {
    expect_false(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }
})

test_that("list of things", {
  good <- list(
    list(
      "/foo",
      list("/foo2", new_regexp("/foo")),
      list(params = named_list())
    ),
    list("/", list(new_regexp("notthis"), "/"), TRUE)
  )
  for (x in good) {
    expect_equal(
      path_match("get", x[[1]], list(method = "get", path = x[[2]])),
      x[[3]]
    )
  }

  bad <- list(
    list("/foo", list())
  )
  for (x in bad) {
    expect_false(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }
})

test_that("regexp with capture groups", {
  good <- list(
    list(
      "/foo/bar",
      new_regexp("^/foo/([a-z]+)/?$"),
      list(params = structure(list("bar"), names = ""))
    ),
    list(
      "/foo/bar",
      new_regexp("^/(?<x>f.*)/(?<y>[a-z]+)/?$"),
      list(params = structure(list("foo", "bar"), names = c("x", "y")))
    )
  )
  for (x in good) {
    expect_equal(
      path_match("get", x[[1]], list(method = "get", path = x[[2]])),
      x[[3]]
    )
  }

  bad <- list(
    list("/foo/bar", new_regexp("/foox/([a-z]+)/?$"))
  )
  for (x in bad) {
    expect_false(path_match("get", x[[1]], list(method = "get", path = x[[2]])))
  }
})

test_that("tokens", {
  good <- list(
    list(
      "/foo",
      "/:x",
      list(params = list(x = "foo"))
    ),
    list(
      "/foo/bar",
      "/:x/:y",
      list(params = list(x = "foo", y = "bar"))
    )
  )
  for (x in good) {
    expect_equal(
      path_match("get", x[[1]], list(method = "get", path = x[[2]])),
      x[[3]]
    )
  }
})

test_response_app <- function() {
  app <- new_app()

  `%||%` <- function(l, r) if (is.null(l)) r else l

  app$locals$applocal <- "foo"

  app$engine("txt", tmpl_glue())
  app$get("/local", function(req, res) {
    res$locals$reslocal <- "bar"
    "next"
  })

  app$get("/local", function(req, res) {
    res$send(paste(res$locals$applocal, res$locals$reslocal))
  })

  app$get("/badengine", function(req, res) {
    txt <- res$render("foobar")
    res$send(txt)
  })

  app$get("/badjson", function(req, res) {
    res$send_json(1:3, text = "foo")
  })

  app$get("/file", function(req, res) {
    res$send_file(
      root = system.file(package = "webfakes"),
      file.path("examples", "static", "public", "foo", "bar.json")
    )
  })

  app$get("/type", function(req, res) {
    res$set_type("json")
    res$send("{ \"foo\": 1 }")
  })

  app$get("/write", function(req, res) {
    res$write("hello ")$write("world!")
  })

  app$get("/write-header", function(req, res) {
    res$set_header("foo", "bar")$write("hello ")$write("world!")
  })

  app$get("/write-wait", function(req, res) {
    res$locals$turn <- (res$locals$turn %||% 0) + 1L
    if (res$locals$turn == 1) {
      res$set_header("content-length", nchar("hello world!"))$write(
        "hell"
      )$delay(0.01)
    } else if (res$locals$turn == 2) {
      res$write("o world")$delay(0.01)
    } else {
      res$send("!")
    }
  })

  app$get("/send-chunk", function(req, res) {
    res$locals$turn <- (res$locals$turn %||% 0) + 1L
    if (res$locals$turn == 1) {
      res$set_header("Content-Type", "text/plain")$send_chunk(
        "first chunk\n"
      )$delay(0.01)
    } else if (res$locals$turn == 2) {
      res$send_chunk("second chunk\n")$delay(0.01)
    } else {
      res$send_chunk("third and final chunk\n")
    }
  })

  app$get("/add-header", function(req, res) {
    res$add_header("foo", "bar")
    res$add_header("foo", "bar2")
    res$add_header("foobar", "baz")
    res$send("ready")
  })

  app
}

httpbin <- local_app_process(httpbin_app(), opts = server_opts(num_threads = 6))
httpbin$local_env(c(FOO = "{url}xxx"))

r_variant <- function() {
  if (getRversion() < "4.2.0") {
    "old-r"
  } else {
    "new-r"
  }
}

callr_curl <- function(url, options = list()) {
  callr::r(
    function(url, options) {
      h <- curl::new_handle()
      curl::handle_setopt(h, .list = options)
      curl::curl_fetch_memory(url, handle = h)
    },
    list(url = url, options = options),
    env = c(
      callr::rcmd_safe_env(),
      CURL_SSL_BACKEND = "openssl",
      CURL_CA_BUNDLE = if ("cainfo" %in% names(options)) options$cainfo
    )
  )
}

#' @title Happy HTTP testing with webfakes
#' @name introduction
#' @section Happy HTTP testing with webfakes:
#'
#' ```{r child = "vignettes/introduction.Rmd"}
#' ```
NULL

#' @title webfakes glossary
#' @name glossary
#' @section Webfakes glossary:
#'
#' ```{r child = "vignettes/glossary.Rmd"}
#' ```
NULL

#' @title How to use webfakes in your tests
#' @name how-to
#' @section How to use webfakes in your tests:
#'
#' ```{r child = "vignettes/how-to.Rmd"}
#' ```
NULL

XX <- 255L
EQ <- 254L
INVALID <- XX

index_64 <- as.integer(c(
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  62,
  XX,
  XX,
  XX,
  63,
  52,
  53,
  54,
  55,
  56,
  57,
  58,
  59,
  60,
  61,
  XX,
  XX,
  XX,
  EQ,
  XX,
  XX,
  XX,
  0,
  1,
  2,
  3,
  4,
  5,
  6,
  7,
  8,
  9,
  10,
  11,
  12,
  13,
  14,
  15,
  16,
  17,
  18,
  19,
  20,
  21,
  22,
  23,
  24,
  25,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  26,
  27,
  28,
  29,
  30,
  31,
  32,
  33,
  34,
  35,
  36,
  37,
  38,
  39,
  40,
  41,
  42,
  43,
  44,
  45,
  46,
  47,
  48,
  49,
  50,
  51,
  XX,
  XX,
  XX,
  XX,
  XX,

  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX,
  XX
))

base64_decode <- function(x) {
  if (is.character(x)) {
    x <- charToRaw(x)
  }

  len <- length(x)
  idx <- 1
  c <- integer(4)
  out <- raw()
  while (idx <= len) {
    i <- 1
    while (i <= 4) {
      uc <- index_64[[as.integer(x[[idx]]) + 1L]]
      idx <- idx + 1
      if (uc != INVALID) {
        c[[i]] <- uc
        i <- i + 1
      }
      if (idx > len) {
        if (i <= 4) {
          if (i <= 2) {
            return(rawToChar(out))
          }
          if (i == 3) {
            c[[3]] <- EQ
            c[[4]] <- EQ
          }
          break
        }
      }
    }

    if (c[[1]] == EQ || c[[2]] == EQ) {
      break
    }

    #print(sprintf("c1=%d,c2=%d,c3=%d,c4=%d\n", c[1],c[2],c[3],c[4]))

    out[[length(out) + 1]] <- as.raw(bitwOr(
      bitwShiftL(c[[1]], 2L),
      bitwShiftR(bitwAnd(c[[2]], 0x30), 4L)
    ))

    if (c[[3]] == EQ) {
      break
    }

    out[[length(out) + 1]] <- as.raw(bitwOr(
      bitwShiftL(bitwAnd(c[[2]], 0x0F), 4L),
      bitwShiftR(bitwAnd(c[[3]], 0x3C), 2L)
    ))

    if (c[[4]] == EQ) {
      break
    }

    out[[length(out) + 1]] <- as.raw(bitwOr(
      bitwShiftL(bitwAnd(c[[3]], 0x03), 6L),
      c[[4]]
    ))
  }
  rawToChar(out)
}

basis64 <- charToRaw(paste(c(LETTERS, letters, 0:9, "+", "/"), collapse = ""))

base64_encode <- function(x) {
  if (is.character(x)) {
    x <- charToRaw(x)
  }

  len <- length(x)
  rlen <- floor((len + 2L) / 3L) * 4L
  out <- raw(rlen)
  ip <- op <- 1L
  c <- integer(4)

  while (len > 0L) {
    c[[1]] <- as.integer(x[[ip]])
    ip <- ip + 1L
    if (len > 1L) {
      c[[2]] <- as.integer(x[ip])
      ip <- ip + 1L
    } else {
      c[[2]] <- 0L
    }
    out[op] <- basis64[1 + bitwShiftR(c[[1]], 2L)]
    op <- op + 1L
    out[op] <- basis64[
      1 +
        bitwOr(
          bitwShiftL(bitwAnd(c[[1]], 3L), 4L),
          bitwShiftR(bitwAnd(c[[2]], 240L), 4L)
        )
    ]
    op <- op + 1L

    if (len > 2) {
      c[[3]] <- as.integer(x[ip])
      ip <- ip + 1L
      out[op] <- basis64[
        1 +
          bitwOr(
            bitwShiftL(bitwAnd(c[[2]], 15L), 2L),
            bitwShiftR(bitwAnd(c[[3]], 192L), 6L)
          )
      ]
      op <- op + 1L
      out[op] <- basis64[1 + bitwAnd(c[[3]], 63)]
      op <- op + 1L
    } else if (len == 2) {
      out[op] <- basis64[1 + bitwShiftL(bitwAnd(c[[2]], 15L), 2L)]
      op <- op + 1L
      out[op] <- charToRaw("=")
      op <- op + 1L
    } else {
      ## len == 1
      out[op] <- charToRaw("=")
      op <- op + 1L
      out[op] <- charToRaw("=")
      op <- op + 1L
    }
    len <- len - 3L
  }

  rawToChar(out)
}

test_that("base64_decode", {
  decode_tests <- c(
    'YWE=' = 'aa',
    ' YWE=' = 'aa',
    'Y WE=' = 'aa',
    'YWE= ' = 'aa',
    "Y\nW\r\nE=" = 'aa',
    'YWE=====' = 'aa', # extra padding
    'YWE' = 'aa', # missing padding
    'YWFh====' = 'aaa',
    'YQ' = 'a',
    'Y' = '',
    'x==' = ''
  )

  for (i in seq_along(decode_tests)) {
    encoded <- names(decode_tests)[[i]]
    expected <- decode_tests[[i]]

    decoded <- base64_decode(encoded)
    expect_equal(decoded, expected)
  }
})

uuid_random <- function() {
  # These uuids are pseudo-random, they are not secure!
  # xxxxxxxx-xxxx-Mxxx-Nxxx-xxxxxxxxxxxx (8-4-4-4-12)
  # The 4 bits of digit M are the UUID version,
  # and the 1 to 3 most significant bits of digit N code the UUID variant.
  digits <- floor(stats::runif(32, 0, 16))
  digits[13] <- 0x4
  digits[17] <- bitwOr(bitwAnd(digits[17], 0x3), 0x8)

  x <- format(as.hexmode(digits))
  paste(
    collapse = "",
    c(x[1:8], "-", x[9:12], "-", x[13:16], "-", x[17:20], "-", x[21:32])
  )
}

#' Run a webfakes app in another process
#'
#' Runs an app in a subprocess, using [callr::r_session].
#'
#' @param app `webfakes_app` object, the web app to run.
#' @param port Port(s) to use. By default the OS assigns a port.
#'   Add an `"s"` suffix to the port to use HTTPS. Use `"0s"` to
#'   use an OS assigned port with HTTPS. See the [how-to] on how
#'   to run the web server on multiple ports.
#' @param opts Server options. See [server_opts()] for the defaults.
#' @param start Whether to start the web server immediately. If this is
#'   `FALSE`, and `auto_start` is `TRUE`, then it is started as neeed.
#' @param auto_start Whether to start the web server process automatically.
#'   If `TRUE` and the process is not running, then `$start()`,
#'   `$get_port()`, `$get_ports()` and `$url()` start the process.
#' @param process_timeout How long to wait for the subprocess to start, in
#'   milliseconds.
#' @param callr_opts Options to pass to [callr::r_session_options()]
#'   when setting up the subprocess.
#' @return A `webfakes_app_process` object.
#'
#' ## Methods
#'
#' The `webfakes_app_process` class has the following methods:
#'
#' ```r
#' get_app()
#' get_port()
#' get_ports()
#' stop()
#' get_state()
#' local_env(envvars)
#' url(path = "/", query = NULL)
#' ```
#'
#' * `envvars`: Named list of environment variables. The `{url}` substring
#'   is replaced by the URL of the app.
#' * `path`: Path to return the URL for.
#' * `query`: Additional query parameters, a named list, to add to the URL.
#'
#' `get_app()` returns the app object.
#'
#' `get_port()` returns the (first) port the web server is running on.
#'
#' `get_ports()` returns all ports the web server is running on, and
#' whether it uses SSL on those ports, in a data frame with columns
#' `ipv4`, `ipv6`, `port` and `ssl`.
#'
#' `stop()` stops the web server, and also the subprocess. If the error
#' log file is not empty, then it dumps its contents to the screen.
#'
#' `get_state()` returns a string, the state of the web server:
#' * `"not running"` the server is not running (because it was stopped
#'   already).
#' * `"live"` means that the server is running.
#' * `"dead"` means that the subprocess has quit or crashed.
#'
#' `local_env()` sets the given environment variables for the duration of
#' the app process. It resets them in `$stop()`. Webfakes replaces `{url}`
#' in the value of the environment variables with the app URL, so you can
#' set environment variables that point to the app.
#'
#' `url()` returns the URL of the web app. You can use the `path`
#' parameter to return a specific path.
#'
#' @aliases webfakes_app_process
#' @seealso [local_app_process()] for automatically cleaning up the
#'   subprocess.
#' @export
#' @examples
#' app <- new_app()
#' app$get("/foo", function(req, res) {
#'   res$send("Hello world!")
#' })
#'
#' proc <- new_app_process(app)
#' url <- proc$url("/foo")
#' resp <- curl::curl_fetch_memory(url)
#' cat(rawToChar(resp$content))
#'
#' proc$stop()

new_app_process <- function(
  app,
  port = NULL,
  opts = server_opts(remote = TRUE),
  start = FALSE,
  auto_start = TRUE,
  process_timeout = NULL,
  callr_opts = NULL
) {
  app
  port
  opts
  start
  auto_start
  process_timeout
  callr_opts

  # WTF
  tmp <- tempfile()
  sink(tmp)
  print(app$.stack)
  sink(NULL)
  unlink(tmp)

  process_timeout <- process_timeout %||%
    as.integer(Sys.getenv(
      "R_WEBFAKES_PROCESS_TIMEOUT",
      if (Sys.getenv("R_COVR") == "true") 600000 else 5000
    ))

  self <- new_object(
    "webfakes_app_process",

    start = function() {
      self$.app <- app
      callr_opts <- do.call(callr::r_session_options, as.list(callr_opts))
      self$.process <- callr::r_session$new(callr_opts, wait = TRUE)
      self$.process$call(
        args = list(app, port, opts),
        function(app, port, opts) {
          library(webfakes)
          .GlobalEnv$app <- app
          app$listen(port = port, opts = opts)
        }
      )

      if (self$.process$poll_process(process_timeout) != "ready") {
        self$.process$kill()
        stop("webfakes app subprocess did not start :(")
      }
      msg <- self$.process$read()
      if (msg$code == 200 && !is.null(msg$error)) {
        msg$error$message <- paste0(
          "failed to start webfakes app process: ",
          msg$error$message
        )
        stop(msg$error)
      }
      if (msg$code != 301) {
        stop(
          "Unexpected message from webfakes app subprocess. ",
          "Report a bug please."
        )
      }
      self$.ports <- msg$message$ports
      self$.port <- msg$message$ports$port[1]
      self$.access_log <- msg$message$access_log
      self$.error_log <- msg$message$error_log
      self$.set_env()

      invisible(self)
    },

    get_app = function() self$.app,

    get_port = function() {
      if (self$get_state() == "not running" && auto_start) {
        self$start()
      }
      self[[".port"]]
    },

    get_ports = function() {
      if (self$get_state() == "not running" && auto_start) {
        self$start()
      }
      self[[".ports"]]
    },

    stop = function() {
      self$.reset_env()
      if (is.null(self$.process)) {
        return(invisible(self))
      }

      if (!self$.process$is_alive()) {
        status <- self$.process$get_exit_status()
        out <- err <- NULL
        try_silently(out <- self$.process$read_output())
        try_silently(err <- self$.process$read_error())
        cat0("webfakes process dead, exit code: ", status, "\n")
        if (!is.null(out)) {
          cat0("stdout:", out, "\n")
        }
        if (!is.null(err)) cat0("stderr:", err, "\n")
      }

      try_silently(self$.process$close(grace = 0))
      try_silently(self$.process$read())
      self$.print_errors()
      self$.process <- NULL
      invisible(self)
    },

    get_state = function() {
      if (is.null(self$.process)) {
        "not running"
      } else if (self$.process$is_alive()) {
        "live"
      } else {
        "dead"
      }
    },

    local_env = function(envvars) {
      if (!is.null(self$.process)) {
        local <- unlist(envvars)
        local[] <- gsub("{url}", self$url(), local, fixed = TRUE)
        self$.old_env <- c(self$.old_env, set_envvar(local))
      } else {
        self$.local_env <- utils::modifyList(
          as.list(self$.local_env),
          as.list(envvars)
        )
      }
      invisible(self)
    },

    .set_env = function() {
      local <- unlist(self$.local_env)
      local[] <- gsub("{url}", self$url(), local, fixed = TRUE)
      self$.old_env <- set_envvar(local)
      invisible(self)
    },

    .reset_env = function() {
      if (!is.null(self$.old_env)) {
        set_envvar(self$.old_env)
        self$.old_env <- NULL
      }
    },

    url = function(
      path = "/",
      query = NULL,
      https = FALSE,
      domain = "127.0.0.1"
    ) {
      if (self$get_state() == "not running" && auto_start) {
        self$start()
      }
      if (!is.null(query)) {
        query <- paste0("?", paste0(names(query), "=", query, collapse = "&"))
      }
      port <- if (https) {
        if (!any(self$.ports$ssl)) {
          stop("Web server does support HTTPS")
        }
        self$.ports$port[self$.ports$ssl][1]
      } else {
        self$.port
      }
      proto <- if (self$.ports$ssl[match(port, self$.ports$port)]) {
        "https"
      } else {
        "http"
      }
      paste0(proto, "://", domain, ":", port, path, query)
    },

    .process = NULL,
    .app = NULL,
    .ports = NULL,
    .port = NULL,
    .old_env = NULL,
    .access_log = NA_character_,
    .error_log = NA_character_,
    .auto_start = auto_start,

    .print_errors = function() {
      if (
        !is.na(self$.error_log) &&
          file.exists(self$.error_log) &&
          file.info(self$.error_log)$size > 0
      ) {
        err <- readLines(self$.error_log, warn = FALSE)
        cat("webfakes web server errors:\n")
        cat(err, sep = "\n")
      }
    }
  )

  reg.finalizer(self, function(x) x$.reset_env())

  if (start) {
    self$start()
  }

  self
}

test_that("parallel requests", {
  url <- httpbin$url("/delay/0.5")
  p <- curl::new_pool()
  handles <- replicate(3, curl::new_handle(url = url, http_version = 2))
  resps <- list()
  for (handle in handles) {
    curl::multi_add(
      handle,
      done = function(x) resps <<- c(resps, list(x)),
      fail = stop,
      pool = p
    )
  }
  st <- system.time(curl::multi_run(timeout = 5, pool = p))
  expect_true(st[["elapsed"]] >= 0.5)
  expect_true(st[["elapsed"]] < 1.5)
  expect_equal(resps[[1]]$status_code, 200L)
  expect_equal(resps[[2]]$status_code, 200L)
  expect_equal(resps[[3]]$status_code, 200L)
})

test_that("error if cannot start", {
  # does not start before the timeout
  app <- new_app()
  app$listen <- function(...) Sys.sleep(1)
  expect_snapshot(
    error = TRUE,
    new_app_process(app, process_timeout = 100, start = TRUE)
  )

  # errors before/while starting
  app <- new_app()
  app$listen <- function(...) stop("oops")
  expect_snapshot(
    error = TRUE,
    new_app_process(app, start = TRUE)
  )

  # sends a different message first
  app <- new_app()
  app$listen <- function(...) "foobar"
  expect_snapshot(
    error = TRUE,
    new_app_process(app, start = TRUE)
  )
})

test_that("get_state", {
  app <- new_app()
  on.exit(proc$stop(), add = TRUE)
  proc <- new_app_process(app, start = TRUE)
  expect_equal(proc$get_state(), "live")
  proc$.process$kill()
  expect_equal(proc$get_state(), "dead")
  expect_output(proc$stop(), "webfakes process dead")
  expect_equal(proc$get_state(), "not running")
})

test_that("env vars", {
  app <- new_app()
  on.exit(proc$stop(), add = TRUE)
  withr::local_envvar(list(FOO = "foo"))
  withr::local_envvar(list(BAR = NA_character_))
  proc <- new_app_process(app, start = TRUE)
  proc$local_env(list(FOO = "bar"))
  proc$local_env(list(BAR = "{url}"))
  expect_equal(Sys.getenv("FOO", ""), "bar")
  expect_equal(Sys.getenv("BAR"), proc$url())
  proc$stop()
  expect_equal(Sys.getenv("FOO", ""), "foo")
  expect_equal(Sys.getenv("BAR", ""), "")
})

test_that("env vars 2", {
  app <- new_app()
  on.exit(proc$stop(), add = TRUE)
  withr::local_envvar(list(FOO = "foo"))
  withr::local_envvar(list(BAR = NA_character_))
  proc <- new_app_process(app)
  proc$local_env(list(FOO = "bar"))
  proc$local_env(list(BAR = "{url}"))
  proc$start()
  expect_equal(Sys.getenv("FOO", ""), "bar")
  expect_equal(Sys.getenv("BAR"), proc$url())
  proc$stop()
  expect_equal(Sys.getenv("FOO", ""), "foo")
  expect_equal(Sys.getenv("BAR", ""), "")
})

test_that("app from setup", {
  # This app was created in setup.R
  url <- httpbin2$url("/get", query = c(q1 = "one", q2 = "two"))
  handle <- curl::new_handle()
  curl::handle_setheaders(handle, "foo" = "bar")
  resp <- curl::curl_fetch_memory(url, handle = handle)
  expect_equal(resp$status_code, 200L)
})

#' Middleware that add an `ETag` header to the response
#'
#' If the response already has an `ETag` header, then it is kept.
#'
#' This middleware handles the `If-None-Match` headers, and it sets the
#' status code of the response to 304 if `If-None-Match` matches the
#' `ETag`. It also removes the response body in this case.
#'
#' @param algorithm Checksum algorithm to use. Only `"crc32"` is
#' implemented currently.
#'
#' @return Handler function.
#'
#' @family middleware
#' @export
#' @examples
#' app <- new_app()
#' app$use(mw_etag())
#' app

mw_etag <- function(algorithm = "crc32") {
  if (algorithm != "crc32") {
    stop("Only the 'crc32' algorithm is implemented in `mw_etag()`")
  }
  function(req, res) {
    do <- function(req, res) {
      etag <- res$get_header("ETag")
      if (is.null(etag)) {
        etag <- paste0("\"", crc32(res$.body), "\"")
        res$set_header("ETag", etag)
      }
      req_etag <- req$get_header("If-None-Match")
      if (!is.null(req_etag) && req_etag == etag) {
        res$.body <- NULL
        res$set_status(304)
      }
    }
    res$on_response(do)

    "next"
  }
}

#! /bin/bash

# https://transang.me/create-a-multiple-domains-self-signed-ssl-certificate-with-testing-scripts/
# https://github.com/civetweb/civetweb/blob/master/docs/OpenSSL.md

openssl genrsa -out server.key 4096
openssl req \
	-new \
	-key server.key \
	-nodes \
	-out server.csr \
	-subj "/C=ES/ST=Barcelona/L=Barcelona/O=webfakes.r-lib.org/CN=localhost" \
	-addext "subjectAltName=IP:127.0.0.1,DNS:localhost,DNS:localhost.localdomain"

openssl genrsa -out ca.key 4096
openssl req \
	-new \
	-x509 \
	-nodes \
	-days 36500 \
	-key ca.key \
	-out ca.crt \
	-subj "/C=ES/ST=Barcelona/L=Barcelona/O=webfakes.r-lib.org/OU=CA/CN=ca.webfakes.r-lib.org"

openssl x509 \
	-req \
	-in server.csr \
	-CAkey ca.key \
	-CA ca.crt \
	-set_serial -01 \
	-out server.crt \
	-days 36500 \
	-sha256 \
	-extfile <(printf "subjectAltName=IP:127.0.0.1,DNS:localhost,DNS:localhost.localdomain")

openssl verify -CAfile ca.crt server.crt

cp server.crt server.pem
cat server.key >> server.pem

See https://github.com/Redocly/redoc#readme for Redoc.

Redoc authors:

* Roman Hotsiy <gotsijroman@gmail.com>
* Cesar <cesarlevel@gmail.com>
* Dimitar Nanov <dimitar.nanov@ecollect.net>
* Ivan Goncharov <ivan.goncharov.ua@gmail.com>
* Anya Stasiuk <stasiukanya@gmail.com>
* Bohdan Khorolets <bogdan@khorolets.com>
* Brendan Abbott <brendan@bloodbone.ws>
* Ben Firshman <ben@firshman.co.uk>
* brushmate <brushmate@gmail.com>
* Adam Altman <adam@ilabsinc.com>
* Jérémy Derussé <jeremy@derusse.com>
* Matthias Mohr <webmaster@mamo-net.de>
* Mike Stead <mike.stead@massive.co>
* kedashoe <kwallace@gmail.com>
* leoliu <leoliu@yunify.com>
* amanganiello <a.manganiello@emminformatica.it>
* Patrick Elam <40776618+patrickelam@users.noreply.github.com>
* Mohamed Zenadi <zeapo@users.noreply.github.com>
* Sergey Dubovyk <knidarkness@gmail.com>
* Zakary Kamal Ismail <zakary.kamal.fs@outlook.com>
* Jon Nicholson <drjonnicholson@users.noreply.github.com>
* Alex <alexgmin@gmail.com>
* TATSUNO Yasuhiro <tatsuno_yasuhiro@kurusugawa.jp>
* Oleksiy Kachynskyy <okachynskyy@users.noreply.github.com>
* Mike Ralphson <mike.ralphson@gmail.com>
* Phil Sturgeon <philsturgeon@users.noreply.github.com>
* Dimitar Nanov <mitko.n@gmail.com>
* Faheem Abrar <faheem.abrar@outlook.com>
* Fredrik Lengstrand <10126466+fredriklengstrand@users.noreply.github.com>
* Gaurav Jain <gaurav@gauravjain.org>
* George Wilson <georgejameswilson@googlemail.com>
* GreenHedgehog <IrishGreenHedgehog@gmail.com>
* Homa Wong <homawong@gmail.com>
* Ingo Claro <miclaro@gmail.com>
* Jacob Baskin <jacob.baskin@gmail.com>
* Jean-Daniel <Jean-Daniel@users.noreply.github.com>
* Joe Honzawa <goflb.jh@gmail.com>
* Josh Price <josh@alembic.com.au>
* Julian <2564520+julmuell@users.noreply.github.com>
* Julian Kühnel <me@juliankuehnel.com>
* Julien Feltesse <jfeltesse@mdsol.com>
* Khoa Tran <khoatran@temando.com>
* Kris Kalavantavanich <krismath.K@gmail.com>
* Kryštof Korb <krystof@korb.cz>
* LeFnord <pscholz.le@gmail.com>
* Leonard Ehrenfried <leonard.ehrenfried@gmail.com>
* Lev Pachmanov <31389480+levpachmanov@users.noreply.github.com>
* Luciano Jr <luciano@lucianojr.com.br>
* Luigi Pinca <luigipinca@gmail.com>
* Making GitHub Delicious <iron@waffle.io>
* Marius Rumpf <MariusRumpf@users.noreply.github.com>
* Mason Malone <mason.malone@gmail.com>
* Mathias Schreck <schreck.mathias@googlemail.com>
* Melvyn Sopacua <40824897+mes3yd@users.noreply.github.com>
* Melvyn Sopacua <mes@3yourmind.com>
* Michael Huynh <43751307+miqh@users.noreply.github.com>
* Morgan Terry <localstatic@users.noreply.github.com>
* Nan Yan <625518543@qq.com>
* Nick Oliver <github@pixnbits.org>
* Oleg Vaskevich <oleg@osv.im>
* Patrick Niklaus <github@execfoo.de>
* Patrick Rodacker <lordrhodos@users.noreply.github.com>
* Pete Nykänen <pete.a.nykanen@gmail.com>
* Peter Golm <golm.peter@gmail.com>
* Peter Wessels <pjotrwes@hotmail.com>
* Petr Flaks <flaks-petr@protonmail.com>
* Primož Pincolič <primozweb@gmail.com>
* Quinn Blenkinsop <quinn@qw-in.com>
* Robert DeRose <RobertDeRose@users.noreply.github.com>
* Sander van de Graaf <sander@vandegraaf.nl>
* Sebastián Ramírez <tiangolo@gmail.com>
* Sergio Regueira <sergio@sergioregueira.com>
* SeungWoon Maz Lee <mazicky@gmail.com>
* Sheila Kelly <36901025+viralanomaly@users.noreply.github.com>
* Siarhei Bautrukevich <bautrukevich@users.noreply.github.com>
* Skyler Lewis <sblnog@gmail.com>
* SoftBrix <andreas@softbrix.se>
* Sérgio Ramos <mail@sergioramos.me>
* Tim Hordern <tim.hordern@gmail.com>
* Vincent Giersch <vincent@giersch.fr>
* Vladimir L <vladimir.lioubitelev@gmail.com>
* William Boman <william+github@redwill.se>
* Zach Pomerantz <zmp@umich.edu>
* bwjohnson-ss <41123899+bwjohnson-ss@users.noreply.github.com>
* davchen51 <46250804+davchen51@users.noreply.github.com>
* duxiaofeng <duxiaofeng-github@users.noreply.github.com>
* fritz-c <fritz-c@users.noreply.github.com>
* jsmartfo <james.jerome.smartfoster@aexp.com>
* kmbenitez <kathleen.benitez@gmail.com>
* lrobledo <lrobledo@cmsonline.com>
* lscholten <info@luukscholten.com>
* mknoszlig <maximilian.karasz@googlemail.com>
* neumond <knifeslaughter@gmail.com>
* pengfluf <pengfluf@tutanota.com>
* russellrobinson <russellrobinson@users.noreply.github.com>
* tomjankes <tomjankes@users.noreply.github.com>
* torbenw <torbenw@users.noreply.github.com>
* travis@localhost <travis@localhost>
* unarist <m.unarist@gmail.com>
* zhzxang <xuzhang3371@163.com>
* Adam Altman <adam@redoc.ly>
* Adam DuVander <adamd@users.noreply.github.com>
* Adrien Gallou <adriengallou@gmail.com>
* Aleksandr Karo <alex@karo-dev.ru>
* Alex Scammon <stackedsax@users.noreply.github.com>
* Anastasiya Mashoshyna <a.mashoshyna@gmail.com>
* Andrew Berry <deviantintegral@gmail.com>
* Andrew Zhukevych <34597767+Hollister009@users.noreply.github.com>
* Andrii Tykhan <andriytixan@gmail.com>
* Antherkiv <antherkiv@gmail.com>
* Anthony Porthouse <anthony@porthou.se>
* Anto <antograssiot@free.fr>
* Anton Komarev <1849174+antonkomarev@users.noreply.github.com>
* Benny Neugebauer <bn@bennyn.de>
* Blake Erickson <o.blakeerickson@gmail.com>
* Brendan Abbott <brendan.abbott@temando.com>
* Cesar Landeros Delgado <chichozell@gmail.com>
* Cesar Level <cesarlevel@gmail.com>
* Chris Faulkner <thefaulkner@gmail.com>
* Cédric Bertolini <41571181+wizacedric@users.noreply.github.com>
* Cédric Fabianski <cfabianski@me.com>
* Daniel Chao <daniel.h.chao@gmail.com>
* Dave Oram <david.oram99@gmail.com>
* David Beacham <dbeacham@dbeacham.co.uk>
* David Cumps <CumpsD@users.noreply.github.com>
* David J. Felix <felix.davidj@gmail.com>
* David Revay <MrBlenny@users.noreply.github.com>
* DeBr0glie <afrimuchkov@yandex.ru>

The authors were extracted from the git history:
```sh
git clone https://github.com/Redocly/redoc.git
cd redoc
git shortlog --summary --numbered --email --all | cut -f2 | sed 's/^/* /'
```

The webfakes package comes with two fake apps that
allow to imitate the OAuth2.0 flow in your test cases.
(See [Aaron Parecki's tutorial](https://aaronparecki.com/oauth-2-simplified/) for a good introduction to OAuth2.0.)
One app (`oauth2_resource_app()`) is the API server that serves both as the resource and provides authorization.
`oauth2_third_party_app()` plays the role of the third-party app.
They are useful when testing or demonstrating code handling OAuth2.0 authorization, token caching, etc. in a package.
The apps can be used in your tests directly,
or you could adapt one or both of them to better mimic a particular OAuth2.0 flow.

# parse_range

    Code
      parse_range("foobar=1-100")
    Output
      NULL
    Code
      parse_range("bytes=0-100, 50-150")
    Output
      NULL
    Code
      parse_range("bytes=200-100")
    Output
      NULL
    Code
      parse_range("bytes=x-100")
    Output
      NULL
    Code
      parse_range("bytes=1-100")
    Output
           [,1] [,2]
      [1,]    1  100
    Code
      parse_range("bytes=1-")
    Output
           [,1] [,2]
      [1,]    1  Inf
    Code
      parse_range("bytes=-100")
    Output
           [,1] [,2]
      [1,]    0 -100
    Code
      parse_range("bytes=0-100, 200-")
    Output
           [,1] [,2]
      [1,]    0  100
      [2,]  200  Inf
    Code
      parse_range("bytes=200-300, 0-100")
    Output
           [,1] [,2]
      [1,]    0  100
      [2,]  200  300

# invalid handler

    Code
      app$use("foobar")
    Condition
      Error in `parse_handlers()`:
      ! Invalid webfakes handler
    Code
      app$get("/foo", 1:100)
    Condition
      Error in `parse_handlers()`:
      ! Invalid webfakes handler

# error if cannot start

    Code
      new_app_process(app, process_timeout = 100, start = TRUE)
    Condition
      Error in `self$start()`:
      ! webfakes app subprocess did not start :(

---

    Code
      new_app_process(app, start = TRUE)
    Condition
      Error:
      ! ! failed to start webfakes app process: in callr subprocess.
      Caused by error in `app$listen(port = port, opts = opts)`:
      ! oops

---

    Code
      new_app_process(app, start = TRUE)
    Condition
      Error in `self$start()`:
      ! Unexpected message from webfakes app subprocess. Report a bug please.

# git_app

    Code
      system(paste("git ls-remote", git$url("/pak-test.git")), intern = TRUE)
    Output
      [1] "3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b\tHEAD"             
      [2] "3f3b0b4ee8a0ff4563073924e5fe069da67a6d8b\trefs/heads/main"  
      [3] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e\trefs/heads/subdir"
      [4] "cefdc0eebcd7f757efb9a80652fd8aaf1a87508e\trefs/tags/v1"     

# /cookies/set/:name/:value

    Code
      curl::handle_cookies(handle)
    Output
           domain  flag path secure expiration name value
      1 127.0.0.1 FALSE    /  FALSE        Inf  foo   bar

# /cookies/set

    Code
      curl::handle_cookies(handle)
    Output
           domain  flag path secure expiration name value
      1 127.0.0.1 FALSE    /  FALSE        Inf  foo   bar
      2 127.0.0.1 FALSE    /  FALSE        Inf  bar   baz

# /cookies/delete

    Code
      curl::handle_cookies(handle)
    Output
           domain  flag path secure expiration name value
      1 127.0.0.1 FALSE    /  FALSE        Inf  bar   baz

# webfakes_app_process

    Code
      proc
    Output
      <webfakes_app_process>
      state:
        not running
      auto_start:
        TRUE
      process id:
        none
      http url:
        NA
      fields and methods:
        get_app()              # get the app object
        get_port()             # query (first) port of the app
        get_ports()            # query all ports of the app
        get_state()            # query web server process state
        local_env(envvars)     # set temporary environment variables
        start()                # start the app
        url(path, query)       # query url for an api path
        stop()                 # stop web server process
      # see ?webfakes_app_process for details

# parse_query

    Code
      parse_query("foo")
    Output
      $foo
      [1] ""
      
    Code
      parse_query("?foo")
    Output
      $foo
      [1] ""
      
    Code
      parse_query("?foo&bar")
    Output
      $foo
      [1] ""
      
      $bar
      [1] ""
      
    Code
      parse_query("?foo&bar=baz")
    Output
      $foo
      [1] ""
      
      $bar
      [1] "baz"
      

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/webfakes-package.R
\docType{package}
\name{webfakes-package}
\alias{webfakes}
\alias{webfakes-package}
\title{webfakes: Fake Web Apps for HTTP Testing}
\description{
Create a web app that makes it easier to test web clients without using the internet. It includes a web app framework with path matching, parameters and templates. Can parse various 'HTTP' request bodies. Can send 'JSON' data or files from the disk. Includes a web app that implements the 'httpbin.org' web service.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://webfakes.r-lib.org/}
  \item \url{https://github.com/r-lib/webfakes}
  \item Report bugs at \url{https://github.com/r-lib/webfakes/issues}
}

}
\author{
\strong{Maintainer}: Gábor Csárdi \email{csardi.gabor@gmail.com}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
  \item  Civetweb contributors (see inst/credits/ciwetweb.md) [contributor]
  \item  Redoc contributors (see inst/credits/redoc.md) [contributor]
  \item L. Peter Deutsch (src/md5.h) [contributor]
  \item Martin Purschke (src/md5.h) [contributor]
  \item  Aladdin Enterprises (src/md5.h) [copyright holder]
  \item Maëlle Salmon \email{maelle.salmon@yahoo.se} (\href{https://orcid.org/0000-0002-2815-0399}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-etag.R
\name{mw_etag}
\alias{mw_etag}
\title{Middleware that add an \code{ETag} header to the response}
\usage{
mw_etag(algorithm = "crc32")
}
\arguments{
\item{algorithm}{Checksum algorithm to use. Only \code{"crc32"} is
implemented currently.}
}
\value{
Handler function.
}
\description{
If the response already has an \code{ETag} header, then it is kept.
}
\details{
This middleware handles the \code{If-None-Match} headers, and it sets the
status code of the response to 304 if \code{If-None-Match} matches the
\code{ETag}. It also removes the response body in this case.
}
\examples{
app <- new_app()
app$use(mw_etag())
app
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_json}()},
\code{\link{mw_log}()},
\code{\link{mw_multipart}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app.R
\name{new_app}
\alias{new_app}
\alias{webfakes_app}
\title{Create a new web application}
\usage{
new_app()
}
\value{
A new \code{webfakes_app}.
}
\description{
Create a new web application
}
\details{
The typical workflow of creating a web application is:
\enumerate{
\item Create a \code{webfakes_app} object with \code{new_app()}.
\item Add middleware and/or routes to it.
\item Start is with the \code{webfakes_app$listen()} method, or start it in
another process with \code{\link[=new_app_process]{new_app_process()}}.
\item Make queries to the web app.
\item Stop it via \code{CTRL+C} / \code{ESC}, or, if it is running in another
process, with the \verb{$stop()} method of \code{\link[=new_app_process]{new_app_process()}}.
}

A web application can be
\itemize{
\item restarted,
\item saved to disk,
\item copied to another process using the callr package, or a similar way,
\item embedded into a package,
\item extended by simply adding new routes and/or middleware.
}

The webfakes API is very much influenced by the
\href{https://expressjs.com/}{express.js} project.
\subsection{Create web app objects}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{new_app()
}\if{html}{\out{</div>}}

\code{new_app()} returns a \code{webfakes_app} object the has the methods listed
on this page.

An app is an environment with S3 class \code{webfakes_app}.
}

\subsection{The handler stack}{

An app has a stack of handlers. Each handler can be a route or
middleware. The differences between the two are:
\itemize{
\item A route is bound to one or more paths on the web server. Middleware
is not (currently) bound to paths, but run for all paths.
\item A route is usually (but not always) the end of the handler stack for
a request. I.e. a route takes care of sending out the response to
the request. Middleware typically performs some action on the request
or the response, and then the next handler in the stack is invoked.
}
}

\subsection{Routes}{

The following methods define routes. Each method corresponds to the
HTTP verb with the same name, except for \code{app$all()}, which creates a
route for all HTTP methods.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$all(path, ...)
app$delete(path, ...)
app$get(path, ...)
app$head(path, ...)
app$patch(path, ...)
app$post(path, ...)
app$put(path, ...)
... (see list below)
}\if{html}{\out{</div>}}
\itemize{
\item \code{path} is a path specification, see 'Path specification' below.
\item \code{...} is one or more handler functions. These will be placed in the
handler stack, and called if they match an incoming HTTP request.
See 'Handler functions' below.
}

webfakes also has methods for the less frequently used HTTP verbs:
\code{CONNECT}, \code{MKCOL}, \code{OPTIONS}, \code{PROPFIND}, \code{REPORT}. (The method
names are always in lowercase.)

If a request is not handled by any routes (or handler functions in
general), then webfakes will send a simple HTTP 404 response.
}

\subsection{Middleware}{

\code{app$use()} adds a middleware to the handler stack. A middleware is
a handler function, see 'Handler functions' below. webfakes comes with
middleware to perform common tasks:
\itemize{
\item \code{\link[=mw_cookie_parser]{mw_cookie_parser()}} parses \code{Cookie} headers.
\item \code{\link[=mw_etag]{mw_etag()}} adds an \code{ETag} header to the response.
\item \code{\link[=mw_json]{mw_json()}} parses JSON request bodies.
\item \code{\link[=mw_log]{mw_log()}} logs each requests to standard output, or another connection.
\item \code{\link[=mw_multipart]{mw_multipart()}} parses multipart request bodies.
\item \code{\link[=mw_range_parser]{mw_range_parser()}} parses \code{Range} headers.
\item \code{\link[=mw_raw]{mw_raw()}} parses raw request bodies.
\item \code{\link[=mw_static]{mw_static()}} serves static files from a directory.
\item \code{\link[=mw_text]{mw_text()}} parses plain text request bodies.
\item \code{\link[=mw_urlencoded]{mw_urlencoded()}} parses URL encoded request bodies.
}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$use(..., .first = FALSE)
}\if{html}{\out{</div>}}
\itemize{
\item \code{...} is a set of (middleware) handler functions. They are added to
the handler stack, and called for every HTTP request. (Unless an HTTP
response is created before reaching this point in the handler stack.)
\item \code{.first} set to \code{TRUE} is you want to add the handler function
to the bottom of the stack.
}
}

\subsection{Handler functions}{

A handler function is a route or middleware. A handler function is
called by webfakes with the incoming HTTP request and the outgoing
HTTP response objects (being built) as arguments. The handler function
may query and modify the members of the request and/or the response
object. If it returns the string \code{"next"}, then it is \emph{not} a terminal
handler, and once it returns, webfakes will move on to call the next
handler in the stack.

A typical route:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$get("/user/:id", function(req, res) \{
  id <- req$params$id
  ...
  res$
    set_status(200L)$
    set_header("X-Custom-Header", "foobar")$
    send_json(response, auto_unbox = TRUE)
\})
}\if{html}{\out{</div>}}
\itemize{
\item The handler belongs to an API path, which is a wildcard path in
this case. It matches \verb{/user/alice}, \verb{/user/bob}, etc. The handler
will be only called for GET methods and matching API paths.
\item The handler receives the request (\code{req}) and the response (\code{res}).
\item It sets the HTTP status, additional headers, and sends the data.
(In this case the \code{webfakes_response$send_json()} method automatically
converts \code{response} to JSON and sets the \code{Content-Type} and
\code{Content-Length} headers.
\item This is a terminal handler, because it does \emph{not} return \code{"next"}.
Once this handler function returns, webfakes will send out the HTTP
response.
}

A typical middleware:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$use(function(req, res) \{
  ...
  "next"
\})
}\if{html}{\out{</div>}}
\itemize{
\item There is no HTTP method and API path here, webfakes will call the
handler for each HTTP request.
\item This is not a terminal handler, it does return \code{"next"}, so after it
returns webfakes will look for the next handler in the stack.
}
}

\subsection{Errors}{

If a handler function throws an error, then the web server will return
a HTTP 500 \code{text/plain} response, with the error message as the
response body.
}

\subsection{Request and response objects}{

See \link{webfakes_request} and \link{webfakes_response} for the methods of the
request and response objects.
}

\subsection{Path specification}{

Routes are associated with one or more API paths. A path specification
can be
\itemize{
\item A "plain" (i.e. without parameters) string. (E.g. \code{"/list"}.)
\item A parameterized string. (E.g. \code{"/user/:id"}.)
\item A regular expression created via \code{\link[=new_regexp]{new_regexp()}} function.
\item A list or character vector of the previous ones. (Regular expressions
must be in a list.)
}
}

\subsection{Path parameters}{

Paths that are specified as parameterized strings or regular expressions
can have parameters.

For parameterized strings the keys may contain letters, numbers and
underscores. When webfakes matches an API path to a handler with a
parameterized string path, the parameters will be added to the
request, as \code{params}. I.e. in the handler function (and subsequent
handler functions, if the current one is not terminal), they are
available in the \code{req$params} list.

For regular expressions, capture groups are also added as parameters.
It is best to use named capture groups, so that the parameters are in
a named list.

If the path of the handler is a list of parameterized strings or
regular expressions, the parameters are set according to the first
matching one.
}

\subsection{Templates}{

webfakes supports templates, using any template engine. It comes with
a template engine that uses the glue package, see \code{\link[=tmpl_glue]{tmpl_glue()}}.

\code{app$engine()} registers a template engine, for a certain file
extension. The \verb{$render()} method of \link{webfakes_response}
can be called from the handler function to evaluate a template from a
file.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$engine(ext, engine)
}\if{html}{\out{</div>}}
\itemize{
\item \code{ext}: the file extension for which the template engine is added.
It should not contain the dot. E.g. \verb{"html"', }"brew"`.
\item \code{engine}: the template engine, a function that takes the file path
(\code{path}) of the template, and a list of local variables (\code{locals})
that can be used in the template. It should return the result.
}

An example template engine that uses glue might look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$engine("txt", function(path, locals) \{
  txt <- readChar(path, nchars = file.size(path))
  glue::glue_data(locals, txt)
\})
}\if{html}{\out{</div>}}

(The built-in \code{\link[=tmpl_glue]{tmpl_glue()}} engine has more features.)

This template engine can be used in a handler:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$get("/view", function(req, res) \{
 txt <- res$render("test")
 res$
   set_type("text/plain")$
   send(txt)
\})
}\if{html}{\out{</div>}}

The location of the templates can be set using the \code{views} configuration
parameter, see the \verb{$set_config()} method below.

In the template, the variables passed in as \code{locals}, and also the
response local variables (see \code{locals} in \link{webfakes_response}), are
available.
}

\subsection{Starting and stopping}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$listen(port = NULL, opts = server_opts(), cleanup = TRUE)
}\if{html}{\out{</div>}}
\itemize{
\item \code{port}: port to listen on. When \code{NULL}, the operating system will
automatically select a free port. Add an \code{"s"} suffix to the port
to use HTTPS. Use \code{"0s"} to use an OS assigned port with HTTPS.
See the \link{how-to} manual page if you want to start the web server
on more than one ports.
\item \code{opts}: options to the web server. See \code{\link[=server_opts]{server_opts()}} for the
list of options and their default values.
\item \code{cleanup}: stop the server (with an error) if the standard input
of the process is closed. This is handy when the app runs in a
\code{callr::r_session} subprocess, because it stops the app (and the
subprocess) if the main process has terminated.
}

This method does not return, and can be interrupted with \code{CTRL+C} / \code{ESC}
or a SIGINT signal. See \code{\link[=new_app_process]{new_app_process()}} for interrupting an app that
is running in another process.

When \code{port} is \code{NULL}, the operating system chooses a port where the
app will listen. To be able to get the port number programmatically,
before the listen method blocks, it advertises the selected port in a
\code{webfakes_port} condition, so one can catch it:

webfakes by default binds only to the loopback interface at 127.0.0.1, so
the webfakes web app is never reachable from the network.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{withCallingHandlers(
  app$listen(),
  "webfakes_port" = function(msg) print(msg$port)
)
}\if{html}{\out{</div>}}
}

\subsection{Logging}{

webfakes can write an access log that contains an entry for all incoming
requests, and also an error log for the errors that happen while
the server is running. This is the default behavior for local app
(the ones started by \code{app$listen()} and for remote apps (the ones
started via \code{new_app_process()}:
\itemize{
\item Local apps do not write an access log by default.
\item Remote apps write an access log into the
\verb{<tmpdir>/webfakes/<pid>/access.log} file, where \verb{<tmpdir>} is the
session temporary directory of the \emph{main process}, and \verb{<pid>} is
the process id of the \emph{subprocess}.
\item Local apps write an error log to \verb{<tmpdir>/webfakes/error.log}, where
\verb{<tmpdir>} is the session temporary directory of the current process.
\item Remote app write an error log to the \verb{<tmpdir>/webfakes/<pid>/error.log},
where \verb{<tmpdir>} is the session temporary directory of the
\emph{main process} and \verb{<pid>} is the process id of the \emph{subprocess}`.
}

See \code{\link[=server_opts]{server_opts()}} for changing the default logging behavior.
}

\subsection{Shared app data}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$locals
}\if{html}{\out{</div>}}

It is often useful to share data between handlers and requests in an
app. \code{app$locals} is an environment that supports this. E.g. a
middleware that counts the number of requests can be implemented as:

\if{html}{\out{<div class="sourceCode">}}\preformatted{app$use(function(req, res) \{
  locals <- req$app$locals
  if (is.null(locals$num)) locals$num <- 0L
  locals$num <- locals$num + 1L
  "next"
\})
}\if{html}{\out{</div>}}

\link{webfakes_response} objects also have a \code{locals} environment, that is
initially populated as a copy of \code{app$locals}.
}

\subsection{Configuration}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app$get_config(key)
app$set_config(key, value)
}\if{html}{\out{</div>}}
\itemize{
\item \code{key}: configuration key.
\item \code{value}: configuration value.
}

Currently used configuration values:
\itemize{
\item \code{views}: path where webfakes searches for templates.
}
}
}
\examples{
# see example web apps in the `/examples` directory in
system.file(package = "webfakes", "examples")

app <- new_app()
app$use(mw_log())

app$get("/hello", function(req, res) {
  res$send("Hello there!")
})

app$get(new_regexp("^/hi(/.*)?$"), function(req, res) {
  res$send("Hi indeed!")
})

app$post("/hello", function(req, res) {
  res$send("Got it, thanks!")
})

app

# Start the app with: app$listen()
# Or start it in another R session: new_app_process(app)
}
\seealso{
\link{webfakes_request} for request objects, \link{webfakes_response} for
response objects.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-text.R
\name{mw_text}
\alias{mw_text}
\title{Middleware to parse a plain text body}
\usage{
mw_text(default_charset = "utf-8", type = "text/plain")
}
\arguments{
\item{default_charset}{Encoding to set on the text.}

\item{type}{Content type to match before parsing. If it does not
match, then the request object is not modified.}
}
\value{
Handler function.
}
\description{
Adds the parsed object as the \code{text} element of the request object.
}
\examples{
app <- new_app()
app$use(mw_text())
app
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_json}()},
\code{\link{mw_log}()},
\code{\link{mw_multipart}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{how-to}
\alias{how-to}
\title{How to use webfakes in your tests}
\description{
How to use webfakes in your tests
}
\section{How to use webfakes in your tests}{

\subsection{How do I use webfakes in my package?}{

First, you need to add webfakes to the \code{DESCRIPTION} file of your package.
Use the \code{Suggests} field, as webfakes is only needed for testing:

\if{html}{\out{<div class="sourceCode">}}\preformatted{...
Suggests:
  webfakes,
  testthat
...
}\if{html}{\out{</div>}}

Then, unless the URL to the web service is an argument of your package functions, you might need to tweak your package code slightly to make sure every call to a real web service can be targeted at another URL instead (of a fake app).
See next subsection.

Last but not least, you need to decide if you want a single web app for all your test cases.
The alternative is to use different apps for some or all test files.
Occasionally you may want to use a special app for a single test case.
Each app runs in a new subprocess, and it takes typically about 100-400ms to start.

See the sections later on on writing tests with a single app or multiple apps.
}

\subsection{How do I make my app connect to webfakes when the tests are running?}{

In the typical scenario, you want your package to connect to the test app only when running the tests.
If the URL to the web service is not an argument of the functions, one way to achieve this is to allow specifying the web server URL(s) via environment variables.
E.g. when writing a GitHub API client, your package can check use \code{GITHUB_URL} environment variable.

E.g.

\if{html}{\out{<div class="sourceCode {.r}">}}\preformatted{service_url <- function() \{
    Sys.getenv("GITHUB_URL", "https://api.github.com")
\}

# rest of the package code

foobar <- function() \{
    httr::GET(service_url())
\}
}\if{html}{\out{</div>}}

When this is not set, the package connects to the proper GitHub API. When testing, you can point it to your test app.

\code{new_app_process()} helps you setting up temporary environment variables.
These are active while the process is running, and they are removed or reset in \verb{$stop()}.
For example:

In \verb{$local_env()} environment variables, webfakes replaces \code{{url}} with the actual app URL.
This is needed by default, because the web server process starts up only later, so the URL is not known yet.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{http <- webfakes::local_app_process(webfakes::httpbin_app(), start = TRUE)
http$local_env(list(GITHUB_API = "\{url\}"))
Sys.getenv("GITHUB_API")
#> [1] "http://127.0.0.1:64362/"
http$stop()
Sys.getenv("GITHUB_API")
#> [1] ""
}\if{html}{\out{</div>}}
}

\subsection{How can I write my own app?}{

You create a new app with \code{new_app()}.
This returns an object with methods to add middleware and API
endpoints to it.
For example, a simple app that returns the current time in JSON would
look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{time <- webfakes::new_app()
time$get("/time", function(req, res) \{
  res$send_json(list(time = format(Sys.time())), auto_unbox = TRUE)
\})
}\if{html}{\out{</div>}}

Now you can start this app on a random port using \code{web$listen()}.
Alternatively, you can start it in a subprocess with \code{new_app_process()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web <- webfakes::new_app_process(time)
web$url()
#> [1] "http://127.0.0.1:64364/"
}\if{html}{\out{</div>}}

Use \code{web$url()} to query the URL of the app. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{url <- web$url("/time")
httr::content(httr::GET(url))
#> $time
#> [1] "2025-01-14 10:07:38"
}\if{html}{\out{</div>}}

\code{web$stop()} stops the app and the subprocess as well:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web$stop()
web$get_state()
#> [1] "not running"
}\if{html}{\out{</div>}}

\code{local_app_process()} is similar to \code{new_app_process()}, but it stops the server process at the end of the calling block.
This means that the process is automatically cleaned up at the end of a \code{test_that()} block or at the end of the test file.

You can create your app at the beginning of your test file.
Or, if you want to use the same app in multiple test files, use a
\href{https://testthat.r-lib.org/reference/test_file.html#special-files}{testthat helper file}.
Sometimes it useful if your users can create and use your test app,
for example to create reproducible examples.
You can include a (possibly internal) function in your package, that
creates the app.

See \code{?new_app()}, \code{?new_app_process()} and \code{?local_app_process} for more details.
}

\subsection{How do I use \code{httpbin_app()} (or another app) with testthat?}{

You can use testthat's setup files. You start the app in a setup file and also register a teardown expression for it. \code{local_app_process()} can do both in one go. Your \code{tests/testthat/setup-http.R} may look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{http <- webfakes::local_app_process(
  webfakes::httpbin_app(),
  .local_envir = testthat::teardown_env()
)
}\if{html}{\out{</div>}}

(Before testthat 3.0.0, you had to write the teardown expression in a \code{tests/testthat/teardown-http.R} file. That still works, but a single setup file is considered to be better practice, see \href{https://testthat.r-lib.org/articles/test-fixtures.html}{this testthat vignette}.)

In the test cases you can query the \code{http} app process to get the URLs you need to connect to:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{test_that("fails on 404", \{
  url <- http$url("/status/404")
  response <- httr::GET(url)
  expect_error(
    httr::stop_for_status(response),
    class = "http_404"
  )
\})
#> Test passed
}\if{html}{\out{</div>}}

When writing your tests interactively, you may create a \code{http} app process in the global environment, for convenience. You can \code{source()} your \code{setup-http.R} file for this. Alternatively, you can start the app process in a helper file. See "How do I start the app when writing the tests?" just below.
}

\subsection{How do I start the app when writing the tests?}{

It is convenient to start the webfakes server process(es) when working on the tests interactively, e.g. when using \code{devtools::load_all()}.
With \code{local_app_process()} in the testthat \code{setup*.R} file this is not automatic, because \code{devtools::load_all()} does not run these files.
So you would need to source the \code{setup*.R} files manually, which is error prone.

One solution is to create server processes in the testthat \code{helper*.R} files.
\code{load_all()} executes the helper files by default.
So instead of using a setup file, you can simply do this in the \code{helper-http.R} file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{httpbin <- local_app_process(httpbin_app())
}\if{html}{\out{</div>}}

If the app process is created in the helper file, then it is ready use after \code{load_all()}, and (by default) the actual process will be started at the first \verb{$url()} or \verb{$get_port()} call.
You can also start it manually with \verb{$start()}.

Processes created in helper files are not cleaned up automatically at the end of the test suite, unless you clean them up by registering a \verb{$stop()} call in a setup file, like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{withr::defer(httpbin$stop(), testthat::teardown_env())
}\if{html}{\out{</div>}}

In practice this is not necessary, because \verb{R CMD check} runs the tests in a separate process, and when that finishes, the webfakes processes are cleaned up as well.

When running \code{devtools::test()}, \code{testthat::test_local()} or another testthat function to run (part of) the test suite in the current session, the \code{helper*.R} files are (re)loaded first.
This will terminate the currently running app processes, if any, and create new app process objects.
Should the test suite auto-start some of the test processes from \code{helper*.R}, these will not be cleaned up at the end of the test suite, but only at the next \code{load_all()} or \code{test()} call, or at the end of the R session.
This lets you run your test code interactively, either via \code{test()} or manually, without thinking too much about the webfakes processes.
}

\subsection{Can I have an app for a single testthat test file?}{

To run a web app for a single test file, start it with \code{new_app_process()} at the beginning of the file, and register its cleanup using \code{withr::defer()}.
Even simpler, use \code{local_app_process()} which is the same as \code{new_app_process()} but it automatically stops the web server process, at the end of the test file:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app <- webfakes::new_app()
app$get("/hello/:user", function(req, res) \{
  res$send(paste0("Hello ", req$params$user, "!"))
\})
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web <- webfakes::local_app_process(app)
}\if{html}{\out{</div>}}

Then in the test cases, use \code{web$url()} to get the URL to connect to.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{test_that("can use hello API", \{
  url <- web$url("/hello/Gabor")
  expect_equal(httr::content(httr::GET(url)), "Hello Gabor!")
\})
#> No encoding supplied: defaulting to UTF-8.
#> Test passed
}\if{html}{\out{</div>}}
}

\subsection{Can I use an app for a single testthat test?}{

Sure.
For this you need to create the app process within the \code{testthat::test_that()} test case.
\code{local_app_process()} automatically cleans it up at the end of the block.
It goes like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{test_that("query works", \{
  app <- webfakes::new_app()
  app$get("/hello", function(req, res) res$send("hello there"))
  web <- webfakes::local_app_process(app)

  echo <- httr::content(httr::GET(web$url("/hello")))
  expect_equal(echo, "hello there")
\})
#> No encoding supplied: defaulting to UTF-8.
#> Test passed
}\if{html}{\out{</div>}}
}

\subsection{How do I test a sequence of requests?}{

To test a sequence of requests, the app needs state information that is kept between requests.
\code{app$locals} is an environment that belongs to the app, and it can be used to record information and then retrieve it in future requests.
You could store anything in \code{app$locals}, something simple like a counter variable, something fancier like a sqlite database.
You can add something to \code{app$locals} via methods or directly after creating the app.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{store <- webfakes::new_app()
store$locals$packages <- list("webfakes")
ls(store$locals)
#> [1] "packages"
store$locals$packages
#> [[1]]
#> [1] "webfakes"
}\if{html}{\out{</div>}}

E.g. here is an end point that fails three times, then succeeds once, fails again three times, etc.

Note that the \code{counter} created by the code below starts at 0, not 1.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{flaky <- webfakes::new_app()
flaky$get("/unstable", function(req, res) \{
  if (identical(res$app$locals$counter, 3L)) \{
    res$app$locals$counter <- NULL
    res$send_json(object = list(result = "ok"))
  \} else \{
    res$app$locals$counter <- c(res$app$locals$counter, 0L)[[1]] + 1L
    res$send_status(401)
  \}
\})
}\if{html}{\out{</div>}}

Let's run this app in another process and connect to it:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{pr <- webfakes::new_app_process(flaky)
url <- pr$url("/unstable")
httr::RETRY("GET", url, times = 4)
#> Request failed [401]. Retrying in 1 seconds...
#> Request failed [401]. Retrying in 1 seconds...
#> Request failed [401]. Retrying in 2.1 seconds...
#> Response [http://127.0.0.1:64374/unstable]
#>   Date: 2025-01-14 10:07
#>   Status: 200
#>   Content-Type: application/json
#>   Size: 17 B
}\if{html}{\out{</div>}}

Another example where we send information to an app and then retrieve it.
On a POST request we store the \code{name} query parameter in \code{app$locals$packages}, which can be queried with a GET request.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{store <- webfakes::new_app()
# Initial "data" for the app
store$locals$packages <- list("webfakes")
# Get method
store$get("/packages", function(req, res) \{
  res$send_json(res$app$locals$packages, auto_unbox = TRUE)
\})
# Post method, store information from the query
store$post("/packages", function(req, res) \{
  res$app$locals$packages <- c(res$app$locals$packages, req$query$name)
  res$send_json(res$app$locals$packages, auto_unbox = TRUE)
\})
}\if{html}{\out{</div>}}

Now we start the app in a subprocess, and run a GET query against it.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web <- webfakes::local_app_process(store, start = TRUE)
# Get current information
get_packages <- function() \{
  httr::content(
    httr::GET(
      httr::modify_url(
        web$url(), 
        path = "packages"
        )
      )
    )
\}
get_packages()
#> [[1]]
#> [1] "webfakes"
}\if{html}{\out{</div>}}

Let's POST some new information.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{post_package <- function(name) \{
  httr::POST(
  httr::modify_url(
    web$url(),  
    path = "packages",
    query = list(name = name)
  )
)
\}
post_package("vcr")
#> Response [http://127.0.0.1:64380/packages?name=vcr]
#>   Date: 2025-01-14 10:07
#>   Status: 200
#>   Content-Type: application/json
#>   Size: 18 B
# Get current information
get_packages()
#> [[1]]
#> [1] "webfakes"
#> 
#> [[2]]
#> [1] "vcr"

post_package("httptest")
#> Response [http://127.0.0.1:64380/packages?name=httptest]
#>   Date: 2025-01-14 10:07
#>   Status: 200
#>   Content-Type: application/json
#>   Size: 29 B
# Get current information
get_packages()
#> [[1]]
#> [1] "webfakes"
#> 
#> [[2]]
#> [1] "vcr"
#> 
#> [[3]]
#> [1] "httptest"
}\if{html}{\out{</div>}}

Stop the app process:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web$stop()

}\if{html}{\out{</div>}}
}

\subsection{How can I debug an app?}{

To debug an app, it is best to run it in the main R process, i.e. \emph{not} via \code{new_app_process()}.
You can add breakpoints, or \code{browser()} calls to your handler functions, and then invoke your app from another process.
You might find the \code{curl} command line tool to send HTTP requests to the app, or you can just use another R process.
Here is an example.
We will simply print the incoming request object to the screen now.
For a real debugging session you probably want to place a \code{browser()} command there.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{app <- webfakes::new_app()
app$get("/debug", function(req, res) \{
  print(req)
  res$send("Got your back")
\})
}\if{html}{\out{</div>}}

Now start the app on port 3000:

\if{html}{\out{<div class="sourceCode {.r}">}}\preformatted{app$listen(port = 3000)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> Running webfakes web app on port 3000
}\if{html}{\out{</div>}}

Connect to the app from another R or \code{curl} process:

\if{html}{\out{<div class="sourceCode {.sh}">}}\preformatted{curl -v http://127.0.0.1:3000/debug
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> *   Trying 127.0.0.1...
#> * TCP_NODELAY set
#> * Connected to 127.0.0.1 (127.0.0.1) port 3000 (#0)
#> > GET /debug HTTP/1.1
#> > Host: 127.0.0.1:3000
#> > User-Agent: curl/7.54.0
#> > Accept: */*
#> >
#> < HTTP/1.1 200 OK
#> < Content-Type: text/plain
#> < Content-Length: 13
#> <
#> * Connection #0 to host 127.0.0.1 left intact
#> Got your back
}\if{html}{\out{</div>}}

Your main R session will print the incoming request:

\if{html}{\out{<div class="sourceCode">}}\preformatted{#> <webfakes_request>
#> method:
#>   get
#> url:
#>   http://127.0.0.1:3000/debug
#> client:
#>   127.0.0.1
#> query:
#> headers:
#>   Host: 127.0.0.1:3000
#>   User-Agent: curl/7.54.0
#>   Accept: */*
#> fields and methods:
#>   app                    # the webfakes_app the request belongs to
#>   headers                # HTTP request headers
#>   hostname               # server hostname, the Host header
#>   method                 # HTTP method of request (lowercase)
#>   path                   # server path
#>   protocol               # http or https
#>   query_string           # raw query string without '?'
#>   query                  # named list of query parameters
#>   remote_addr            # IP address of the client
#>   url                    # full URL of the request
#>   get_header(field)      # get a request header
#>  # see ?webfakes_request for details
}\if{html}{\out{</div>}}

Press \code{CTRL+C} or \code{ESC} to interrupt the app in the main session.
}

\subsection{How can I test HTTPS requests?}{

Serving HTTPS from \code{localhost} or \verb{127.0.0.1} instead of HTTP is easy,
all you need to do is
\enumerate{
\item Set the port to an HTTPS port by adding an \code{"s"} suffix
to the port number. Use \code{"0s"} for an OS assigned free port:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{new_app_process(app, port = "0s")
}\if{html}{\out{</div>}}

By default webfakes uses the server key + certificate in the file at

\if{html}{\out{<div class="sourceCode r">}}\preformatted{system.file("cert/localhost/server.pem", package = "webfakes")
}\if{html}{\out{</div>}}

This certificate includes \code{localhost}, \verb{127.0.0.1} and
\code{localhost.localdomain}. If you need another domain or IP address,
you'll need to create your own certificate. The \code{generate.sh}
file in the same directory helps with that.
\item Specify the certificate bundle to the HTTP client you are using.
For the default server key use the \code{ca.crt} file from
the webfakes package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{system.file("cert/localhost/ca.crt", package = "webfakes")
}\if{html}{\out{</div>}}

See examples for HTTP clients below.
}

If you are using the curl package, use the \code{ca_info} option in
\code{curl::new_handle()} or \code{curl::handle_setopt()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{cainfo <- system.file("cert/localhost/ca.crt", package = "webfakes")
curl::curl_fetch_memory(
  http$url("/path/to/endpoint"), 
  handle = curl::new_handle(cainfo = cainfo)
)
}\if{html}{\out{</div>}}

For the httr package, use \code{httr::config(cainfo = ...)}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{httr::GET(
  http$url("/headers", https = TRUE), 
  httr::config(cainfo = cainfo)
)
}\if{html}{\out{</div>}}

For the httr2 package:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{httr2::request("https://example.com") |>
  httr2::req_options(cainfo = cainfo) |>
  httr2::req_perform()
}\if{html}{\out{</div>}}

For \code{utils::download.file} point the \code{CURL_CA_BUNDLE} environment variable
to the \code{ca.crt} file. Don't forget to undo this, once the HTTP request
is done.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{Sys.setenv(
  CURL_CA_BUNDLE = system.file("cert/localhost/ca.crt", package = "webfakes")
)
download.file(http$url("/path/to/endpoint"), res <- tempfile())
}\if{html}{\out{</div>}}
\subsection{Special considerations for tests on Windows}{

Unfortunately things are not that simple on Windows, for the HTTP clients.
As far as I can tell, it is not easily possible to make the HTTP clients
accept a new self-signed certificate. It is possible with libcurl, though,
but you need to set the \code{CURL_SSL_BACKEND=openssl} environment variable.
(And libcurl must be built with openssl support of course.) You need to
set this env var \emph{before} loading libcurl, so it is best to set it before
starting R. One way to do this in the tests is to run the tests in a
subprocess, with the callr package. Look at the \code{test-https.R} file in
webfakes for a complete, current example. The tests use this helper
function, defined in \code{helper.R}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{callr_curl <- function(url, options = list()) \{
  callr::r(
    function(url, options) \{
      h <- curl::new_handle()
      curl::handle_setopt(h, .list = options)
      curl::curl_fetch_memory(url, handle = h)
    \},
    list(url = url, options = options),
    env = c(
      callr::rcmd_safe_env(),
      CURL_SSL_BACKEND = "openssl",
      CURL_CA_BUNDLE = if ("cainfo" \%in\% names(options)) options$cainfo
    )
  )
\}
}\if{html}{\out{</div>}}

Example test case:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# ...
  cainfo <- system.file("cert/localhost/ca.crt", package = "webfakes")
  resp <- if (.Platform$OS.type == "windows") \{
    callr_curl(http$url("/hello"), list(cainfo = cainfo))
  \} else \{
    curl::curl_fetch_memory(
      http$url("/hello"),
      handle = curl::new_handle(cainfo = cainfo)
    )
  \}
# ...
}\if{html}{\out{</div>}}

It seems like a good idea to \code{skip_on_cran()} HTTPS tests, at least on
Windows, because this setup is not yet tested enough to consider it
robust.

webfakes uses \href{https://www.trustedfirmware.org/projects/mbed-tls/}{Mbed TLS}
for serving HTTPS.
}

}

\subsection{How can I run a server on multiple ports?}{

You can specify multiple port numbers, in a vector. webfakes will
then listen on all those ports. You can also mix HTTP and HTTP ports.

To redirect from an HTTP port to an HTTPS port, append an \code{"r"} suffix
to the HTTP port number. THis port will be redirected to the next HTTPS
port. E.g.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{new_app_process(app, port = c("3000r", "3001s"))
}\if{html}{\out{</div>}}

will redirect HTTP from port 3000 to HTTPS on port 3001.

To redirect from an OS assigned HTTP port to an OS assigned HTTPS port,
use zeros for the port numbers:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{http <- new_app_process(app, port = c("0r", "0s"))
}\if{html}{\out{</div>}}

Then you can use \code{http$get_ports()} to query all port numbers.

You can also use

\if{html}{\out{<div class="sourceCode r">}}\preformatted{http$url(..., https = TRUE)
}\if{html}{\out{</div>}}

to get an HTTPS URL instead of the default one (the one with the
first port).
}

\subsection{Can I test asynchronous or parallel HTTP requests?}{

R is single threaded and a webfakes app runs an R interpreter, so it cannot process multiple requests at the same time.
The web server itself runs in a separate thread, and it can also process each request in a separate thread, but at any time only one request can use the R interpreter.

This is important, because sometimes test requests may take longer to process.
For example the \verb{/delay/:secs} end point of \code{httpbin_app()} wait for the specified number of seconds before responding, to simulate a slow web server.
If this wait is implemented via the standard \code{Sys.sleep()} R function, then no other requests can be processed until the sleep is over.
To avoid this, webfakes can put the waiting request on hold, return from the R interpreter, and respond to other incoming requests.
Indeed, the \verb{/delay/} end point is implemented using this feature.

However, the request thread of the web server is still busy while on hold, so to take advantage of this, you need to allow multiple threads.
The \code{num_threads} argument of the \verb{$listen()} method of \code{webfakes_app} lets you specify the number of request threads the web server will use.
Similarly, the \code{num_threads} argument of \code{local_app_process()} lets you modify the number of threads.

When testing asynchronous or parallel code, that might invoke multiple, possibly delayed requests, it is best to increase the number of threads.
The code below calls the same API request concurrently, three times.
Each request takes 1 second to answer, but if the web server has more than three threads, together they'll still take about 1 second.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web <- webfakes::local_app_process(
  webfakes::httpbin_app(),
  opts = webfakes::server_opts(num_threads = 6, enable_keep_alive = TRUE)
)
}\if{html}{\out{</div>}}

\if{html}{\out{<div class="sourceCode r">}}\preformatted{testthat::test_that("parallel requests", \{
  url <- web$url("/delay/0.5")
  p <- curl::new_pool()
  handles <- replicate(3, curl::new_handle(url = url))
  resps <- list()
  for (handle in handles) \{
    curl::multi_add(
      handle,
      done = function(x) resps <<- c(resps, list(x)),
      fail = stop,
      pool = p
    )
  \}
  st <- system.time(curl::multi_run(timeout = 3, pool = p))
  testthat::expect_true(st[["elapsed"]] < 1.5)
\})
#> Test passed
}\if{html}{\out{</div>}}

(If this should fail for you and webfakes appears to process the
requests sequentially, see \href{https://github.com/r-lib/webfakes/issues/108}{issue #108} for possible workarounds.)
}

\subsection{How to make sure that my code works with the \emph{real} API?}{

Indeed, if you use webfakes for your test cases, then they never touch the real web server.
As you might suspect, this is not ideal, especially when you do not control the server.
The web service might change their API, and your test cases will fail to warn you.

One practical solution is to write (at least some) flexible tests, that can run against a local fake webserver, or a real one, and you have a quick switch to change their behavior.
I have found that environment variables work great for this.

E.g. if the \code{FAKE_HTTP_TESTS} environment variable is not set, the tests run with the real web server, otherwise they use a fake one.
Another solution, that works best is the HTTP requests are in the downstream package code, is to introduce one environment variable for each API you need to connect to.
These might be set to the real API servers, or to the fake ones.

Once you have some tests that can use both kinds or servers, you can set up your continuous integration (CI) framework, to run the tests agains the real server (say) once a day.
This special CI run makes sure that your code works well with the real API.
You can run all the other tests, locally and in the CI, against the fake local web server.

See the question on \href{#how-do-i-make-my-app-connect-to-webfakes-when-the-tests-are-running-}{how webfakes helps you setting environment variables that point to your local server}.
}

\subsection{How do I simulate a slow internet connection?}{

You need to use the \code{throttle} server option when you start your web app.
This means that you can run the very same app with different connection speed.
This is how it goes:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(webfakes)
slow <- new_app_process(
  httpbin_app(),
  opts = server_opts(throttle = 100)
)
resp <- curl::curl_fetch_memory(slow$url("/bytes/200"))
resp$times
#>      redirect    namelookup       connect   pretransfer starttransfer 
#>      0.000000      0.000087      0.000266      0.000284      0.004878 
#>         total 
#>      2.013756
}\if{html}{\out{</div>}}

\code{throttle} gives the number of bytes per second, so downloading 200 random bytes from the fake app will take about 2 seconds.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-raw.R
\name{mw_raw}
\alias{mw_raw}
\title{Middleware to read the raw body of a request}
\usage{
mw_raw(type = "application/octet-stream")
}
\arguments{
\item{type}{Content type to match. If it does not match, then the
request object is not modified.}
}
\value{
Handler function.
}
\description{
Adds the raw body, as a raw object to the \code{raw} field of the request.
}
\examples{
app <- new_app()
app$use(mw_raw())
app
}
\seealso{
Other middleware:
\code{\link[=mw_cgi]{mw_cgi()}},
\code{\link[=mw_cookie_parser]{mw_cookie_parser()}},
\code{\link[=mw_etag]{mw_etag()}},
\code{\link[=mw_json]{mw_json()}},
\code{\link[=mw_log]{mw_log()}},
\code{\link[=mw_multipart]{mw_multipart()}},
\code{\link[=mw_range_parser]{mw_range_parser()}},
\code{\link[=mw_static]{mw_static()}},
\code{\link[=mw_text]{mw_text()}},
\code{\link[=mw_urlencoded]{mw_urlencoded()}}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/local-app-process.R
\name{local_app_process}
\alias{local_app_process}
\title{App process that is cleaned up automatically}
\usage{
local_app_process(app, ..., .local_envir = parent.frame())
}
\arguments{
\item{app}{\code{webfakes_app} object, the web app to run.}

\item{...}{Passed to \code{\link[=new_app_process]{new_app_process()}}.}

\item{.local_envir}{The environment to attach the process cleanup to.
Typically a frame. When this frame finishes, the process is stopped.}
}
\description{
You can start the process with an explicit \verb{$start()} call.
Alternatively it starts up at the first \verb{$url()} or \verb{$get_port()}
call.
}
\seealso{
\code{\link[=new_app_process]{new_app_process()}} for more details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-urlencoded.R
\name{mw_urlencoded}
\alias{mw_urlencoded}
\title{Middleware to parse an url-encoded request body}
\usage{
mw_urlencoded(type = "application/x-www-form-urlencoded")
}
\arguments{
\item{type}{Content type to match before parsing. If it does not
match, then the request object is not modified.}
}
\value{
Handler function.
}
\description{
This is typically data from a form. The parsed data is added
as the \code{form} element of the request object.
}
\examples{
app <- new_app()
app$use(mw_urlencoded())
app
}
\seealso{
Other middleware:
\code{\link[=mw_cgi]{mw_cgi()}},
\code{\link[=mw_cookie_parser]{mw_cookie_parser()}},
\code{\link[=mw_etag]{mw_etag()}},
\code{\link[=mw_json]{mw_json()}},
\code{\link[=mw_log]{mw_log()}},
\code{\link[=mw_multipart]{mw_multipart()}},
\code{\link[=mw_range_parser]{mw_range_parser()}},
\code{\link[=mw_raw]{mw_raw()}},
\code{\link[=mw_static]{mw_static()}},
\code{\link[=mw_text]{mw_text()}}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{webfakes_response}
\alias{webfakes_response}
\title{A webfakes response object}
\description{
webfakes creates a \code{webfakes_response} object for every incoming HTTP
request. This object is passed to every matched route and middleware,
until the HTTP response is sent. It has reference semantics, so handlers
can modify it.
}
\details{
Fields and methods:
\itemize{
\item \code{app}: The \code{webfakes_app} object itself.
\item \code{req}: The request object.
\item \code{headers_sent}: Whether the response headers were already sent out.
\item \code{locals}: Local variables, the are shared between the handler
functions. This is for the end user, and not for the middlewares.
\item \code{delay(secs)}: delay the response for a number of seconds. If a
handler calls \code{delay()}, the same handler will be called again,
after the specified number of seconds have passed. Use the \code{locals}
environment to distinguish between the calls. If you are using
\code{delay()}, and want to serve requests in parallel, then you probably
need a multi-threaded server, see \code{\link[=server_opts]{server_opts()}}.
\item \code{add_header(field, value)}: Add a response header. Note that
\code{add_header()} may create duplicate headers. You usually want
\code{set_header()}.
\item \code{get_header(field)}: Query the currently set response headers. If
\code{field} is not present it return \code{NULL}.
\item \code{on_response(fun)}: Run the \code{fun} handler function just before the
response is sent out. At this point the headers and the body are
already properly set.
\item \code{redirect(path, status = 302)}: Send a redirect response. It sets
the \code{Location} header, and also sends a \code{text/plain} body.
\item \code{render(view, locals = list())}: Render a template page. Searches
for the \code{view} template page, using all registered engine extensions,
and calls the first matching template engine. Returns the filled
template.
\item \code{send(body)}. Send the specified body. \code{body} can be a raw vector,
or HTML or other text. For raw vectors it sets the content type to
\code{application/octet-stream}.
\item \code{send_json(object = NULL, text = NULL, ...)}: Send a JSON response.
Either \code{object} or \code{text} must be given. \code{object} will be converted
to JSON using \code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. \code{...} are passed to
\code{\link[jsonlite:fromJSON]{jsonlite::toJSON()}}. It sets the content type appropriately.
\item \code{send_file(path, root = ".")}: Send a file. Set \code{root = "/"} for
absolute file names. It sets the content type automatically, based
on the extension of the file, if it is not set already.
\item \code{send_status(status)}: Send the specified HTTP status code, without
a response body.
\item \code{send_chunk(data)}: Send a chunk of a response in chunked encoding.
The first chunk will automatically send the HTTP response headers.
Webfakes will automatically send a final zero-lengh chunk, unless
\verb{$delay()} is called.
\item \code{set_header(field, value)}: Set a response header. If the headers have
been sent out already, then it throws a warning, and does nothing.
\item \code{set_status(status)}: Set the response status code. If the headers
have been sent out already, then it throws a warning, and does nothing.
\item \code{set_type(type)}: Set the response content type. If it contains a \code{/}
character then it is set as is, otherwise it is assumed to be a file
extension, and the corresponding MIME type is set. If the headers have
been sent out already, then it throws a warning, and does nothing.
\item \code{add_cookie(name, value, options)}: Adds a cookie to the response.
\code{options} is a named list, and may contain:
\itemize{
\item \code{domain}: Domain name for the cookie, not set by default.
\item \code{expires}: Expiry date in GMT. It must be a POSIXct object, and
will be formatted correctly.
\item 'http_only': if TRUE, then it sets the 'HttpOnly' attribute, so
Javasctipt cannot access the cookie.
\item \code{max_age}: Maximum age, in number of seconds.
\item \code{path}: Path for the cookie, defaults to "/".
\item \code{same_site}: The 'SameSite' cookie attribute. Possible values are
"strict", "lax" and "none".
\item \code{secure}: if TRUE, then it sets the 'Secure' attribute.
}
\item \code{clear_cookie(name, options = list())}: clears a cookie. Typically,
web browsers will only clear a cookie if the options also match.
\item \code{write(data)}: writes (part of) the body of the response. It also
sends out the response headers, if they haven't been sent out before.
}

Usually you need one of the \code{send()} methods, to send out the HTTP
response in one go, first the headers, then the body.

Alternatively, you can use \verb{$write()} to send the response in parts.
}
\examples{
# This is how you can see the request and response objects:
app <- new_app()
app$get("/", function(req, res) {
  browser()
  res$send("done")
})
app

# Now start this app on a port:
# app$listen(3000)
# and connect to it from a web browser: http://127.0.0.1:3000
# You can also use another R session to connect:
# httr::GET("http://127.0.0.1:3000")
# or the command line curl tool:
# curl -v http://127.0.0.1:3000
# The app will stop while processing the request.
}
\seealso{
\link{webfakes_request} for the webfakes request object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{introduction}
\alias{introduction}
\title{Happy HTTP testing with webfakes}
\description{
Happy HTTP testing with webfakes
}
\section{Happy HTTP testing with webfakes}{

\subsection{What is webfakes?}{

Webfakes is an R package that can spin up web servers on your machine to facilitate testing R code.
R code that needs an HTTP connection is not trivial to test:
\itemize{
\item Connectivity problems might prevent the tests from accessing the web server.
\item The web server might need authentication, and it is not easy to convey login information to the test suite in a secure way.
\item The web server might have rate limits, i.e, it limits the number of queries per hour or day, causing some spurious test failures.
\item You might want to test in non-normal conditions, e.g. with low bandwidth, or when the client is rate limited. These conditions don't normally happen on the web server and they are hard to trigger.
}

With webfakes you can easily start a custom web app, that is running on the local machine.
\itemize{
\item Webfakes does not need a network connection.
\item Webfakes does not need authentication. Well, unless you want it to.
\item Webfakes does not have rate limits.
\item Webfakes can simulate low bandwidth, or a broken connection.
}
}

\subsection{Webfakes vs mocking}{

Mocking is a general technique to mimic the behavior of a function or object that is needed in test case.
In the case of HTTP requests, this typically means that both the request and its response are recorded when the tests run the first time, and saved to disk.
Subsequent test runs intercept the HTTP requests, match them against the recorded requests and then replay the corresponding recorded response.
See for example the \href{https://docs.ropensci.org/vcr/}{vcr} and \href{https://enpiar.com/r/httptest/}{httptest} R packages.

The advantages of using your own webfakes server, over mocking:
\itemize{
\item Simpler infrastructure. No separate recording and replaying phases, no recorded files. No request matching.
\item You can use any web client you want. E.g. curl and base R's HTTP functions do not explicitly support mocking currently.
\item No need to worry about sensitive information in recorded requests or responses.
\item Often easier to use when testing non-normal conditions, e.g. errors that are hard to trigger, low bandwidth, or rate limits.
\item Works if you stream data from a HTTP connection, instead of reading the whole response at once.
\item You can reuse the same app for multiple tests, in multiple packages.
\item Easier to use for tests that require multiple rounds of requests.
\item Comes with a built-in \verb{https://httpbin.org} compatible app, chances are, you don't even need to write your testing app, just start writing tests right away.
\item Better test writing experience. This is subjective, and your mileage may vary.
}
}

\subsection{Webfakes vs the real API}{
\itemize{
\item No network needed. No more \code{skip_if_offline()}.
\item Much faster.
\item No rate limits. But you can simulate one if you want to.
\item You can write your custom app.
\item Simulate low bandwidth or a broken connection.
}
}

\subsection{Webfakes vs httpbin.org}{
\itemize{
\item No network needed. No more \code{skip_if_offline()}.
\item Much faster.
\item You can use the built-in \code{webfakes::httpbin_app()} app, so it is easy to switch from httpbin.org.
\item You can write your custom app, httpbin.org might not have what you need.
}
}

\subsection{Using \code{webfakes::httpbin_app()} with testthat}{

You can use testthat's setup files. You start the app in a setup file and also register a teardown expression for it. \code{local_app_process()} can do both in one go. Your \code{tests/testthat/setup-http.R} may look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{http <- webfakes::local_app_process(
  webfakes::httpbin_app(),
  .local_envir = testthat::teardown_env()
)
}\if{html}{\out{</div>}}

(Before testthat 3.0.0, you had to write the teardown expression in a \code{tests/testthat/teardown-http.R} file. That still works, but a single setup file is considered to be better practice, see \href{https://testthat.r-lib.org/articles/test-fixtures.html}{this testthat vignette}.)

In the test cases you can query the \code{http} app process to get the URLs you need to connect to:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{test_that("fails on 404", \{
  url <- http$url("/status/404")
  response <- httr::GET(url)
  expect_error(
    httr::stop_for_status(response),
    class = "http_404"
  )
\})
#> Test passed
}\if{html}{\out{</div>}}

When writing your tests interactively, you may create a \code{http} app process in the global environment, for convenience. You can \code{source()} your \code{setup-http.R} file for this. Alternatively, you can start the app process in a helper file. See "How do I start the app when writing the tests?" just below.

You can also create a web server for a test file, or even for a single test case.
See \code{vignette("how-to")} for details how.
}

\subsection{Writing apps}{

If the builtin \code{httpbin_app()} is not appropriate for your tests, you can write your own app.
You can also extend the \code{httpbin_app()} app, if you don't want to start from scratch.

You create a new app with \code{new_app()}.
This returns an object with methods to add middleware and API
endpoints to it.
For example, a simple app that returns the current time in JSON would
look like this:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{time <- webfakes::new_app()
time$get("/time", function(req, res) \{
  res$send_json(list(time = format(Sys.time())), auto_unbox = TRUE)
\})
}\if{html}{\out{</div>}}

Now you can start this app on a random port using \code{web$listen()}.
Alternatively, you can start it in a subprocess with \code{new_app_process()}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web <- webfakes::new_app_process(time)
web$url()
#> [1] "http://127.0.0.1:64358/"
}\if{html}{\out{</div>}}

Use \code{web$url()} to query the URL of the app. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{url <- web$url("/time")
httr::content(httr::GET(url))
#> $time
#> [1] "2025-01-14 10:07:33"
}\if{html}{\out{</div>}}

\code{web$stop()} stops the app and the subprocess as well:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{web$stop()
web$get_state()
#> [1] "not running"
}\if{html}{\out{</div>}}

\code{local_app_process()} is similar to \code{new_app_process()}, but it stops the server process at the end of the calling block.
This means that the process is automatically cleaned up at the end of a \code{test_that()} block or at the end of the test file.

You can create your app at the beginning of your test file.
Or, if you want to use the same app in multiple test files, use a
\href{https://testthat.r-lib.org/reference/test_file.html#special-files}{testthat helper file}.
Sometimes it useful if your users can create and use your test app,
for example to create reproducible examples.
You can include a (possibly internal) function in your package, that
creates the app.

See \code{?new_app()}, \code{?new_app_process()} and \code{?local_app_process} for more details.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-cgi.R
\name{mw_cgi}
\alias{mw_cgi}
\title{Middleware that calls a CGI script}
\usage{
mw_cgi(command, args = character(), timeout = as.difftime(Inf, units = "secs"))
}
\arguments{
\item{command}{External command to run.}

\item{args}{Arguments to pass to the external command.}

\item{timeout}{Timeout for the external command. If the command does
not terminate in time, the web server kills it and returns an 500
response.}
}
\value{
A function with signature

\if{html}{\out{<div class="sourceCode">}}\preformatted{function(req, res, env = character())
}\if{html}{\out{</div>}}

See \href{https://datatracker.ietf.org/doc/html/rfc3875}{RFC 3875}
for details on the CGI protocol.

The request body (if any) is passed to the external command as standard
intput. \code{mw_cgi()} sets \code{CONTENT_LENGTH}, \code{CONTENT_TYPE},
\code{GATEWAY_INTERFACE}, \code{PATH_INFO}, \code{QUERY_STRING}, \code{REMOTE_ADDR},
\code{REMOTE_HOST}, \code{REMOTE_USER}, \code{REQUEST_METHOD}, \code{SERVER_NAME},
\code{SERVER_PORT}, \code{SERVER_PROTOCOL}, \code{SERVER_SOFTEWARE}.

It does not currently set the \code{AUTH_TYPE}, \code{PATH_TRANSLATED},
\code{REMOTE_IDENT}, \code{SCRIPT_NAME} environment variables.

The standard output of the external command is used to set the
response status code, the response headers and the response body.
Example output from git's CGI:

\if{html}{\out{<div class="sourceCode">}}\preformatted{Status: 200 OK
Expires: Fri, 01 Jan 1980 00:00:00 GMT
Pragma: no-cache
Cache-Control: no-cache, max-age=0, must-revalidate
Content-Type: application/x-git-upload-pack-advertisement

000eversion 2
0015agent=git/2.42.0
0013ls-refs=unborn
0020fetch=shallow wait-for-done
0012server-option
0017object-format=sha1
0010object-info
0000
}\if{html}{\out{</div>}}
}
\description{
You can use it as an unconditional middleware in \code{app$use()},
as a handler on \code{app$get()}, \code{app$post()}, etc., or you can call it
from a handler. See examples below.
}
\examples{
app <- new_app()
app$use(mw_cgi("echo", "Status: 200\n\nHello"))
app

app2 <- new_app()
app2$get("/greet", mw_cgi("echo", "Status: 200\n\nHello"))
app2

# Using `mw_cgi()` in a handler, you can pass extra environment variables
app3 <- new_app()
cgi <- mw_cgi("echo", "Status: 200\n\nHello")
app2$get("/greet", function(req, res) {
  cgi(req, res, env = c("EXTRA_VAR" = "EXTRA_VALUE"))
})
app3
}
\seealso{
Other middleware: 
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_json}()},
\code{\link{mw_log}()},
\code{\link{mw_multipart}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-cookie-parser.R
\name{mw_cookie_parser}
\alias{mw_cookie_parser}
\title{Middleware to parse Cookies}
\usage{
mw_cookie_parser()
}
\value{
Handler function.
}
\description{
Adds the cookies as the \code{cookies} element of the request object.
}
\details{
It ignores cookies in an invalid format. It ignores duplicate cookies:
if two cookies have the same name, only the first one is included.
}
\seealso{
Other middleware:
\code{\link[=mw_cgi]{mw_cgi()}},
\code{\link[=mw_etag]{mw_etag()}},
\code{\link[=mw_json]{mw_json()}},
\code{\link[=mw_log]{mw_log()}},
\code{\link[=mw_multipart]{mw_multipart()}},
\code{\link[=mw_range_parser]{mw_range_parser()}},
\code{\link[=mw_raw]{mw_raw()}},
\code{\link[=mw_static]{mw_static()}},
\code{\link[=mw_text]{mw_text()}},
\code{\link[=mw_urlencoded]{mw_urlencoded()}}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git-app.R
\name{git_app}
\alias{git_app}
\title{Web app that acts as a git http server}
\usage{
git_app(
  git_root,
  git_cmd = "git",
  git_timeout = as.difftime(1, units = "mins"),
  filter = TRUE,
  cleanup = TRUE
)
}
\arguments{
\item{git_root}{Path to the root of the directory tree to be served.}

\item{git_cmd}{Command to call, by default it is \code{"git"}. It may also
be a full path to git.}

\item{git_timeout}{A \code{difftime} object, time limit for the git
command.}

\item{filter}{Whether to support the \code{filter} capability in the server.}

\item{cleanup}{Whether to clean up \code{git_root} when the app is
garbage collected.}
}
\description{
It is useful for tests that need an HTTP git server.
}
\examples{
\dontshow{if (FALSE) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
dir.create(tmp <- tempfile())
setwd(tmp)
system("git clone --bare https://github.com/cran/crayon")
system("git clone --bare https://github.com/cran/glue")
app <- git_app(tmp)
git <- new_app_process(app)
system(paste("git ls-remote", git$url("/crayon")))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth2_login}
\alias{oauth2_login}
\title{Helper function to log in to a third party OAuth2.0 app without a
browser}
\usage{
oauth2_login(login_url)
}
\arguments{
\item{login_url}{The login URL of the third party app.}
}
\value{
A named list with
\itemize{
\item \code{login_response} The curl HTTP response object for the login
page.
\item \code{token_response} The curl HTTP response object for submitting
the login page.
}
}
\description{
It works with \code{\link[=oauth2_resource_app]{oauth2_resource_app()}}, and any third party app,
including the fake \code{\link[=oauth2_third_party_app]{oauth2_third_party_app()}}.
}
\details{
See \code{test-oauth.R} in webfakes for an example.
}
\seealso{
Other OAuth2.0 functions: 
\code{\link{oauth2_httr_login}()},
\code{\link{oauth2_resource_app}()},
\code{\link{oauth2_third_party_app}()}
}
\concept{OAuth2.0 functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-json.R
\name{mw_json}
\alias{mw_json}
\title{Middleware to parse a JSON body}
\usage{
mw_json(type = "application/json", simplifyVector = FALSE, ...)
}
\arguments{
\item{type}{Content type to match before parsing. If it does not
match, then the request object is not modified.}

\item{simplifyVector}{Whether to simplify lists to vectors, passed to
\code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}}.}

\item{...}{Arguments to pass to \code{\link[jsonlite:fromJSON]{jsonlite::fromJSON()}}, that performs
the JSON parsing.}
}
\value{
Handler function.
}
\description{
Adds the parsed object as the \code{json} element of the request object.
}
\examples{
app <- new_app()
app$use(mw_json())
app
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_log}()},
\code{\link{mw_multipart}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/app-process.R
\name{new_app_process}
\alias{new_app_process}
\alias{webfakes_app_process}
\title{Run a webfakes app in another process}
\usage{
new_app_process(
  app,
  port = NULL,
  opts = server_opts(remote = TRUE),
  start = FALSE,
  auto_start = TRUE,
  process_timeout = NULL,
  callr_opts = NULL
)
}
\arguments{
\item{app}{\code{webfakes_app} object, the web app to run.}

\item{port}{Port(s) to use. By default the OS assigns a port.
Add an \code{"s"} suffix to the port to use HTTPS. Use \code{"0s"} to
use an OS assigned port with HTTPS. See the \link{how-to} on how
to run the web server on multiple ports.}

\item{opts}{Server options. See \code{\link[=server_opts]{server_opts()}} for the defaults.}

\item{start}{Whether to start the web server immediately. If this is
\code{FALSE}, and \code{auto_start} is \code{TRUE}, then it is started as neeed.}

\item{auto_start}{Whether to start the web server process automatically.
If \code{TRUE} and the process is not running, then \verb{$start()},
\verb{$get_port()}, \verb{$get_ports()} and \verb{$url()} start the process.}

\item{process_timeout}{How long to wait for the subprocess to start, in
milliseconds.}

\item{callr_opts}{Options to pass to \code{\link[callr:r_session_options]{callr::r_session_options()}}
when setting up the subprocess.}
}
\value{
A \code{webfakes_app_process} object.
\subsection{Methods}{

The \code{webfakes_app_process} class has the following methods:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{get_app()
get_port()
get_ports()
stop()
get_state()
local_env(envvars)
url(path = "/", query = NULL)
}\if{html}{\out{</div>}}
\itemize{
\item \code{envvars}: Named list of environment variables. The \code{{url}} substring
is replaced by the URL of the app.
\item \code{path}: Path to return the URL for.
\item \code{query}: Additional query parameters, a named list, to add to the URL.
}

\code{get_app()} returns the app object.

\code{get_port()} returns the (first) port the web server is running on.

\code{get_ports()} returns all ports the web server is running on, and
whether it uses SSL on those ports, in a data frame with columns
\code{ipv4}, \code{ipv6}, \code{port} and \code{ssl}.

\code{stop()} stops the web server, and also the subprocess. If the error
log file is not empty, then it dumps its contents to the screen.

\code{get_state()} returns a string, the state of the web server:
\itemize{
\item \code{"not running"} the server is not running (because it was stopped
already).
\item \code{"live"} means that the server is running.
\item \code{"dead"} means that the subprocess has quit or crashed.
}

\code{local_env()} sets the given environment variables for the duration of
the app process. It resets them in \verb{$stop()}. Webfakes replaces \code{{url}}
in the value of the environment variables with the app URL, so you can
set environment variables that point to the app.

\code{url()} returns the URL of the web app. You can use the \code{path}
parameter to return a specific path.
}
}
\description{
Runs an app in a subprocess, using \link[callr:r_session]{callr::r_session}.
}
\examples{
app <- new_app()
app$get("/foo", function(req, res) {
  res$send("Hello world!")
})

proc <- new_app_process(app)
url <- proc$url("/foo")
resp <- curl::curl_fetch_memory(url)
cat(rawToChar(resp$content))

proc$stop()
}
\seealso{
\code{\link[=local_app_process]{local_app_process()}} for automatically cleaning up the
subprocess.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth2_httr_login}
\alias{oauth2_httr_login}
\title{Helper function to use httr's OAuth2.0 functions
non-interactively, e.g. in test cases}
\usage{
oauth2_httr_login(expr)
}
\arguments{
\item{expr}{Expression that calls \code{\link[httr:oauth2.0_token]{httr::oauth2.0_token()}},
either directly, or indirectly.}
}
\value{
The return value of \code{expr}.
}
\description{
To perform an automatic acknowledgement and log in for a
local OAuth2.0 app, run by httr, wrap the expression that
obtains the OAuth2.0 token in \code{oauth2_httr_login()}.
}
\details{
In interactive sessions, \code{oauth2_httr_login()} overrides the
\code{browser} option, and when httr opens a browser page, it
calls \code{\link[=oauth2_login]{oauth2_login()}} in a subprocess.

In non-interactive sessions, httr does not open a browser page,
only messages the user to do it manually. \code{oauth2_httr_login()}
listens for these messages, and calls \code{\link[=oauth2_login]{oauth2_login()}} in a
subprocess.
}
\seealso{
See \code{?vignette("oauth", package = "webfakes")} for a case
study that uses this function.

Other OAuth2.0 functions: 
\code{\link{oauth2_login}()},
\code{\link{oauth2_resource_app}()},
\code{\link{oauth2_third_party_app}()}
}
\concept{OAuth2.0 functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/path.R
\name{new_regexp}
\alias{new_regexp}
\alias{webfakes_regexp}
\title{Create a new regular expression to use in webfakes routes}
\usage{
new_regexp(x)
}
\arguments{
\item{x}{String scalar containing a regular expression.}
}
\value{
String with class \code{webfakes_regexp}.
}
\description{
Note that webfakes uses PERL regular expressions.
}
\details{
As R does not have data type or class for regular expressions,
you can use \code{new_regexp()} to mark a string as a regular expression,
when adding routes.
}
\examples{
new_regexp("^/api/match/(?<pattern>.*)$")
}
\seealso{
The 'Path specification' and 'Path parameters' chapters
of the manual of \code{\link[=new_app]{new_app()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/server.R
\name{server_opts}
\alias{server_opts}
\title{Webfakes web server options}
\usage{
server_opts(
  remote = FALSE,
  port = NULL,
  num_threads = 1,
  interfaces = "127.0.0.1",
  enable_keep_alive = FALSE,
  access_log_file = remote,
  error_log_file = TRUE,
  tcp_nodelay = FALSE,
  throttle = Inf,
  decode_url = TRUE,
  ssl_certificate = NULL
)
}
\arguments{
\item{remote}{Meta-option. If set to \code{TRUE}, webfakes uses slightly
different defaults, that are more appropriate for a background
server process.}

\item{port}{Port to start the web server on. Defaults to a randomly
chosen port.}

\item{num_threads}{Number of request handler threads to use. Typically
you don't need more than one thread, unless you run test cases in
parallel or you make concurrent HTTP requests.}

\item{interfaces}{The network interfaces to listen on. Being a test
web server, it defaults to the localhost. Only bind to a public
interface if you know what you are doing. webfakes was not designed
to serve public web pages.}

\item{enable_keep_alive}{Whether the server keeps connections alive.}

\item{access_log_file}{\code{TRUE}, \code{FALSE}, or a path. See 'Logging'
below.}

\item{error_log_file}{\code{TRUE}, \code{FALSE}, or a path. See 'Logging'
below.}

\item{tcp_nodelay}{if \code{TRUE} then packages will be sent as soon as
possible, instead of waiting for a full buffer or timeout to occur.}

\item{throttle}{Limit download speed for clients. If not \code{Inf},
then it is the maximum number of bytes per second, that is sent to
as connection.}

\item{decode_url}{Whether the server should automatically decode
URL-encodded URLs. If \code{TRUE} (the default), \verb{/foo\%2fbar} will be
converted to \verb{/foo/bar} automatically. If \code{FALSE}, URLs as not
URL-decoded.}

\item{ssl_certificate}{Path to the SSL certificate of the server,
needed if you want to server HTTPS requests.}
}
\value{
List of options that can be passed to \code{webfakes_app$listen()}
(see \code{\link[=new_app]{new_app()}}), and \code{\link[=new_app_process]{new_app_process()}}.
}
\description{
Webfakes web server options
}
\section{Logging}{

\itemize{
\item For \code{access_log_file}, \code{TRUE} means \verb{<log-dir>/access.log}.
\item For \code{error_log_file}, \code{TRUE} means \verb{<log-dir>/error.log}.
}

\verb{<log-dir>} is set to the contents of the \code{WEBFAKES_LOG_DIR}
environment variable, if it is set. Otherwise it is set to
\verb{<tmpdir>/webfakes} for local apps and \verb{<tmpdir>/<pid>/webfakes} for
remote apps (started with \code{new_app_procss()}).

\verb{<tmpdir>} is the session temporary directory of the \emph{main process}.

\verb{<pid>} is the process id of the subprocess.
}

\examples{
# See the defaults
server_opts()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs.R
\name{glossary}
\alias{glossary}
\title{webfakes glossary}
\description{
webfakes glossary
}
\section{Webfakes glossary}{


The webfakes package uses vocabulary that is standard for web apps, especially those developed with Express.js, but not necessarily well known to all R package developers.
\subsection{app}{

(Also: fake web app, webfakes app.) A web application that can be served by webfakes's web server, typically in another process, an \emph{app process}.
Sometimes we call it a \emph{fake} web app, to emphasize that we use it for testing real web apps and APIs.

You can create a webfakes app with the \code{new_app()} function.
A webfakes app is an R object that you can save to disk with \code{saveRDS()} , and you can also include it in your package.

You can start an with its \verb{$listen()} method.
Since the main R process runs that test suite code, you usually run them in a subprocess, see \code{new_app_process()} or \code{local_app_process()}.
}

\subsection{app process}{

(Also: web server process, webfakes subprocess.) An app process is an R subprocess, started from the main R process, to serve a webfakes \emph{app}.

You can create an app process object with \code{new_app_process()} or \code{local_app_process()}.
By default the actual process does not start yet, when you create it.
You can start it explicitly with the \verb{$start} method of the app process object, or by querying its URL with \verb{$url()} or its port with \verb{$get_port()}.

For test cases, you typically start app processes at these places:
\itemize{
\item In a \code{setup*.R} file, to start an app that the whole test suite can use.
\item Alternatively, in a \code{helper*.R} file, to start an app that the whole test suite can use, and it works better for interactive development.
\item At the beginning of a test file, to create an app for a single test file.
\item Inside \code{test_that()}, to create an app for a single test block.
}

See the How-to for details about each.
}

\subsection{handler}{

(Or handler function.) A handler is a \emph{route} or a \emph{middleware}.
}

\subsection{handler stack}{

This is a stack of handler functions, which are called by the app one after the other, passing the request and response objects to them.
Handlers typically manipulate the request and/or response objects.
A terminal handler instructs the app to return the response to the HTTP client.
A non-terminal handler tells the app to keep calling handlers, by returning the string \code{"next"}.
}

\subsection{httpbin app}{

This is an example app, which implements the excellent \verb{https://httpbin.org/} web service.
You can use it to simulate certain HTTP responses.
It is most handy for HTTP clients, but potentially useful for other tools as well.

Use \code{httpbin_app()} to create an instance of this app.
}

\subsection{middleware}{

A middleware is a handler function that is not bound to a path.
It is called by the router, like other handler functions.
It may manipulate the request or the response, or can have a side effect.
Some example built-in middleware functions in webfakes:
\itemize{
\item \code{mw_json()} parses a request's JSON body into an R object.
\item \code{mw_log()} logs requests and responses to the screen or to a file.
\item \code{mw_static()} serves static files from the directory.
}

You can also write your own middleware functions.
}

\subsection{path matching}{

The router performs path matching when it goes over the handler stack.
If the HTTP method and path of a \emph{route} match the HTTP method and URL of the request, then the handler is called, otherwise it is not.
Paths can have parameters and be regular expressions.
See \code{?new_regexp()} for regular expressions and "Path parameters" in \code{?new_app()} for parameters.
}

\subsection{route}{

A route is a handler function that is bound to certain paths of you web app.
If the request URL matches the path of the route, then the handler function is called, to give it a chance to send the appropriate response.
Route paths may have parameters or they can be regular expressions in webfakes.
}

\subsection{routing}{

Routing is the process of going over the handlers stack, and calling handler functions, one after the other, until one handles the request.
If a handler function is a \emph{route}, then the router only calls it if its path matches the request URL.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth2_resource_app}
\alias{oauth2_resource_app}
\title{Fake OAuth 2.0 resource and authorization app}
\usage{
oauth2_resource_app(
  access_duration = 3600L,
  refresh_duration = 7200L,
  refresh = TRUE,
  seed = NULL,
  authorize_endpoint = "/authorize",
  token_endpoint = "/token"
)
}
\arguments{
\item{access_duration}{After how many seconds should access tokens
expire.}

\item{refresh_duration}{After how many seconds should refresh
tokens expire (ignored if \code{refresh} is \code{FALSE}).}

\item{refresh}{Should a refresh token be returned (logical).}

\item{seed}{Random seed used when creating tokens. If \code{NULL},
we rely on R to provide a seed. The app uses its own RNG stream,
so it does not affect reproducibility of the tests.}

\item{authorize_endpoint}{The authorization endpoint of the resource
server. Change this from the default if the real app that you
are faking does not use \verb{/authorize}.}

\item{token_endpoint}{The endpoint to request tokens. Change this if the
real app that you are faking does not use \verb{/token}.}
}
\value{
a \code{webfakes} app

webfakes app
}
\description{
The webfakes package comes with two fake apps that allow to imitate the
OAuth2.0 flow in your test cases. (See \href{https://aaronparecki.com/oauth-2-simplified/}{Aaron Parecki’s tutorial} for a good
introduction to OAuth2.0.) One app (\code{oauth2_resource_app()}) is the API
server that serves both as the resource and provides authorization.
\code{oauth2_third_party_app()} plays the role of the third-party app. They
are useful when testing or demonstrating code handling OAuth2.0
authorization, token caching, etc. in a package. The apps can be used in
your tests directly, or you could adapt one or both of them to better
mimic a particular OAuth2.0 flow.
}
\details{
The app has the following endpoints:
\itemize{
\item \code{GET /register} is the endpoint that you can use to register
your third party app. It needs to receive the \code{name} of the
third party app, and its \code{redirect_uri} as query parameters,
otherwise returns an HTTP 400 error. On success it returns a
JSON dictionary with entries \code{name} (the name of the third party
app), \code{client_id}, \code{client_secret} and \code{redirect_uri}.
\item \code{GET /authorize} is the endpoint where the user of the third
party app is sent. You can change the URL of this endpoint with
the \code{authorize_endpoint} argument. It needs to receive the \code{client_id}
of the third party app, and its correct \code{redirect_uri} as query
parameters. It may receive a \code{state} string as well, which can
be used by a client to identify the request. Otherwise it
generates a random \code{state} string. On error it fails with a HTTP
400 error. On success it returns a simple HTML login page.
\item \code{POST /authorize/decision} is the endpoint where the HTML login
page generated at \verb{/authorize} connects back to, either with a
positive or negative result. The form on the login page will send
the \code{state} string and the user's choice in the \code{action} variable.
If the user authorized the third party app, then they are
redirected to the \code{redirect_uri} of the app, with a temporary
\code{code} and the \code{state} string supplied as query parameters.
Otherwise a simple HTML page is returned.
\item \code{POST /token} is the endpoint where the third party app requests
a temporary access token. It is also uses for refreshing an
access token with a refresh token. You can change the URL of this
endpoint with the \code{token_endpoint} argument.
To request a new token or refresh an existing one, the following
data must be included in either a JSON or an URL encoded request body:
\itemize{
\item \code{grant_type}, this must be \code{authorization_code} for new tokens,
and \code{refresh_token} for refreshing.
\item \code{code}, this must be the temporary code obtained from the
\verb{/authorize/decision} redirection, for new tokens. It is not
needed when refreshing.
\item \code{client_id} must be the client id of the third party app.
\item \code{client_secret} must be the client secret of the third party
app.
\item \code{redirect_uri} must be the correct redirection URI of the
third party app. It is not needed when refreshing tokens.
\item \code{refresh_token} must be the refresh token obtained previously,
when refreshing a token. It is not needed for new tokens.
On success a JSON dictionary is returned with entries:
\code{access_token}, \code{expiry} and \code{refresh_token}. (The latter is
omitted if the \code{refresh} argument is \code{FALSE}).
}
\item \code{GET /locals} returns a list of current apps, access tokens and
refresh tokens.
\item \code{GET /data} is an endpoint that returns a simple JSON response,
and needs authorization.
}
\subsection{Notes}{
\itemize{
\item Using this app in your tests requires the glue package, so you
need to put it in \code{Suggests}.
\item You can add custom endpoints to the app, as needed.
\item If you need authorization in your custom endpoint, call
\code{app$is_authorized()} in your handler:

\if{html}{\out{<div class="sourceCode">}}\preformatted{if (!app$is_authorized(req, res)) return()
}\if{html}{\out{</div>}}

\code{app$is_authorized()} returns an HTTP 401 response if the
client is not authorized, so you can simply return from your
handler.
}

For more details see \code{vignette("oauth", package = "webfakes")}.
}
}
\section{\code{oauth2_resource_app()}}{

App representing the API server (resource/authorization)
}

\seealso{
Other OAuth2.0 functions: 
\code{\link{oauth2_httr_login}()},
\code{\link{oauth2_login}()},
\code{\link{oauth2_third_party_app}()}
}
\concept{OAuth2.0 functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tmpl-glue.R
\name{tmpl_glue}
\alias{tmpl_glue}
\title{glue based template engine}
\usage{
tmpl_glue(
  sep = "",
  open = "{",
  close = "}",
  na = "NA",
  transformer = NULL,
  trim = TRUE
)
}
\arguments{
\item{sep}{Separator used to separate elements.}

\item{open}{The opening delimiter. Doubling the full delimiter escapes
it.}

\item{close}{The closing delimiter. Doubling the full delimiter escapes
it.}

\item{na}{Value to replace NA values with. If \code{NULL} missing values are
propagated, that is an \code{NA} result will cause \code{NA} output.
Otherwise the value is replaced by the value of \code{na}.}

\item{transformer}{A function taking three parameters \code{code}, \code{envir}
and \code{data} used to transform the output of each block before during or
after evaluation.}

\item{trim}{Whether to trim the input template with \code{\link[glue:trim]{glue::trim()}} or not.}
}
\value{
Template function.
}
\description{
Use this template engine to create pages with glue templates.
See \code{\link[glue:glue]{glue::glue()}} for the syntax.
}
\examples{
# See th 'hello' app at
hello_root <- system.file(package = "webfakes", "examples", "hello")
hello_root

app <- new_app()
app$engine("txt", tmpl_glue())
app$use(mw_log())


app$get("/view", function(req, res) {
  txt <- res$render("test")
  res$
    set_type("text/plain")$
    send(txt)
})

# Switch to the app's root: setwd(hello_root)
# Now start the app with: app$listen(3000L)
# Or start it in another process: new_process(app)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oauth.R
\name{oauth2_third_party_app}
\alias{oauth2_third_party_app}
\title{App representing the third-party app}
\usage{
oauth2_third_party_app(name = "Third-Party app")
}
\arguments{
\item{name}{Name of the third-party app}
}
\value{
webfakes app
}
\description{
The webfakes package comes with two fake apps that allow to imitate the
OAuth2.0 flow in your test cases. (See \href{https://aaronparecki.com/oauth-2-simplified/}{Aaron Parecki’s tutorial} for a good
introduction to OAuth2.0.) One app (\code{oauth2_resource_app()}) is the API
server that serves both as the resource and provides authorization.
\code{oauth2_third_party_app()} plays the role of the third-party app. They
are useful when testing or demonstrating code handling OAuth2.0
authorization, token caching, etc. in a package. The apps can be used in
your tests directly, or you could adapt one or both of them to better
mimic a particular OAuth2.0 flow.
}
\details{
Endpoints:
\itemize{
\item \code{POST /login/config} Use this endpoint to configure the client ID
and the client secret of the app, received from
\code{\link[=oauth2_resource_app]{oauth2_resource_app()}} (or another resource app). You need to
send in a JSON or URL encoded body:
\itemize{
\item \code{auth_url}, the authorization URL of the resource app.
\item \code{token_url}, the token URL of the resource app.
\item \code{client_id}, the client ID, received from the resource app.
\item \code{client_secret} the client secret, received from the resource
app.
}
\item \code{GET /login} Use this endpoint to start the login process. It
will redirect to the resource app for authorization and after the
OAuth2.0 dance to \verb{/login/redirect}.
\item \code{GET /login/redirect}, \code{POST /login/redirect} This is the
redirect URI of the third party app. (Some HTTP clients redirect
a \code{POST} to a \code{GET}, others don't, so it has both.) This endpoint
is used by the resource app, and it received the \code{code} that can
be exchanged to an access token and the \code{state} which was
generated in \verb{/login}. It contacts the resource app to get an
access token, and then stores the token in its \code{app$locals}
local variables. It fails with HTTP code 500 if it cannot obtain
an access token. On success it returns a JSON dictionary with
\code{access_token}, \code{expiry} and \code{refresh_token} (optionally) by
default. This behavior can be changed by redefining the
\code{app$redirect_hook()} function.
\item \code{GET /locals} returns the tokens that were obtained from the
resource app.
\item \code{GET /data} is an endpoint that uses the obtained token(s) to
connect to the \verb{/data} endpoint of the resource app. The \verb{/data}
endpoint of the resource app needs authorization. It responds
with the response of the resource app. It tries to refresh the
access token of the app if needed.
}

For more details see \code{vignette("oauth", package = "webfakes")}.
}
\seealso{
Other OAuth2.0 functions: 
\code{\link{oauth2_httr_login}()},
\code{\link{oauth2_login}()},
\code{\link{oauth2_resource_app}()}
}
\concept{OAuth2.0 functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-multipart.R
\name{mw_multipart}
\alias{mw_multipart}
\title{Parse a multipart HTTP request body}
\usage{
mw_multipart(type = "multipart/form-data")
}
\arguments{
\item{type}{Content type to match before parsing. If it does not
match, then the request object is not modified.}
}
\value{
Handler function.
}
\description{
Adds the parsed form fields in the \code{form} element of the request and
the parsed files to the \code{files} element.
}
\examples{
app <- new_app()
app$use(mw_multipart())
app
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_json}()},
\code{\link{mw_log}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-static.R
\name{mw_static}
\alias{mw_static}
\title{Middleware function to serve static files}
\usage{
mw_static(root, set_headers = NULL)
}
\arguments{
\item{root}{Root path of the served files. Everything under this
directory is served automatically. Directory lists are not currently
supports.}

\item{set_headers}{Callback function to call before a file is served.}
}
\value{
Handler function.
}
\description{
The content type of the response is set automatically from the
extension of the file. Note that this is a terminal middleware
handler function. If a file is served, then the rest of the handler
functions will not be called. If a file was not found, however,
the rest of the handlers are still called.
}
\examples{
root <- system.file(package = "webfakes", "examples", "static", "public")
app <- new_app()
app$use(mw_static(root = root))
app
}
\seealso{
Other middleware:
\code{\link[=mw_cgi]{mw_cgi()}},
\code{\link[=mw_cookie_parser]{mw_cookie_parser()}},
\code{\link[=mw_etag]{mw_etag()}},
\code{\link[=mw_json]{mw_json()}},
\code{\link[=mw_log]{mw_log()}},
\code{\link[=mw_multipart]{mw_multipart()}},
\code{\link[=mw_range_parser]{mw_range_parser()}},
\code{\link[=mw_raw]{mw_raw()}},
\code{\link[=mw_text]{mw_text()}},
\code{\link[=mw_urlencoded]{mw_urlencoded()}}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/request.R
\name{webfakes_request}
\alias{webfakes_request}
\title{A webfakes request object}
\description{
webfakes creates a \code{webfakes_request} object for every incoming HTTP
request. This object is passed to every matched route and middleware,
until the response is sent. It has reference semantics, so handlers
can modify it.
}
\details{
Fields and methods:
\itemize{
\item \code{app}: The \code{webfakes_app} object itself.
\item \code{headers}: Named list of HTTP request headers.
\item \code{hostname}: The Host header, the server hostname and maybe port.
\item \code{method}: HTTP method.
\item \code{path}: Server path.
\item \code{protocol}: \code{"http"} or \code{"https"}.
\item \code{query_string}: The raw query string, without the starting \verb{?}.
\item \code{query}: Parsed query parameters in a named list.
\item \code{remote_addr}: String, the domain name or IP address of the client.
webfakes runs on the localhost, so this is \verb{127.0.0.1}.
\item \code{url}: The full URL of the request.
\item \code{get_header(field)}: Function to query a request header. Returns
\code{NULL} if the header is not present.
}

Body parsing middleware adds additional fields to the request object.
See \code{\link[=mw_raw]{mw_raw()}}, \code{\link[=mw_text]{mw_text()}}, \code{\link[=mw_json]{mw_json()}}, \code{\link[=mw_multipart]{mw_multipart()}} and
\code{\link[=mw_urlencoded]{mw_urlencoded()}}.
}
\examples{
# This is how you can see the request and response objects:
app <- new_app()
app$get("/", function(req, res) {
  browser()
  res$send("done")
})
app

# Now start this app on a port:
# app$listen(3000)
# and connect to it from a web browser: http://127.0.0.1:3000
# You can also use another R session to connect:
# httr::GET("http://127.0.0.1:3000")
# or the command line curl tool:
# curl -v http://127.0.0.1:3000
# The app will stop while processing the request.
}
\seealso{
\link{webfakes_response} for the webfakes response object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/httpbin.R
\name{httpbin_app}
\alias{httpbin_app}
\title{Generic web app for testing HTTP clients}
\usage{
httpbin_app(log = interactive())
}
\arguments{
\item{log}{Whether to log requests to the standard output.}
}
\value{
A \code{webfakes_app}.
}
\description{
A web app similar to \verb{https://httpbin.org}.
See \href{https://webfakes.r-lib.org/httpbin.html}{its specific docs}.
You can also see these docs locally, by starting the app:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{httpbin <- new_app_process(httpbin_app())
browseURL(httpbin$url())
}\if{html}{\out{</div>}}
}
\examples{
app <- httpbin_app()
proc <- new_app_process(app)
url <- proc$url("/get")
resp <- curl::curl_fetch_memory(url)
curl::parse_headers_list(resp$headers)
cat(rawToChar(resp$content))
proc$stop()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-range-parser.R
\name{mw_range_parser}
\alias{mw_range_parser}
\title{Middleware to parse a Range header}
\usage{
mw_range_parser()
}
\value{
Handler function.
}
\description{
Adds the requested ranges to the \code{ranges} element of the request
object. \code{request$ranges} is a data frame with two columns, \code{from} and
\code{to}. Each row corresponds one requested interval.
}
\details{
When the last \code{n} bytes of the file are requested, the matrix row is set
to \code{c(0, -n)}. When all bytes after a \code{p} position are requested, the
matrix row is set to \code{c(p, Inf)}.

If the intervals overlap, then \code{ranges} is not set, i.e. the \code{Range}
header is ignored.

If its syntax is invalid or the unit is not \code{bytes}, then the
\code{Range} header is ignored.
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_json}()},
\code{\link{mw_log}()},
\code{\link{mw_multipart}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mw-log.R
\name{mw_log}
\alias{mw_log}
\title{Log requests to the standard output or other connection}
\usage{
mw_log(format = "dev", stream = "stdout")
}
\arguments{
\item{format}{Log format. Not implemented currently.}

\item{stream}{R connection to log to. \code{"stdout"} means the standard
output, \code{"stderr"} is the standard error. You can also supply a
connection object, but then you need to be sure that it will be
valid when the app is actually running.}
}
\value{
Handler function.
}
\description{
A one line log entry for every request. The output looks like this:

\if{html}{\out{<div class="sourceCode">}}\preformatted{GET http://127.0.0.1:3000/image 200 3 ms - 4742
}\if{html}{\out{</div>}}

and contains
\itemize{
\item the HTTP method,
\item the full request URL,
\item the HTTP status code of the response,
\item how long it took to process the response, in ms,
\item and the size of the response body, in bytes.
}
}
\examples{
app <- new_app()
app$use(mw_log())
app
}
\seealso{
Other middleware: 
\code{\link{mw_cgi}()},
\code{\link{mw_cookie_parser}()},
\code{\link{mw_etag}()},
\code{\link{mw_json}()},
\code{\link{mw_multipart}()},
\code{\link{mw_range_parser}()},
\code{\link{mw_raw}()},
\code{\link{mw_static}()},
\code{\link{mw_text}()},
\code{\link{mw_urlencoded}()}
}
\concept{middleware}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{http_time_stamp}
\alias{http_time_stamp}
\title{Format a time stamp for HTTP}
\usage{
http_time_stamp(t = Sys.time())
}
\arguments{
\item{t}{Date-time value to format, defaults to the current date and
time. It must be a POSIXct object.}
}
\value{
Character vector, formatted date-time.
}
\description{
Format a time stamp for HTTP
}
